#include "CompressHeightfield.h"

#include <cuda_runtime.h>

#include "tools/BitStream.h"
#include "tools/Vec.h"
//#include "tools/rawfile.h"

#include "reduce/reduce.h"

#include "cudaUtil.h"
#include "util.h"

#include "Instance.h"
#include "Bits.h"
#include "DWT.h"
#include "Encode.h"
#include "Quantize.h"

using namespace tum3D;


namespace cudaCompress {


size_t compressHeightfieldGetRequiredMemory(const void* pCudaCompressInstance)
{
	Instance* pInstance = (Instance*)pCudaCompressInstance;

	const uint blockCount = 3;

	Vec2i sizeMax(pInstance->m_sizeMax);
	uint elemCountTotal = sizeMax.area();
	Vec2i blockSizeMax = ((sizeMax + 1) / 2);
	uint elemCountPerBlock = blockSizeMax.area();

	uint bitsSize = (blockSizeMax.area() + 31) / 32;

	size_t size = 0;

	// dpDWT
	size += getAlignedSize(elemCountTotal * sizeof(short), 128);
	size += getAlignedSize(elemCountTotal * sizeof(short), 128);
	// dpSymbolStreams
	size += getAlignedSize(blockCount * elemCountPerBlock * sizeof(Symbol), 128);
	// dpBits
	size += getAlignedSize(bitsSize * sizeof(uint), 128);

	return size;
}

bool compressHeightfieldInit(void* pCudaCompressInstance)
{
	Instance* pInstance = (Instance*)pCudaCompressInstance;

	const uint blockCount = 3;

	Vec2i sizeMax(pInstance->m_sizeMax);
	uint elemCountTotal = sizeMax.area();
	Vec2i blockSizeMax = ((sizeMax + 1) / 2);
	uint elemCountPerBlock = blockSizeMax.area();

	uint bitsSize = (blockSizeMax.area() + 31) / 32;

	cudaSafeCall(cudaMallocHost(&pInstance->CompressHeightfield.pBitsUpload, bitsSize * sizeof(uint), cudaHostAllocWriteCombined));
	cudaSafeCall(cudaMallocHost(&pInstance->CompressHeightfield.pBitsDownload, bitsSize * sizeof(uint)));

	cudaSafeCall(cudaEventCreateWithFlags(&pInstance->CompressHeightfield.syncEventUpload, cudaEventDisableTiming));
	cudaSafeCall(cudaEventCreateWithFlags(&pInstance->CompressHeightfield.syncEventDownload, cudaEventDisableTiming));
	// record upload event immediately to signal that buffer is ready for use
	cudaSafeCall(cudaEventRecord(pInstance->CompressHeightfield.syncEventUpload));

	return true;
}

bool compressHeightfieldShutdown(void* pCudaCompressInstance)
{
	Instance* pInstance = (Instance*)pCudaCompressInstance;

	cudaSafeCall(cudaEventDestroy(pInstance->CompressHeightfield.syncEventDownload));
	pInstance->CompressHeightfield.syncEventDownload = 0;
	cudaSafeCall(cudaEventDestroy(pInstance->CompressHeightfield.syncEventUpload));
	pInstance->CompressHeightfield.syncEventUpload = 0;

	cudaSafeCall(cudaFreeHost(pInstance->CompressHeightfield.pBitsDownload));
	pInstance->CompressHeightfield.pBitsDownload = nullptr;
	cudaSafeCall(cudaFreeHost(pInstance->CompressHeightfield.pBitsUpload));
	pInstance->CompressHeightfield.pBitsUpload = nullptr;

	return true;
}


bool compressHeightfieldOneLevel(void* pCudaCompressInstance, const short* dpImage, uint sizeX, uint sizeY, short* dpLowpass, short lowpassShift, std::vector<uint>& highpassBitStream)
{
	const uint blockCount = 3;

	Vec2i size(sizeX, sizeY);
	uint elemCountTotal = size.area();
	Vec2i blockSize = ((size + 1) / 2);
	uint elemCountPerBlock = blockSize.area();

	uint bitsSize = (elemCountPerBlock + 31) / 32;

	Instance* pInstance = (Instance*)pCudaCompressInstance;

	short* dpDWT0 = pInstance->getBuffer<short>(elemCountTotal);
	short* dpDWT1 = pInstance->getBuffer<short>(elemCountTotal);
	std::vector<Symbol*> dpSymbolStreams(blockCount);
	for(uint i = 0; i < blockCount; i++) {
		dpSymbolStreams[i] = pInstance->getBuffer<Symbol>(elemCountPerBlock);
	}
	uint* dpBits = pInstance->getBuffer<uint>((blockSize.area() + 31) / 32);

	CudaScopedTimer timer(pInstance->CompressHeightfield.timerEncode);

	// perform DWT
	timer("DWT");
	dwtIntForward(dpDWT1, dpDWT0, dpImage, size.x(), size.y(), 1);

	// copy lowpass band
	timer("Copy LP");
	cudaSafeCall(cudaMemcpy2DAsync(dpLowpass, blockSize.x() * sizeof(short), dpDWT1, size.x() * sizeof(short), blockSize.x() * sizeof(short), blockSize.y(), cudaMemcpyDeviceToDevice));

	// truncate and download lsb of lowpass and apply shift
	timer("Process LP");
	removeLSB(dpLowpass, dpBits, elemCountPerBlock, lowpassShift);
	cudaSafeCall(cudaMemcpyAsync(pInstance->CompressHeightfield.pBitsDownload, dpBits, bitsSize * sizeof(uint), cudaMemcpyDeviceToHost));
	cudaSafeCall(cudaEventRecord(pInstance->CompressHeightfield.syncEventDownload));

	// symbolize highpass bands
	timer("Symbolize HP");
	uint nextSymbolStream = 0;
	for(uint block = 1; block < 1 + blockCount; block++) {
		uint x = block % 2;
		uint y = block / 2;
		uint offset = x * blockSize.x() + y * blockSize.y() * size.x();

		// make (unsigned!) symbols
		symbolize2D(dpSymbolStreams[nextSymbolStream++], dpDWT1 + offset, blockSize.x(), blockSize.y(), size.x());
	}

	BitStream bitStream(&highpassBitStream);

	// store lsb of lowpass
	timer("Store LP LSB");
	cudaSafeCall(cudaEventSynchronize(pInstance->CompressHeightfield.syncEventDownload));
	bitStream.writeAligned(pInstance->CompressHeightfield.pBitsDownload, bitsSize);

	// encode
	timer("Encode");
	bitStream.reserveBitSize(bitStream.getBitPosition() + sizeX * sizeY * 16);
	BitStream* pBitStream = &bitStream;
	bool result = encodeRLHuff(pInstance, &pBitStream, true, &dpSymbolStreams[0], blockCount, elemCountPerBlock);

	timer();

	pInstance->releaseBuffers(blockCount + 3);

	return result;
}

void decompressHeightfieldOneLevel(void* pCudaCompressInstance, short* dpImage, uint sizeX, uint sizeY, short* dpLowpass, short lowpassShift, const std::vector<uint>& highpassBitStream)
{
	decompressHeightfieldOneLevel(pCudaCompressInstance, dpImage, sizeX, sizeY, dpLowpass, lowpassShift, highpassBitStream.data(), uint(highpassBitStream.size() * sizeof(uint) * 8));
}

void decompressHeightfieldOneLevel(void* pCudaCompressInstance, short* dpImage, uint sizeX, uint sizeY, short* dpLowpass, short lowpassShift, const uint* pHighpassBits, uint highpassBitCount)
{
	const uint blockCount = 3;

	Vec2i size(sizeX, sizeY);
	uint elemCountTotal = size.area();
	Vec2i blockSize = ((size + 1) / 2);
	uint elemCountPerBlock = blockSize.area();

	uint bitsSize = (blockSize.area() + 31) / 32;

	Instance* pInstance = (Instance*)pCudaCompressInstance;

	short* dpDWT0 = pInstance->getBuffer<short>(elemCountTotal);
	short* dpDWT1 = pInstance->getBuffer<short>(elemCountTotal);
	std::vector<Symbol*> dpSymbolStreams(blockCount);
	for(uint i = 0; i < blockCount; i++) {
		dpSymbolStreams[i] = pInstance->getBuffer<Symbol>(elemCountPerBlock);
	}
	uint* dpBits = pInstance->getBuffer<uint>((blockSize.area() + 31) / 32);

	CudaScopedTimer timer(pInstance->CompressHeightfield.timerDecode);

	BitStreamReadOnly bitStream(pHighpassBits, highpassBitCount);

	// read lowpass lsb from bitstream and upload
	// wait for previous upload to complete before overwriting cpu buffer
	timer("Sync Prev");
	cudaSafeCall(cudaEventSynchronize(pInstance->CompressHeightfield.syncEventUpload));
	bitStream.readAligned(pInstance->CompressHeightfield.pBitsUpload, bitsSize);
	cudaSafeCall(cudaMemcpyAsync(dpBits, pInstance->CompressHeightfield.pBitsUpload, bitsSize * sizeof(uint), cudaMemcpyHostToDevice));
	cudaSafeCall(cudaEventRecord(pInstance->CompressHeightfield.syncEventUpload));

	// append lsb to lowpass and apply shift
	timer("Append LP LSB");
	appendLSB(dpLowpass, dpBits, blockSize.area(), lowpassShift);

	// copy lowpass band into larger buffer
	timer("Copy LP");
	cudaSafeCall(cudaMemcpy2DAsync(dpDWT1, size.x() * sizeof(short), dpLowpass, blockSize.x() * sizeof(short), blockSize.x() * sizeof(short), blockSize.y(), cudaMemcpyDeviceToDevice));

	// decode
	timer("Decode");
	BitStreamReadOnly* pBitStream = &bitStream;
	decodeRLHuff(pInstance, &pBitStream, true, &dpSymbolStreams[0], blockCount, elemCountPerBlock);

	// unsymbolize highpass bands
	timer("Unsymbolize HP");
	uint nextSymbolStream = 0;
	for(uint block = 1; block < 1 + blockCount; block++) {
		uint x = block % 2;
		uint y = block / 2;
		uint offset = x * blockSize.x() + y * blockSize.y() * size.x();

		// get signed values back from unsigned symbols
		unsymbolize2D(dpDWT1 + offset, dpSymbolStreams[nextSymbolStream++], blockSize.x(), blockSize.y(), size.x());
	}

	// perform IDWT
	timer("IDWT");
	dwtIntInverse(dpImage, dpDWT0, dpDWT1, size.x(), size.y(), 1);

	pInstance->releaseBuffers(blockCount + 3);
}

void getLowpassHeightfield(void* pCudaCompressInstance, const short* dpImage, uint sizeX, uint sizeY, short* dpLowpass, short lowpassShift)
{
	Vec2i size(sizeX, sizeY);
	uint elemCountTotal = size.area();
	Vec2i blockSize = ((size + 1) / 2);

	Instance* pInstance = (Instance*)pCudaCompressInstance;

	short* dpDWT = pInstance->getBuffer<short>(elemCountTotal);

	// perform DWT
	dwtIntForwardLowpassOnly(dpLowpass, dpDWT, dpImage, size.x(), size.y());

	// truncate lsb of lowpass and apply shift
	removeLSB(dpLowpass, 0, blockSize.area(), lowpassShift);
	
	pInstance->releaseBuffer();
}

void getHeightfieldMinMax(void* pCudaCompressInstance, const short* dpImage, uint sizeX, uint sizeY, short* dpValMin, short* dpValMax)
{
	Instance* pInstance = (Instance*)pCudaCompressInstance;

	if(dpValMin != 0) {
		reduceMinShort(dpValMin, dpImage, sizeX * sizeY, pInstance->m_pReducePlan);
		cudaCheckMsg("getHeightfieldMinMax: Error in reduceMinShort");
	}
	if(dpValMax != 0) {
		reduceMaxShort(dpValMax, dpImage, sizeX * sizeY, pInstance->m_pReducePlan);
		cudaCheckMsg("getHeightfieldMinMax: Error in reduceMaxShort");
	}
}

void shiftHeightfield(void* pCudaCompressInstance, short* dpImage, uint sizeX, uint sizeY, short shift)
{
	shiftElems(dpImage, sizeX * sizeY, shift);
}

}
