#include "CompressImage.h"

//#include <string>

#include <cuda_runtime.h>

#include "tools/BitStream.h"
#include "tools/Vec.h"

#include "cudaUtil.h"
#include "util.h"

#include "Instance.h"
#include "DWT.h"
#include "Quantize.h"
#include "Encode.h"

using namespace tum3D;


namespace cudaCompress {


static const uint g_blockCountPerChannel = 4;


static uint getUploadSizeBytes(const Instance* pInstance)
{
	uint blockCountTotal = pInstance->m_channelCountMax * g_blockCountPerChannel;
	return blockCountTotal * sizeof(Symbol*);
}


size_t compressImageGetRequiredMemory(const void* pCudaCompressInstance)
{
	const Instance* pInstance = (const Instance*)pCudaCompressInstance;

	const uint blockCountTotal = pInstance->m_channelCountMax * g_blockCountPerChannel;

	Vec2i sizeMax(pInstance->m_sizeMax);
	uint elemCountTotal = sizeMax.area();
	Vec2i blockSizeMax = ((sizeMax + 1) / 2);
	uint elemCountPerBlock = blockSizeMax.area();

	size_t size = 0;

	// dpDWT
	size += getAlignedSize(elemCountTotal * sizeof(float), 128);
	size += getAlignedSize(elemCountTotal * sizeof(float), 128);

	// dpSymbolStreams
	for(uint i = 0; i < blockCountTotal; i++) {
		size += getAlignedSize(elemCountPerBlock * sizeof(Symbol), 128);
	}

	return size;
}

bool compressImageInit(void* pCudaCompressInstance)
{
	Instance* pInstance = (Instance*)pCudaCompressInstance;

	uint uploadSize = getUploadSizeBytes(pInstance);
	cudaSafeCall(cudaMallocHost(&pInstance->CompressImage.pUpload, uploadSize, cudaHostAllocWriteCombined));

	cudaSafeCall(cudaEventCreateWithFlags(&pInstance->CompressImage.syncEventUpload, cudaEventDisableTiming));
	// immediately record to signal that buffers are ready to use (ie first cudaEventSynchronize works)
	cudaSafeCall(cudaEventRecord(pInstance->CompressImage.syncEventUpload));

	return true;
}

bool compressImageShutdown(void* pCudaCompressInstance)
{
	Instance* pInstance = (Instance*)pCudaCompressInstance;

	cudaSafeCall(cudaEventDestroy(pInstance->CompressImage.syncEventUpload));
	pInstance->CompressImage.syncEventUpload = 0;

	cudaSafeCall(cudaFreeHost(pInstance->CompressImage.pUpload));
	pInstance->CompressImage.pUpload = nullptr;

	return true;
}

bool compressImageOneLevel(void* pCudaCompressInstance, byte* dpImage, uint sizeX, uint sizeY, uint channelCountTotal, uint channelCountToProcess, const byte* dpLowpassReference, std::vector<uint>& highpassBitStream, float quantizationStep)
{
	const uint blockCountTotal = channelCountToProcess * g_blockCountPerChannel;

	Vec2i size(sizeX, sizeY);
	uint elemCountTotal = size.area();
	Vec2i blockSize = ((size + 1) / 2);
	uint elemCountPerBlock = blockSize.area();

	Instance* pInstance = (Instance*)pCudaCompressInstance;

	float* dpDWT0 = pInstance->getBuffer<float>(size.area());
	float* dpDWT1 = pInstance->getBuffer<float>(size.area());
	std::vector<Symbol*> dpSymbolStreams(blockCountTotal);
	for(uint i = 0; i < blockCountTotal; i++) {
		dpSymbolStreams[i] = pInstance->getBuffer<Symbol>(elemCountPerBlock);
	}

	CudaScopedTimer timer(pInstance->CompressImage.timerEncode);

	uint nextSymbolStreamLowpass = 0;
	uint nextSymbolStreamHighpass = channelCountToProcess;
	for(uint channel = 0; channel < channelCountToProcess; channel++) {
		// perform DWT
		timer("DWT");
		dwtFloat2DForwardFromByte(dpDWT1, dpDWT0, dpImage + channel, size.x(), size.y(), channelCountTotal);

		// roundtrip-quantize lowpass band, store differences to reference lowpass
		timer("Quantize LP");
		quantizeDifferenceToSymbolsRoundtrip2D(dpSymbolStreams[nextSymbolStreamLowpass++], dpDWT1, quantizationStep, dpLowpassReference, channelCountTotal, channel, blockSize.x(), blockSize.y(), size.x());

		// roundtrip-quantize highpass bands
		timer("Quantize HP");
		int offsetX = blockSize.x();
		int offsetY = blockSize.y() * size.x();
		quantizeToSymbolsRoundtrip2D(dpSymbolStreams[nextSymbolStreamHighpass++], dpDWT1 + offsetX,           quantizationStep, blockSize.x(), blockSize.y(), size.x());
		quantizeToSymbolsRoundtrip2D(dpSymbolStreams[nextSymbolStreamHighpass++], dpDWT1 +           offsetY, quantizationStep, blockSize.x(), blockSize.y(), size.x());
		quantizeToSymbolsRoundtrip2D(dpSymbolStreams[nextSymbolStreamHighpass++], dpDWT1 + offsetX + offsetY, quantizationStep, blockSize.x(), blockSize.y(), size.x());

		// perform IDWT to complete roundtrip
		timer("IDWT");
		dwtFloat2DInverseToByte(dpImage + channel, dpDWT0, dpDWT1, size.x(), size.y(), channelCountTotal);
	}

	timer();

	BitStream bitStream(&highpassBitStream);
	bitStream.reserveBitSize(bitStream.getBitPosition() + sizeX * sizeY * 24);
	BitStream* pBitStream = &bitStream;

	// encode lowpass increments + highpass
	timer("Encode LP Increments + HP");
	bool result = encodeRLHuff(pInstance, &pBitStream, true, dpSymbolStreams.data(), blockCountTotal, elemCountPerBlock);

	timer();

	pInstance->releaseBuffers(blockCountTotal + 2);

	return result;
}

void decompressImageOneLevel(void* pCudaCompressInstance, byte* dpImage, uint sizeX, uint sizeY, uint channelCountTotal, uint channelCountToProcess, const byte* dpLowpassReference, const std::vector<uint>& highpassBitStream, float quantizationStep)
{
	decompressImageOneLevel(pCudaCompressInstance, dpImage, sizeX, sizeY, channelCountTotal, channelCountToProcess, dpLowpassReference, highpassBitStream.data(), uint(highpassBitStream.size() * sizeof(uint) * 8), quantizationStep);
}

void decompressImageOneLevel(void* pCudaCompressInstance, byte* dpImage, uint sizeX, uint sizeY, uint channelCountTotal, uint channelCountToProcess, const byte* dpLowpassReference, const uint* pHighpassBits, uint highpassBitCount, float quantizationStep)
{
	const uint blockCountTotal = channelCountToProcess * g_blockCountPerChannel;

	Vec2i size(sizeX, sizeY);
	uint elemCountTotal = size.area();
	Vec2i blockSize = ((size + 1) / 2);
	uint elemCountPerBlock = blockSize.area();

	Instance* pInstance = (Instance*)pCudaCompressInstance;

	float* dpDWT0 = pInstance->getBuffer<float>(size.area());
	float* dpDWT1 = pInstance->getBuffer<float>(size.area());
	std::vector<Symbol*> dpSymbolStreams(blockCountTotal);
	for(uint i = 0; i < blockCountTotal; i++) {
		dpSymbolStreams[i] = pInstance->getBuffer<Symbol>(elemCountPerBlock);
	}

	Symbol** dppSymbolStreams = pInstance->getBuffer<Symbol*>(blockCountTotal);
	cudaSafeCall(cudaEventSynchronize(pInstance->CompressImage.syncEventUpload));
	memcpy(pInstance->CompressImage.pUpload, dpSymbolStreams.data(), blockCountTotal * sizeof(Symbol*));
	cudaSafeCall(cudaMemcpyAsync(dppSymbolStreams, pInstance->CompressImage.pUpload, blockCountTotal * sizeof(Symbol*), cudaMemcpyHostToDevice));
	cudaSafeCall(cudaEventRecord(pInstance->CompressImage.syncEventUpload));

	BitStreamReadOnly bitStream(pHighpassBits, highpassBitCount);
	BitStreamReadOnly* pBitStream = &bitStream;

	CudaScopedTimer timer(pInstance->CompressImage.timerEncode);

	// decode lowpass increments + highpass
	timer("Decode LP Increments + HP");
	decodeRLHuff(pInstance, &pBitStream, true, dpSymbolStreams.data(), blockCountTotal, elemCountPerBlock);

	uint nextSymbolStreamLowpass = 0;
	uint nextSymbolStreamHighpass = channelCountToProcess;
	for(uint channel = 0; channel < channelCountToProcess; channel++) {
		// unquantize lowpass band, add reference lowpass
		timer("Unquantize LP");
		unquantizeDifferenceFromSymbols2D(dpDWT0, dpSymbolStreams[nextSymbolStreamLowpass++], quantizationStep, dpLowpassReference, channelCountTotal, channel, blockSize.x(), blockSize.y(), size.x());

		// unquantize highpass bands and perform IDWT
		timer("Unquantize HP + IDWT");
		dwtFloat2DInverseFromSymbolsToByte(dpImage + channel, dpDWT1, dpDWT0, dppSymbolStreams + nextSymbolStreamHighpass, quantizationStep, size.x(), size.y(), channelCountTotal);
		nextSymbolStreamHighpass += 3;
	}

	timer();

	pInstance->releaseBuffers(blockCountTotal + 3);
}

void getLowpassImage(void* pCudaCompressInstance, const byte* dpImage, uint sizeX, uint sizeY, uint channelCountTotal, uint channelCountToProcess, byte* dpLowpass)
{
	Vec2i size(sizeX, sizeY);
	uint elemCountTotal = size.area();
	Vec2i blockSize = ((size + 1) / 2);

	Instance* pInstance = (Instance*)pCudaCompressInstance;

	float* dpDWT0 = pInstance->getBuffer<float>(size.area());
	float* dpDWT1 = pInstance->getBuffer<float>(size.area());

	for(uint channel = 0; channel < channelCountToProcess; channel++) {
		// perform DWT
		dwtFloat2DForwardLowpassOnlyFromByte(dpDWT1, dpDWT0, dpImage + channel, size.x(), size.y(), channelCountTotal);

		// quantize lowpass band
		quantize2D(dpLowpass, channelCountTotal, channel, dpDWT1, blockSize.x(), blockSize.y(), blockSize.x());
	}

	pInstance->releaseBuffers(2);
}


}
