#include "CompressVolume.h"

#include <cuda_runtime.h>

#include "tools/BitStream.h"
#include "tools/Vec.h"

#include "reduce/reduce.h"

#include "cudaUtil.h"
#include "util.h"

#include "Instance.h"
#include "DWT.h"
#include "Encode.h"
#include "Quantize.h"

using namespace tum3D;


namespace cudaCompress {


static uint getUploadSizeBytes(const Instance* pInstance)
{
	const uint blockCount = 7;
	return blockCount * sizeof(Symbol*);
}


size_t compressVolumeGetRequiredMemory(const void* pCudaCompressInstance)
{
	const Instance* pInstance = (const Instance*)pCudaCompressInstance;

	if(pInstance->m_sizeMax.z() <= 1) return 0;

	const uint blockCount = 7;

	uint elemCountTotal = pInstance->m_sizeMax.volume();
	Vec3i blockSizeMax = ((pInstance->m_sizeMax + 1) / 2);
	uint elemCountPerBlock = blockSizeMax.volume();

	size_t size = 0;

	// dpDWT
	size += getAlignedSize(elemCountTotal * sizeof(float), 128);
	size += getAlignedSize(elemCountTotal * sizeof(float), 128);

	// dpSymbolStreams
	for(uint i = 0; i < blockCount; i++) {
		size += getAlignedSize(elemCountPerBlock * sizeof(Symbol), 128);
	}

	return size;
}

bool compressVolumeInit(void* pCudaCompressInstance)
{
	Instance* pInstance = (Instance*)pCudaCompressInstance;

	uint uploadSize = getUploadSizeBytes(pInstance);
	cudaSafeCall(cudaMallocHost(&pInstance->CompressVolume.pUpload, uploadSize, cudaHostAllocWriteCombined));

	cudaSafeCall(cudaEventCreateWithFlags(&pInstance->CompressVolume.syncEventUpload, cudaEventDisableTiming));
	// immediately record to signal that buffers are ready to use (ie first cudaEventSynchronize works)
	cudaSafeCall(cudaEventRecord(pInstance->CompressVolume.syncEventUpload));

	return true;
}

bool compressVolumeShutdown(void* pCudaCompressInstance)
{
	Instance* pInstance = (Instance*)pCudaCompressInstance;

	cudaSafeCall(cudaEventDestroy(pInstance->CompressVolume.syncEventUpload));
	pInstance->CompressVolume.syncEventUpload = 0;

	cudaSafeCall(cudaFreeHost(pInstance->CompressVolume.pUpload));
	pInstance->CompressVolume.pUpload = nullptr;

	return true;
}


bool compressVolumeLosslessOneLevel(void* pCudaCompressInstance, const short* dpImage, uint sizeX, uint sizeY, uint sizeZ, short* dpLowpass, std::vector<uint>& highpassBitStream)
{
	const uint blockCount = 7;

	Vec3i size(sizeX, sizeY, sizeZ);
	Vec3i blockSize = ((size + 1) / 2);


	Instance* pInstance = (Instance*)pCudaCompressInstance;

	short* dpDWT0 = pInstance->getBuffer<short>(size.volume());
	short* dpDWT1 = pInstance->getBuffer<short>(size.volume());
	std::vector<Symbol*> dpSymbolStreams(blockCount);
	for(uint i = 0; i < blockCount; i++) {
		dpSymbolStreams[i] = pInstance->getBuffer<Symbol>(blockSize.volume());
	}

	CudaScopedTimer timer(pInstance->CompressVolume.timerEncode);


	// perform DWT
	timer("DWT");
	dwtIntForward(dpDWT1, dpDWT0, dpImage, size.x(), size.y(), size.z());

	// copy lowpass band
	timer("Copy LP");
	cudaMemcpy3DParms params = { 0 };
	params.srcPtr.ptr = dpDWT1;
	params.srcPtr.pitch = size.x() * sizeof(short);
	params.srcPtr.xsize = size.x() * sizeof(short);
	params.srcPtr.ysize = size.y();
	params.dstPtr.ptr = dpLowpass;
	params.dstPtr.pitch = blockSize.x() * sizeof(short);
	params.dstPtr.xsize = blockSize.x() * sizeof(short);
	params.dstPtr.ysize = blockSize.y();
	params.extent.width  = blockSize.x() * sizeof(short);
	params.extent.height = blockSize.y();
	params.extent.depth  = blockSize.z();
	params.kind = cudaMemcpyDeviceToDevice;
	cudaSafeCall(cudaMemcpy3DAsync(&params));

	// symbolize highpass bands
	timer("Symbolize HP");
	uint nextSymbolStream = 0;
	for(uint block = 1; block < 1 + blockCount; block++) {
		uint x = block % 2;
		uint y = block / 2 % 2;
		uint z = block / 4;
		uint offset = x * blockSize.x() + y * blockSize.y() * size.x() + z * blockSize.z() * size.x() * size.y();

		// make (unsigned!) symbols
		symbolize3D(dpSymbolStreams[nextSymbolStream++], dpDWT1 + offset, blockSize.x(), blockSize.y(), blockSize.z(), size.x(), size.x() * size.y());
	}

	BitStream bitStream(&highpassBitStream);

	// encode
	timer("Encode");
	bitStream.reserveBitSize(bitStream.getBitPosition() + sizeX * sizeY * sizeZ * 16);
	BitStream* pBitStream = &bitStream;
	bool result = encodeRLHuff(pInstance, &pBitStream, true, dpSymbolStreams.data(), blockCount, blockSize.volume());

	timer();

	pInstance->releaseBuffers(blockCount + 2);

	return result;
}

void decompressVolumeLosslessOneLevel(void* pCudaCompressInstance, short* dpImage, uint sizeX, uint sizeY, uint sizeZ, const short* dpLowpass, const std::vector<uint>& highpassBitStream, uint& bitStreamOffset)
{
	const uint blockCount = 7;

	Vec3i size(sizeX, sizeY, sizeZ);
	Vec3i blockSize = ((size + 1) / 2);


	Instance* pInstance = (Instance*)pCudaCompressInstance;

	short* dpDWT0 = pInstance->getBuffer<short>(size.volume());
	short* dpDWT1 = pInstance->getBuffer<short>(size.volume());
	std::vector<Symbol*> dpSymbolStreams(blockCount);
	for(uint i = 0; i < blockCount; i++) {
		dpSymbolStreams[i] = pInstance->getBuffer<Symbol>(blockSize.volume());
	}

	CudaScopedTimer timer(pInstance->CompressVolume.timerDecode);


	// copy lowpass band into larger buffer
	timer("Copy LP");
	cudaMemcpy3DParms params = { 0 };
	params.srcPtr.ptr = const_cast<short*>(dpLowpass);
	params.srcPtr.pitch = blockSize.x() * sizeof(short);
	params.srcPtr.xsize = blockSize.x() * sizeof(short);
	params.srcPtr.ysize = blockSize.y();
	params.dstPtr.ptr = dpDWT1;
	params.dstPtr.pitch = size.x() * sizeof(short);
	params.dstPtr.xsize = size.x() * sizeof(short);
	params.dstPtr.ysize = size.y();
	params.extent.width  = blockSize.x() * sizeof(short);
	params.extent.height = blockSize.y();
	params.extent.depth  = blockSize.z();
	params.kind = cudaMemcpyDeviceToDevice;
	cudaSafeCall(cudaMemcpy3DAsync(&params));

	BitStreamReadOnly bitStream(highpassBitStream.data(), uint(highpassBitStream.size() * sizeof(uint) * 8));
	bitStream.setBitPosition(bitStreamOffset * 32);

	// decode
	timer("Decode");
	BitStreamReadOnly* pBitStream = &bitStream;
	decodeRLHuff(pInstance, &pBitStream, true, dpSymbolStreams.data(), blockCount, blockSize.volume());

	bitStream.align<uint>();
	bitStreamOffset = bitStream.getBitPosition() / 32;

	// unsymbolize highpass bands
	timer("Unsymbolize HP");
	uint nextSymbolStream = 0;
	for(uint block = 1; block < 1 + blockCount; block++) {
		uint x = block % 2;
		uint y = block / 2 % 2;
		uint z = block / 4;
		uint offset = x * blockSize.x() + y * blockSize.y() * size.x() + z * blockSize.z() * size.x() * size.y();

		// get signed values back from unsigned symbols
		unsymbolize3D(dpDWT1 + offset, dpSymbolStreams[nextSymbolStream++], blockSize.x(), blockSize.y(), blockSize.z(), size.x(), size.x() * size.y());
	}

	// perform IDWT
	timer("IDWT");
	dwtIntInverse(dpImage, dpDWT0, dpDWT1, size.x(), size.y(), size.z());

	pInstance->releaseBuffers(blockCount + 2);
}

void decompressVolumeLosslessOneLevel(void* pCudaCompressInstance, short* dpImage, uint sizeX, uint sizeY, uint sizeZ, const short* dpLowpass, const std::vector<uint>& highpassBitStream)
{
	uint bitStreamOffset = 0;
	decompressVolumeLosslessOneLevel(pCudaCompressInstance, dpImage, sizeX, sizeY, sizeZ, dpLowpass, highpassBitStream, bitStreamOffset);
}

bool compressVolumeLossless(void* pCudaCompressInstance, const short* dpImage, uint sizeX, uint sizeY, uint sizeZ, uint numLevels, std::vector<uint>& bitStream)
{
	Instance* pInstance = (Instance*)pCudaCompressInstance;

	bitStream.clear(); //FIXME?

	uint elemCount = sizeX * sizeY * sizeZ;
	uint sizeBytes = elemCount * sizeof(short);

	short* dpBuffer0 = pInstance->getBuffer<short>(elemCount);
	short* dpBuffer1 = pInstance->getBuffer<short>(elemCount / 8);

	std::vector<uint>* levelBitStreams = new std::vector<uint>[numLevels];

	uint factor = 1;
	for(uint level = 0; level < numLevels; level++) {
		const short* dpSource = (level == 0 ? dpImage : dpBuffer0);
		if(!compressVolumeLosslessOneLevel(pInstance, dpSource, sizeX / factor, sizeY / factor, sizeZ / factor, dpBuffer1, levelBitStreams[level])) {
			return false;
		}
		factor *= 2;
		// copy lowpass to buffer0
		cudaMemcpy(dpBuffer0, dpBuffer1, sizeBytes / (factor*factor*factor), cudaMemcpyDeviceToDevice);
	}

	symbolize3D((ushort*)dpBuffer1, dpBuffer0, sizeX / factor, sizeY / factor, sizeZ / factor, sizeX / factor, sizeX * sizeY / (factor*factor));

	BitStream bitStreams(&bitStream);
	BitStream* pBitStreams = &bitStreams;
	bool result = encodeHuff(pInstance, &pBitStreams, true, (Symbol**)&dpBuffer1, 1, sizeX * sizeY * sizeZ / (factor * factor * factor));
	if(!result) {
		return false;
	}

	// append level bit streams
	for(int level = numLevels - 1; level >= 0; level--) {
		size_t oldSize = bitStream.size();
		bitStream.resize(bitStream.size() + levelBitStreams[level].size());
		memcpy(bitStream.data() + oldSize, levelBitStreams[level].data(), levelBitStreams[level].size() * sizeof(uint));
	}

	delete[] levelBitStreams;

	pInstance->releaseBuffers(2);

	return true;
}

void decompressVolumeLossless(void* pCudaCompressInstance, short* dpImage, uint sizeX, uint sizeY, uint sizeZ, uint numLevels, const std::vector<uint>& bitStream)
{
	Instance* pInstance = (Instance*)pCudaCompressInstance;

	uint elemCount = sizeX * sizeY * sizeZ;
	uint sizeBytes = elemCount * sizeof(short);

	short* dpBuffer0 = pInstance->getBuffer<short>(elemCount);
	short* dpBuffer1 = pInstance->getBuffer<short>(elemCount / 8);

	uint factor = (1 << numLevels);

	// huffman decode lowpass band
	BitStreamReadOnly bitStreams(bitStream.data(), uint(bitStream.size() * sizeof(uint) * 8));
	BitStreamReadOnly* pBitStreams = &bitStreams;
	decodeHuff(pInstance, &pBitStreams, true, (Symbol**)&dpBuffer1, 1, sizeX * sizeY * sizeZ / (factor * factor * factor));

	// convert (unsigned) symbols to signed values for IDWT
	unsymbolize3D(dpBuffer0, (ushort*)dpBuffer1, sizeX / factor, sizeY / factor, sizeZ / factor, sizeX / factor, sizeX * sizeY / (factor * factor));

	bitStreams.align<uint>();
	uint bitStreamOffset = bitStreams.getBitPosition() / 32;

	for(int level = numLevels - 1; level >= 0; level--) {
		// copy lowpass from previous level to buffer1
		cudaMemcpy(dpBuffer1, dpBuffer0, sizeBytes / (factor*factor*factor), cudaMemcpyDeviceToDevice);
		factor /= 2;
		// decode highpass bands and perform IDWT
		short* dpTarget = (level == 0 ? dpImage : dpBuffer0);
		decompressVolumeLosslessOneLevel(pCudaCompressInstance, dpTarget, sizeX / factor, sizeY / factor, sizeZ / factor, dpBuffer1, bitStream, bitStreamOffset);
	}

	pInstance->releaseBuffers(2);
}


bool compressVolume(void* pCudaCompressInstance, const byte* dpImage, uint sizeX, uint sizeY, uint sizeZ, uint numLevels, std::vector<uint>& bitStream, float quantizationStepLevel0)
{
	const uint blockCount = 7;

	Vec3i size(sizeX, sizeY, sizeZ);
	Vec3i blockSize = ((size + 1) / 2);


	Instance* pInstance = (Instance*)pCudaCompressInstance;

	float* dpDWT[2];
	dpDWT[0] = pInstance->getBuffer<float>(size.volume());
	dpDWT[1] = pInstance->getBuffer<float>(size.volume());
	std::vector<Symbol*> dpSymbolStreams(blockCount);
	for(uint i = 0; i < blockCount; i++) {
		dpSymbolStreams[i] = pInstance->getBuffer<Symbol>(blockSize.volume());
	}

	CudaScopedTimer timer(pInstance->CompressVolume.timerEncode);


	bitStream.clear(); //FIXME?

	// perform full dwt
	timer("DWT");
	int d = 0;
	dwtFloat3DForwardFromByte(dpDWT[d], dpDWT[1-d], dpDWT[d], dpImage, sizeX, sizeY, sizeZ);
	uint factor = 2;
	for(uint level = 1; level < numLevels; level++) {
		dwtFloat3DForward(dpDWT[1-d], dpDWT[d], dpDWT[1-d], dpDWT[d], sizeX / factor, sizeY / factor, sizeZ / factor, 1, sizeX, sizeX * sizeY, sizeX, sizeX * sizeY);
		d = 1 - d;
		factor *= 2;
	}

	BitStream bitStreams(&bitStream);
	BitStream* pBitStreams = &bitStreams;

	// quantize and encode lowpass band
	timer("Quantize LP");
	float quantizationStep = quantizationStepLevel0 / float(factor/2);
	quantizeToSymbols3D(dpSymbolStreams[0], dpDWT[d], quantizationStep, sizeX / factor, sizeY / factor, sizeZ / factor, sizeX, sizeX * sizeY);
	timer("Encode LP");
	bool result = encodeHuff(pInstance, &pBitStreams, true, dpSymbolStreams.data(), 1, size.volume() / (factor * factor * factor));
	if(!result) {
		pInstance->releaseBuffers(blockCount + 1);
		return false;
	}

	// highpass bands
	for(int level = numLevels - 1; level >= 0; level--) {
		quantizationStep = quantizationStepLevel0 / float(factor/2);

		Vec3i blockSize = size / factor;

		// symbolize highpass bands
		timer("Symbolize HP");
		uint nextSymbolStream = 0;
		for(uint block = 1; block < 1 + blockCount; block++) {
			uint x = block % 2;
			uint y = block / 2 % 2;
			uint z = block / 4;
			uint offset = x * blockSize.x() + y * blockSize.y() * size.x() + z * blockSize.z() * size.x() * size.y();

			// make (unsigned!) symbols
			quantizeToSymbols3D(dpSymbolStreams[nextSymbolStream++], dpDWT[d] + offset, quantizationStep, blockSize.x(), blockSize.y(), blockSize.z(), size.x(), size.x() * size.y());
		}

		// encode
		timer("Encode HP");
		bool result = encodeRLHuff(pInstance, &pBitStreams, true, dpSymbolStreams.data(), blockCount, blockSize.volume());
		if(!result) {
			pInstance->releaseBuffers(blockCount + 1);
			return false;
		}

		factor /= 2;
	}

	timer();

	pInstance->releaseBuffers(blockCount + 2);

	return true;
}

void decompressVolume(void* pCudaCompressInstance, byte* dpImage, uint sizeX, uint sizeY, uint sizeZ, uint numLevels, const std::vector<uint>& bitStream, float quantizationStepLevel0)
{
	decompressVolume(pCudaCompressInstance, dpImage, sizeX, sizeY, sizeZ, numLevels, bitStream.data(), uint(bitStream.size() * sizeof(uint) * 8), quantizationStepLevel0);
}

void decompressVolume(void* pCudaCompressInstance, byte* dpImage, uint sizeX, uint sizeY, uint sizeZ, uint numLevels, const uint* pBits, uint bitCount, float quantizationStepLevel0)
{
	const uint blockCount = 7;

	Vec3i size(sizeX, sizeY, sizeZ);
	Vec3i blockSize = ((size + 1) / 2);


	Instance* pInstance = (Instance*)pCudaCompressInstance;

	float* dpDWT[2];
	dpDWT[0] = pInstance->getBuffer<float>(size.volume());
	dpDWT[1] = pInstance->getBuffer<float>(size.volume());
	std::vector<Symbol*> dpSymbolStreams(blockCount);
	for(uint i = 0; i < blockCount; i++) {
		dpSymbolStreams[i] = pInstance->getBuffer<Symbol>(blockSize.volume());
	}

	CudaScopedTimer timer(pInstance->CompressVolume.timerDecode);


	BitStreamReadOnly bitStreams(pBits, bitCount);
	BitStreamReadOnly* pBitStreams = &bitStreams;

	int factor = (1 << numLevels);

	int d = 0;

	// lowpass band
	timer("Decode LP");
	decodeHuff(pInstance, &pBitStreams, true, dpSymbolStreams.data(), 1, size.volume() / (factor * factor * factor));

	timer("Unquantize LP");
	float quantizationStep = quantizationStepLevel0 / float(factor/2);
	unquantizeFromSymbols3D(dpDWT[d], dpSymbolStreams[0], quantizationStep, sizeX / factor, sizeY / factor, sizeZ / factor, sizeX, sizeX * sizeY);

	// highpass bands
	for(int level = numLevels - 1; level >= 0; level--) {
		float quantizationStep = quantizationStepLevel0 / float(factor/2);

		Vec3i blockSize = size / factor;

		// decode
		timer("Decode HP");
		decodeRLHuff(pInstance, &pBitStreams, true, dpSymbolStreams.data(), blockCount, blockSize.volume());

		// unsymbolize
		timer("Unsymbolize HP");
		uint nextSymbolStream = 0;
		for(uint block = 1; block < 1 + blockCount; block++) {
			uint x = block % 2;
			uint y = block / 2 % 2;
			uint z = block / 4;
			uint offset = x * blockSize.x() + y * blockSize.y() * size.x() + z * blockSize.z() * size.x() * size.y();

			// get signed values back from unsigned symbols
			unquantizeFromSymbols3D(dpDWT[d] + offset, dpSymbolStreams[nextSymbolStream++], quantizationStep, blockSize.x(), blockSize.y(), blockSize.z(), size.x(), size.x() * size.y());
		}

		factor /= 2;
	}

	// perform full idwt
	timer("IDWT");
	factor = (1 << numLevels);
	for(int level = numLevels - 1; level >= 1; level--) {
		factor /= 2;
		dwtFloat3DInverse(dpDWT[1-d], dpDWT[d], dpDWT[1-d], dpDWT[d], sizeX / factor, sizeY / factor, sizeZ / factor, 1, sizeX, sizeX * sizeY, sizeX, sizeX * sizeY);
		d = 1 - d;
	}

	dwtFloat3DInverseToByte(dpImage, dpDWT[d], dpDWT[1-d], dpDWT[d], sizeX, sizeY, sizeZ);

	timer();

	pInstance->releaseBuffers(blockCount + 2);
}


bool compressVolumeFloat(void* pCudaCompressInstance, const float* dpImage, uint sizeX, uint sizeY, uint sizeZ, uint numLevels, std::vector<uint>& bitStream, float quantizationStepLevel0)
{
	const uint blockCount = 7;

	Vec3i size(sizeX, sizeY, sizeZ);
	Vec3i blockSize = ((size + 1) / 2);


	Instance* pInstance = (Instance*)pCudaCompressInstance;


	float* dpDWT[2];
	dpDWT[0] = pInstance->getBuffer<float>(size.volume());
	dpDWT[1] = pInstance->getBuffer<float>(size.volume());
	std::vector<Symbol*> dpSymbolStreams(blockCount);
	for(uint i = 0; i < blockCount; i++) {
		dpSymbolStreams[i] = pInstance->getBuffer<Symbol>(blockSize.volume());
	}

	CudaScopedTimer timer(pInstance->CompressVolume.timerEncode);


	bitStream.clear(); //FIXME?

	// perform full dwt
	timer("DWT");
	int d = 0;
	dwtFloat3DForward(dpDWT[d], dpDWT[1-d], dpDWT[d], dpImage, sizeX, sizeY, sizeZ);
	uint factor = 2;
	for(uint level = 1; level < numLevels; level++) {
		dwtFloat3DForward(dpDWT[1-d], dpDWT[d], dpDWT[1-d], dpDWT[d], sizeX / factor, sizeY / factor, sizeZ / factor, 1, sizeX, sizeX * sizeY, sizeX, sizeX * sizeY);
		d = 1 - d;
		factor *= 2;
	}
	// note: the correct coefficients are now distributed over both DWT buffers:
	//       the coarsest level is in d, the next finest in 1-d, then d again etc.

	BitStream bitStreams(&bitStream);
	BitStream* pBitStreams = &bitStreams;

	// quantize and encode lowpass band
	timer("Quantize LP");
	float quantizationStep = quantizationStepLevel0 / float(factor/2);
	quantizeToSymbols3D(dpSymbolStreams[0], dpDWT[d], quantizationStep, sizeX / factor, sizeY / factor, sizeZ / factor, sizeX, sizeX * sizeY);
	timer("Encode LP");
	bool result = encodeHuff(pInstance, &pBitStreams, true, dpSymbolStreams.data(), 1, size.volume() / (factor * factor * factor));
	if(!result) {
		pInstance->releaseBuffers(blockCount + 1);
		return false;
	}
	//encodePrintBitCounts();
	//encodeResetBitCounts();

	// highpass bands
	for(int level = numLevels - 1; level >= 0; level--) {
		quantizationStep = quantizationStepLevel0 / float(factor/2);

		Vec3i blockSize = size / factor;

		// symbolize highpass bands
		timer("Symbolize HP");
		uint nextSymbolStream = 0;
		for(uint block = 1; block < 1 + blockCount; block++) {
			uint x = block % 2;
			uint y = block / 2 % 2;
			uint z = block / 4;
			uint offset = x * blockSize.x() + y * blockSize.y() * size.x() + z * blockSize.z() * size.x() * size.y();

			// make (unsigned!) symbols
			quantizeToSymbols3D(dpSymbolStreams[nextSymbolStream++], dpDWT[d] + offset, quantizationStep, blockSize.x(), blockSize.y(), blockSize.z(), size.x(), size.x() * size.y());
		}
		// switch DWT buffer for next level
		d = 1 - d;

		// encode
		timer("Encode HP");
		bool result = encodeRLHuff(pInstance, &pBitStreams, true, dpSymbolStreams.data(), blockCount, blockSize.volume());
		if(!result) {
			pInstance->releaseBuffers(blockCount + 1);
			return false;
		}
		//encodePrintBitCounts();
		//encodeResetBitCounts();

		factor /= 2;
	}

	timer();

	pInstance->releaseBuffers(blockCount + 2);

	return true;
}

void decompressVolumeFloat(void* pCudaCompressInstance, float* dpImage, uint sizeX, uint sizeY, uint sizeZ, uint numLevels, const std::vector<uint>& bitStream, float quantizationStepLevel0)
{
	decompressVolumeFloat(pCudaCompressInstance, dpImage, sizeX, sizeY, sizeZ, numLevels, bitStream.data(), uint(bitStream.size() * sizeof(uint) * 8), quantizationStepLevel0);
}

void decompressVolumeFloat(void* pCudaCompressInstance, float* dpImage, uint sizeX, uint sizeY, uint sizeZ, uint numLevels, const uint* pBits, uint bitCount, float quantizationStepLevel0)
{
	const uint blockCount = 7;

	Vec3i size(sizeX, sizeY, sizeZ);
	Vec3i blockSize = ((size + 1) / 2);


	Instance* pInstance = (Instance*)pCudaCompressInstance;

	float* dpDWT[2];
	dpDWT[0] = pInstance->getBuffer<float>(size.volume());
	dpDWT[1] = pInstance->getBuffer<float>(size.volume());
	std::vector<Symbol*> dpSymbolStreams(blockCount);
	for(uint i = 0; i < blockCount; i++) {
		dpSymbolStreams[i] = pInstance->getBuffer<Symbol>(blockSize.volume());
	}

	Symbol** dppSymbolStreams = pInstance->getBuffer<Symbol*>(blockCount);
	cudaSafeCall(cudaEventSynchronize(pInstance->CompressVolume.syncEventUpload));
	memcpy(pInstance->CompressVolume.pUpload, dpSymbolStreams.data(), blockCount * sizeof(Symbol*));
	cudaSafeCall(cudaMemcpyAsync(dppSymbolStreams, pInstance->CompressVolume.pUpload, blockCount * sizeof(Symbol*), cudaMemcpyHostToDevice));
	cudaSafeCall(cudaEventRecord(pInstance->CompressVolume.syncEventUpload));

	CudaScopedTimer timer(pInstance->CompressVolume.timerDecode);


	BitStreamReadOnly bitStreams(pBits, bitCount);
	BitStreamReadOnly* pBitStreams = &bitStreams;

	int factor = (1 << numLevels);

	// lowpass band
	timer("Decode LP");
	decodeHuff(pInstance, &pBitStreams, true, dpSymbolStreams.data(), 1, size.volume() / (factor * factor * factor));

	timer("Unquantize LP");
	float quantizationStep = quantizationStepLevel0 / float(factor/2);
	unquantizeFromSymbols3D(dpImage, dpSymbolStreams[0], quantizationStep, sizeX / factor, sizeY / factor, sizeZ / factor, sizeX, sizeX * sizeY);

	// highpass bands
	for(int level = numLevels - 1; level >= 0; level--) {
		float quantizationStep = quantizationStepLevel0 / float(factor/2);

		Vec3i blockSize = size / factor;

		// decode
		timer("Decode HP");
		decodeRLHuff(pInstance, &pBitStreams, true, dpSymbolStreams.data(), blockCount, blockSize.volume());

		factor /= 2;

		// unsymbolize and idwt
		timer("IDWT");
		dwtFloat3DInverseFromSymbols(dpImage, dpDWT[1], dpDWT[0], dpImage, dppSymbolStreams, quantizationStep, size.x() / factor, size.y() / factor, size.z() / factor, 1, size.x(), size.x() * size.y(), size.x(), size.x() * size.y());
	}

	timer();

	pInstance->releaseBuffers(blockCount + 3);
}


void getVolumeFloatMaxAbs(void* pCudaCompressInstance, const float* dpImage, uint sizeX, uint sizeY, uint sizeZ, float* dpValMax)
{
	Instance* pInstance = (Instance*)pCudaCompressInstance;

	if(dpValMax != 0) {
		reduceMaxAbsFloat(dpValMax, dpImage, sizeX * sizeY * sizeZ, pInstance->m_pReducePlan);
		cudaCheckMsg("getVolumeFloatMaxAbs: Error in reduceMaxAbsFloat");
	}
}

}
