#include "Encode.h"

#include <cassert>

#include <cuda_runtime.h>

#include "tools/Vec.h"

#include "cudaUtil.h"
#include "util.h"

#include "Instance.h"
#include "Histogram.h"
#include "Huffman.h"
#include "HuffmanTable.h"
#include "PackInc.h"
#include "RunLength.h"

using namespace tum3D;


namespace cudaCompress {

//static uint g_bitsSymbolEncodeTables = 0;
//static uint g_bitsSymbolCodewords = 0;
//static uint g_bitsSymbolOffsets = 0;
//static uint g_bitsZeroCountEncodeTables = 0;
//static uint g_bitsZeroCountCodewords = 0;
//static uint g_bitsZeroCountOffsets = 0;
//static uint g_bitsTotal = 0;
//static uint g_totalEncodedCount = 0;


static uint getNumUintsForBits(uint bitsize)
{
	uint bitsPerUint = sizeof(uint) * 8;
	return (bitsize + bitsPerUint - 1) / bitsPerUint;
}

static uint getNumUintsForBytes(uint bytesize)
{
	uint bytesPerUint = sizeof(uint);
	return (bytesize + bytesPerUint - 1) / bytesPerUint;
}

static uint getNumOffsets(uint symbolCount, uint offsetInterval)
{
	return (symbolCount + offsetInterval - 1) / offsetInterval;
}


size_t encodeGetRequiredMemory(const Instance* pInstance)
{
	bool is3D = (pInstance->m_sizeMax.z() > 1);
	uint blockCountMax = (is3D ? 7 : 4);
	blockCountMax *= max(1u, pInstance->m_channelCountMax);

	Vec3i blockSizeMax = (pInstance->m_sizeMax + 1) / 2;
	uint symbolCountPerBlockMax = blockSizeMax.volume();

	uint symbolStreamMaxBytes = symbolCountPerBlockMax * sizeof(Symbol);
	uint offsetStreamMaxBytes = getNumOffsets(symbolCountPerBlockMax, pInstance->m_offsetInterval) * sizeof(uint);

	size_t size = 0;

	for(uint block = 0; block < blockCountMax; block++) {
		// pStreamInfos.dpEncodeCodeWords
		size += getAlignedSize(MAX_DISTINCT_SYMBOLS * sizeof(uint), 128);
		// pStreamInfos.dpEncodeCodeWordLengths
		size += getAlignedSize(MAX_DISTINCT_SYMBOLS * sizeof(uint), 128);

		// pStreamInfos.dpDecodeTable
		size += getAlignedSize(HuffmanDecodeTable::computeMaxGPUSize(), 128);

		// pStreamInfos.dpCodewordStream
		size += getAlignedSize(symbolStreamMaxBytes, 128);
		// pStreamInfos.dpOffsets
		size += getAlignedSize(offsetStreamMaxBytes, 128);

		// pdpSymbolStreamsCompacted
		size += getAlignedSize(symbolStreamMaxBytes, 128);
		// pdpZeroCounts
		size += getAlignedSize(symbolStreamMaxBytes, 128);
	}

	return size;
}

bool encodeInit(Instance* pInstance)
{
	bool is3D = (pInstance->m_sizeMax.z() > 1);
	uint blockCountMax = (is3D ? 7 : 4);
	blockCountMax *= max(1u, pInstance->m_channelCountMax);

	Vec3i blockSizeMax = (pInstance->m_sizeMax + 1) / 2;
	uint symbolCountPerBlockMax = blockSizeMax.volume();


	uint symbolStreamMaxBytes = symbolCountPerBlockMax * sizeof(Symbol);
	uint offsetStreamMaxBytes = getNumOffsets(symbolCountPerBlockMax, pInstance->m_offsetInterval) * sizeof(uint);

	pInstance->Encode.symbolEncodeTables.resize(blockCountMax);
	pInstance->Encode.zeroCountEncodeTables.resize(blockCountMax);

	pInstance->Encode.symbolDecodeTables.resize(blockCountMax);
	pInstance->Encode.zeroCountDecodeTables.resize(blockCountMax);

	pInstance->Encode.ppCodewordBuffers.resize(2 * blockCountMax);
	pInstance->Encode.ppOffsetBuffers.resize(2 * blockCountMax);

	pInstance->Encode.pSyncEvents.resize(2 * blockCountMax);

	for(uint block = 0; block < 2 * blockCountMax; block++) {
		cudaSafeCall(cudaMallocHost(&pInstance->Encode.ppCodewordBuffers[block], symbolStreamMaxBytes));
		cudaSafeCall(cudaMallocHost(&pInstance->Encode.ppOffsetBuffers  [block], offsetStreamMaxBytes));

		cudaSafeCall(cudaEventCreate(&pInstance->Encode.pSyncEvents[block], cudaEventDisableTiming));
	}

	cudaSafeCall(cudaMallocHost(&pInstance->Encode.pUpload0, blockCountMax * sizeof(HuffmanGPUStreamInfo)));
	cudaSafeCall(cudaMallocHost(&pInstance->Encode.pUpload1, blockCountMax * sizeof(HuffmanGPUStreamInfo)));

	cudaSafeCall(cudaEventCreate(&pInstance->Encode.encodeFinishedEvent, cudaEventDisableTiming));
	cudaSafeCall(cudaEventRecord(pInstance->Encode.encodeFinishedEvent));

	return true;
}

bool encodeShutdown(Instance* pInstance)
{
	cudaSafeCall(cudaEventDestroy(pInstance->Encode.encodeFinishedEvent));
	pInstance->Encode.encodeFinishedEvent = 0;

	cudaSafeCall(cudaFreeHost(pInstance->Encode.pUpload1));
	pInstance->Encode.pUpload1 = nullptr;

	cudaSafeCall(cudaFreeHost(pInstance->Encode.pUpload0));
	pInstance->Encode.pUpload0 = nullptr;

	bool is3D = (pInstance->m_sizeMax.z() > 1);
	uint blockCountMax = (is3D ? 7 : 4);
	blockCountMax *= max(1u, pInstance->m_channelCountMax);

	for(uint block = 0; block < 2 * blockCountMax; block++) {
		cudaSafeCall(cudaEventDestroy(pInstance->Encode.pSyncEvents[block]));

		cudaSafeCall(cudaFreeHost(pInstance->Encode.ppOffsetBuffers  [block]));
		cudaSafeCall(cudaFreeHost(pInstance->Encode.ppCodewordBuffers[block]));
	}
	pInstance->Encode.pSyncEvents.clear();
	pInstance->Encode.zeroCountDecodeTables.clear();
	pInstance->Encode.symbolDecodeTables.clear();
	pInstance->Encode.zeroCountEncodeTables.clear();
	pInstance->Encode.symbolEncodeTables.clear();
	pInstance->Encode.ppOffsetBuffers.clear();
	pInstance->Encode.ppCodewordBuffers.clear();

	return true;
}

bool encodeRLHuff(Instance* pInstance, BitStream* ppBitStreams[], bool singleBitStream, const Symbol* const pdpSymbolStreams[], uint blockCount, uint symbolCountPerBlock)
{
	uint symbolStreamMaxBytes = symbolCountPerBlock * sizeof(Symbol);
	uint offsetStreamMaxBytes = getNumOffsets(symbolCountPerBlock, pInstance->m_offsetInterval) * sizeof(uint);

	std::vector<Symbol*> pdpSymbolStreamsCompacted(blockCount);
	std::vector<ushort*> pdpZeroCounts(blockCount);
	std::vector<HuffmanGPUStreamInfo> pStreamInfos(blockCount);
	for(uint block = 0; block < blockCount; block++) {
		pdpSymbolStreamsCompacted[block] = pInstance->getBuffer<Symbol>(symbolCountPerBlock);
		pdpZeroCounts[block]             = pInstance->getBuffer<Symbol>(symbolCountPerBlock);

		HuffmanGPUStreamInfo& streamInfo = pStreamInfos[block];

		streamInfo.dpEncodeCodeWords       = pInstance->getBuffer<uint>(MAX_DISTINCT_SYMBOLS);
		streamInfo.dpEncodeCodeWordLengths = pInstance->getBuffer<uint>(MAX_DISTINCT_SYMBOLS);

		streamInfo.dpDecodeTable = pInstance->getBuffer<byte>(HuffmanDecodeTable::computeMaxGPUSize());

		streamInfo.dpCodewordStream = (uint*)pInstance->getBuffer<byte>(symbolStreamMaxBytes);
		streamInfo.dpOffsets        = (uint*)pInstance->getBuffer<byte>(offsetStreamMaxBytes);
	}

	cudaSafeCall(cudaEventSynchronize(pInstance->Encode.encodeFinishedEvent));

	CudaScopedTimer timer(pInstance->Encode.timerEncode);

	timer("Run Length Encode");

	// run length encode
	std::vector<uint> symbolCountsPerBlock(blockCount, symbolCountPerBlock);
	std::vector<uint> symbolCountsCompact(blockCount);
	runLengthEncode(pInstance, pdpSymbolStreamsCompacted.data(), pdpZeroCounts.data(), (const Symbol**)pdpSymbolStreams, symbolCountsPerBlock.data(), blockCount, ZERO_COUNT_MAX, symbolCountsCompact.data());

	timer("Symbols:    Design Huffman Tables");

	for(uint block = 0; block < blockCount; block++) {
		// padding for histogram (which wants the element count to be a multiple of 8)
		histogramPadData(pdpSymbolStreamsCompacted[block], symbolCountsCompact[block]);
		histogramPadData(pdpZeroCounts[block],             symbolCountsCompact[block]);
	}

	// 1. compacted symbols
	// build encode tables
	std::vector<HuffmanEncodeTable>& symbolEncodeTables = pInstance->Encode.symbolEncodeTables;
	if(!HuffmanEncodeTable::design(pInstance, symbolEncodeTables.data(), blockCount, (const Symbol**)pdpSymbolStreamsCompacted.data(), symbolCountsCompact.data())) {
		pInstance->releaseBuffers(7 * blockCount);
		return false;
	}

	timer("Symbols:    Upload Huffman Tables");

	// fill stream infos, upload encode tables
	for(uint block = 0; block < blockCount; block++) {
		HuffmanGPUStreamInfo& streamInfo = pStreamInfos[block];

		streamInfo.dpSymbolStream = pdpSymbolStreamsCompacted[block];
		streamInfo.symbolCount = symbolCountsCompact[block];

		symbolEncodeTables[block].uploadToGPUAsync(streamInfo.dpEncodeCodeWords, streamInfo.dpEncodeCodeWordLengths);
	}

	timer("Symbols:    Huffman Encode");

	// encode the symbols
	std::vector<uint> codewordBitsizeSymbols(blockCount);
	huffmanEncode(pInstance, pStreamInfos.data(), blockCount, pInstance->m_offsetInterval, codewordBitsizeSymbols.data());

	timer("Symbols:    Pack Offsets");

	// make offsets incremental and pack
	for(uint block = 0; block < blockCount; block++) {
		HuffmanGPUStreamInfo& streamInfo = pStreamInfos[block];

		packInc16(pInstance, streamInfo.dpOffsets, (ushort*)streamInfo.dpOffsets, getNumOffsets(symbolCountsCompact[block], pInstance->m_offsetInterval));
	}

	timer("Symbols:    Download");

	// download encoded symbols and offsets
	// (GPU buffers will be used again for the zero counts)
	for(uint block = 0; block < blockCount; block++) {
		HuffmanGPUStreamInfo& streamInfo = pStreamInfos[block];

		uint codewordUints = getNumUintsForBits(codewordBitsizeSymbols[block]);
		cudaSafeCall(cudaMemcpyAsync(pInstance->Encode.ppCodewordBuffers[block], streamInfo.dpCodewordStream, codewordUints * sizeof(uint), cudaMemcpyDeviceToHost));
		uint numOffsets = getNumOffsets(symbolCountsCompact[block], pInstance->m_offsetInterval);
		cudaSafeCall(cudaMemcpyAsync(pInstance->Encode.ppOffsetBuffers[block], streamInfo.dpOffsets, numOffsets * sizeof(ushort), cudaMemcpyDeviceToHost));
		cudaSafeCall(cudaEventRecord(pInstance->Encode.pSyncEvents[block]));
	}


	timer("ZeroCounts: Design Huffman Tables");

	// 2. zero counts
	assert(sizeof(Symbol) == sizeof(ushort));
	// build encode tables
	std::vector<HuffmanEncodeTable>& zeroCountEncodeTables = pInstance->Encode.zeroCountEncodeTables;
	if(!HuffmanEncodeTable::design(pInstance, zeroCountEncodeTables.data(), blockCount, (const Symbol**)pdpZeroCounts.data(), symbolCountsCompact.data())) {
		pInstance->releaseBuffers(7 * blockCount);
		return false;
	}

	timer("ZeroCounts: Upload Huffman Tables");

	// fill stream infos, upload encode tables
	for(uint block = 0; block < blockCount; block++) {
		HuffmanGPUStreamInfo& streamInfo = pStreamInfos[block];

		streamInfo.dpSymbolStream = pdpZeroCounts[block];
		streamInfo.symbolCount = symbolCountsCompact[block];

		zeroCountEncodeTables[block].uploadToGPUAsync(streamInfo.dpEncodeCodeWords, streamInfo.dpEncodeCodeWordLengths);
	}

	timer("ZeroCounts: Huffman Encode");

	// encode the zero counts
	std::vector<uint> codewordBitsizeZeroCounts(blockCount);
	huffmanEncode(pInstance, pStreamInfos.data(), blockCount, pInstance->m_offsetInterval, codewordBitsizeZeroCounts.data());

	timer("ZeroCounts: Pack Offsets");

	// make offsets incremental and pack
	for(uint block = 0; block < blockCount; block++) {
		HuffmanGPUStreamInfo& streamInfo = pStreamInfos[block];
		packInc16(pInstance, streamInfo.dpOffsets, (ushort*)streamInfo.dpOffsets, getNumOffsets(symbolCountsCompact[block], pInstance->m_offsetInterval));
	}

	timer("ZeroCounts: Download");

	// download zero count codeword stream and offsets
	for(uint block = 0; block < blockCount; block++) {
		HuffmanGPUStreamInfo& streamInfo = pStreamInfos[block];

		uint codewordUints = getNumUintsForBits(codewordBitsizeZeroCounts[block]);
		cudaSafeCall(cudaMemcpyAsync(pInstance->Encode.ppCodewordBuffers[blockCount + block], streamInfo.dpCodewordStream, codewordUints * sizeof(uint), cudaMemcpyDeviceToHost));
		uint numOffsets = getNumOffsets(symbolCountsCompact[block], pInstance->m_offsetInterval);
		cudaSafeCall(cudaMemcpyAsync(pInstance->Encode.ppOffsetBuffers[blockCount + block], streamInfo.dpOffsets, numOffsets * sizeof(ushort), cudaMemcpyDeviceToHost));
		cudaSafeCall(cudaEventRecord(pInstance->Encode.pSyncEvents[blockCount + block]));
	}

	timer();


	// write to bitstream
	for(uint block = 0; block < blockCount; block++) {
		BitStream& bitStream = *ppBitStreams[singleBitStream ? 0 : block];
		uint bitStreamPosStart = bitStream.getBitPosition();

		uint bitStreamPosOld = bitStreamPosStart;
		uint bitStreamPos;

		// write compacted symbol count
		bitStream.writeAligned(&symbolCountsCompact[block], 1);

		// 1. compacted symbols
		// write encode table
		symbolEncodeTables[block].writeToBitstream(bitStream);

		bitStreamPos = bitStream.getBitPosition();
		//g_bitsSymbolEncodeTables += bitStreamPos - bitStreamPosOld;
		bitStreamPosOld = bitStreamPos;

		// write symbol codeword stream
		bitStream.writeAligned(&codewordBitsizeSymbols[block], 1);
		uint codewordUints = getNumUintsForBits(codewordBitsizeSymbols[block]);
		cudaSafeCall(cudaEventSynchronize(pInstance->Encode.pSyncEvents[block]));
		bitStream.writeAligned(pInstance->Encode.ppCodewordBuffers[block], codewordUints);

		bitStreamPos = bitStream.getBitPosition();
		//g_bitsSymbolCodewords += bitStreamPos - bitStreamPosOld;
		bitStreamPosOld = bitStreamPos;

		// write symbol offsets
		uint numOffsets = getNumOffsets(symbolCountsCompact[block], pInstance->m_offsetInterval);
		bitStream.writeAligned((ushort*)pInstance->Encode.ppOffsetBuffers[block], numOffsets);

		bitStreamPos = bitStream.getBitPosition();
		//g_bitsSymbolOffsets += bitStreamPos - bitStreamPosOld;
		bitStreamPosOld = bitStreamPos;


		// 2. zero counts
		// write encode table
		zeroCountEncodeTables[block].writeToBitstream(bitStream);

		bitStreamPos = bitStream.getBitPosition();
		//g_bitsZeroCountEncodeTables += bitStreamPos - bitStreamPosOld;
		bitStreamPosOld = bitStreamPos;

		// write zero count codeword stream
		bitStream.writeAligned(&codewordBitsizeZeroCounts[block], 1);
		codewordUints = getNumUintsForBits(codewordBitsizeZeroCounts[block]);
		cudaSafeCall(cudaEventSynchronize(pInstance->Encode.pSyncEvents[blockCount + block]));
		bitStream.writeAligned(pInstance->Encode.ppCodewordBuffers[blockCount + block], codewordUints);

		bitStreamPos = bitStream.getBitPosition();
		//g_bitsZeroCountCodewords += bitStreamPos - bitStreamPosOld;
		bitStreamPosOld = bitStreamPos;

		// write zero count offsets
		bitStream.writeAligned((ushort*)pInstance->Encode.ppOffsetBuffers[blockCount + block], numOffsets);

		bitStreamPos = bitStream.getBitPosition();
		//g_bitsZeroCountOffsets += bitStreamPos - bitStreamPosOld;
		bitStreamPosOld = bitStreamPos;

		//g_bitsTotal += bitStreamPos - bitStreamPosStart;
	}
	//g_totalEncodedCount += blockCount * symbolCountPerBlock;

	cudaSafeCall(cudaEventRecord(pInstance->Encode.encodeFinishedEvent));

	pInstance->releaseBuffers(7 * blockCount);

	return true;
}

bool decodeRLHuff(Instance* pInstance, BitStreamReadOnly* ppBitStreams[], bool singleBitStream, Symbol* pdpSymbolStreams[], uint blockCount, uint symbolCountPerBlock)
{
	uint symbolStreamMaxBytes = symbolCountPerBlock * sizeof(Symbol);
	uint offsetStreamMaxBytes = getNumOffsets(symbolCountPerBlock, pInstance->m_offsetInterval) * sizeof(uint);

	std::vector<Symbol*> pdpSymbolStreamsCompacted(blockCount);
	std::vector<ushort*> pdpZeroCounts(blockCount);
	std::vector<HuffmanGPUStreamInfo> pStreamInfos(blockCount);
	for(uint block = 0; block < blockCount; block++) {
		pdpSymbolStreamsCompacted[block] = pInstance->getBuffer<Symbol>(symbolCountPerBlock);
		pdpZeroCounts[block]             = pInstance->getBuffer<Symbol>(symbolCountPerBlock);

		HuffmanGPUStreamInfo& streamInfo = pStreamInfos[block];

		streamInfo.dpEncodeCodeWords       = pInstance->getBuffer<uint>(MAX_DISTINCT_SYMBOLS);
		streamInfo.dpEncodeCodeWordLengths = pInstance->getBuffer<uint>(MAX_DISTINCT_SYMBOLS);

		streamInfo.dpDecodeTable = pInstance->getBuffer<byte>(HuffmanDecodeTable::computeMaxGPUSize());

		streamInfo.dpCodewordStream = (uint*)pInstance->getBuffer<byte>(symbolStreamMaxBytes);
		streamInfo.dpOffsets        = (uint*)pInstance->getBuffer<byte>(offsetStreamMaxBytes);
	}

	cudaSafeCall(cudaEventSynchronize(pInstance->Encode.encodeFinishedEvent));

	CudaScopedTimer timer(pInstance->Encode.timerDecode);

	timer("Symbols:    Upload (+read BitStream)");

	// read and upload decode tables, upload codeword streams and offsets, and fill stream info for symbols
	std::vector<uint> symbolCountsCompact(blockCount);
	std::vector<HuffmanDecodeTable>& symbolDecodeTables = pInstance->Encode.symbolDecodeTables;
	std::vector<HuffmanDecodeTable>& zeroCountDecodeTables = pInstance->Encode.zeroCountDecodeTables;
	std::vector<const uint*> pCodewordStreams(blockCount);
	std::vector<uint> codewordUints(blockCount);
	std::vector<const uint*> pOffsets(blockCount);
	for(uint block = 0; block < blockCount; block++) {
		BitStreamReadOnly& bitStream = *ppBitStreams[singleBitStream ? 0 : block];

		// read compacted symbol count
		bitStream.readAligned(&symbolCountsCompact[block], 1);

		// 1. compacted symbols
		// read symbol decode table
		symbolDecodeTables[block].readFromBitstream(bitStream);

		HuffmanGPUStreamInfo& streamInfo = pStreamInfos[block];

		// upload symbol decode table
		symbolDecodeTables[block].uploadToGPUAsync(streamInfo.dpDecodeTable);
		streamInfo.decodeSymbolTableSize = symbolDecodeTables[block].m_symbolTableSize;

		// upload symbol codewords
		uint codewordBitsize;
		bitStream.readAligned(&codewordBitsize, 1);
		codewordUints[block] = getNumUintsForBits(codewordBitsize);
		pCodewordStreams[block] = bitStream.getRaw() + bitStream.getBitPosition() / (sizeof(uint)*8);
		// assume the bitstream memory is already page-locked
		cudaSafeCall(cudaMemcpyAsync(streamInfo.dpCodewordStream, pCodewordStreams[block], codewordUints[block] * sizeof(uint), cudaMemcpyHostToDevice));
		bitStream.skipBits(codewordUints[block] * sizeof(uint) * 8);

		// upload symbol offsets
		uint numOffsets = getNumOffsets(symbolCountsCompact[block], pInstance->m_offsetInterval);
		bitStream.align<uint>();
		pOffsets[block] = bitStream.getRaw() + bitStream.getBitPosition() / (sizeof(uint)*8);
		// assume the bitstream memory is already page-locked
		cudaSafeCall(cudaMemcpyAsync(streamInfo.dpOffsets, pOffsets[block], numOffsets * sizeof(ushort), cudaMemcpyHostToDevice));
		bitStream.skipAligned<ushort>(numOffsets);

		// unpack incremental offsets
		unpackInc16(pInstance, streamInfo.dpOffsets, (ushort*)streamInfo.dpOffsets, numOffsets);

		// fill stream infos
		streamInfo.dpSymbolStream = pdpSymbolStreamsCompacted[block];
		streamInfo.symbolCount = symbolCountsCompact[block];

		// 2. zero counts
		// read zero count decode table
		zeroCountDecodeTables[block].readFromBitstream(bitStream);

		// read zero count codewords pointer
		bitStream.readAligned(&codewordBitsize, 1);
		codewordUints[block] = getNumUintsForBits(codewordBitsize);
		pCodewordStreams[block] = bitStream.getRaw() + bitStream.getBitPosition() / (sizeof(uint)*8);
		bitStream.skipBits(codewordUints[block] * sizeof(uint) * 8);

		// read zero count offsets pointer
		bitStream.align<uint>();
		pOffsets[block] = bitStream.getRaw() + bitStream.getBitPosition() / (sizeof(uint)*8);
		bitStream.skipAligned<ushort>(numOffsets);
	}

	timer("Symbols:    Huffman Decode");

	// decode symbols
	memcpy(pInstance->Encode.pUpload0, pStreamInfos.data(), blockCount * sizeof(HuffmanGPUStreamInfo));
	huffmanDecode(pInstance, (HuffmanGPUStreamInfo*)pInstance->Encode.pUpload0, blockCount, pInstance->m_offsetInterval);

	timer("ZeroCounts: Upload");

	// upload decode tables, codeword streams and offsets, and fill stream infos for zero counts
	// (before overwriting stream infos, wait for previous upload to complete)
	for(uint block = 0; block < blockCount; block++) {
		HuffmanGPUStreamInfo& streamInfo = pStreamInfos[block];

		// upload zero count decode table
		zeroCountDecodeTables[block].uploadToGPUAsync(streamInfo.dpDecodeTable);
		streamInfo.decodeSymbolTableSize = zeroCountDecodeTables[block].m_symbolTableSize;

		// upload zero count codewords and offsets
		// assume the bitstream memory is already page-locked
		cudaSafeCall(cudaMemcpyAsync(streamInfo.dpCodewordStream, pCodewordStreams[block], codewordUints[block] * sizeof(uint), cudaMemcpyHostToDevice));

		uint numOffsets = getNumOffsets(symbolCountsCompact[block], pInstance->m_offsetInterval);
		// assume the bitstream memory is already page-locked
		cudaSafeCall(cudaMemcpyAsync(streamInfo.dpOffsets, pOffsets[block], numOffsets * sizeof(ushort), cudaMemcpyHostToDevice));

		// unpack incremental offsets
		unpackInc16(pInstance, streamInfo.dpOffsets, (ushort*)streamInfo.dpOffsets, numOffsets);

		// fill stream infos
		streamInfo.dpSymbolStream = pdpZeroCounts[block];
		streamInfo.symbolCount = symbolCountsCompact[block];
	}

	timer("ZeroCounts: Huffman Decode");

	// decode zero counts
	memcpy(pInstance->Encode.pUpload1, pStreamInfos.data(), blockCount * sizeof(HuffmanGPUStreamInfo));
	huffmanDecode(pInstance, (HuffmanGPUStreamInfo*)pInstance->Encode.pUpload1, blockCount, pInstance->m_offsetInterval);

	timer("Run Length Decode");

	// run length decode
	std::vector<uint> symbolCountsPerBlock(blockCount, symbolCountPerBlock);
	runLengthDecode(pInstance, (const Symbol**)pdpSymbolStreamsCompacted.data(), (const ushort**)pdpZeroCounts.data(), symbolCountsCompact.data(), pdpSymbolStreams, symbolCountsPerBlock.data(), blockCount);

	timer();

	cudaSafeCall(cudaEventRecord(pInstance->Encode.encodeFinishedEvent));

	pInstance->releaseBuffers(7 * blockCount);

	return true;
}

bool encodeHuff(Instance* pInstance, BitStream* ppBitStreams[], bool singleBitStream, /*const*/ Symbol* const pdpSymbolStreams[], uint blockCount, uint symbolCountPerBlock)
{
	uint symbolStreamMaxBytes = symbolCountPerBlock * sizeof(Symbol);
	uint offsetStreamMaxBytes = getNumOffsets(symbolCountPerBlock, pInstance->m_offsetInterval) * sizeof(uint);

	std::vector<HuffmanGPUStreamInfo> pStreamInfos(blockCount);
	for(uint block = 0; block < blockCount; block++) {
		HuffmanGPUStreamInfo& streamInfo = pStreamInfos[block];

		streamInfo.dpEncodeCodeWords       = pInstance->getBuffer<uint>(MAX_DISTINCT_SYMBOLS);
		streamInfo.dpEncodeCodeWordLengths = pInstance->getBuffer<uint>(MAX_DISTINCT_SYMBOLS);

		streamInfo.dpDecodeTable = pInstance->getBuffer<byte>(HuffmanDecodeTable::computeMaxGPUSize());

		streamInfo.dpCodewordStream = (uint*)pInstance->getBuffer<byte>(symbolStreamMaxBytes);
		streamInfo.dpOffsets        = (uint*)pInstance->getBuffer<byte>(offsetStreamMaxBytes);
	}

	cudaSafeCall(cudaEventSynchronize(pInstance->Encode.encodeFinishedEvent));

	CudaScopedTimer timer(pInstance->Encode.timerEncode);

	timer("Symbols:    Design Huffman Tables");

	for(uint block = 0; block < blockCount; block++) {
		// padding for histogram (which wants the element count to be a multiple of 8)
		histogramPadData(pdpSymbolStreams[block], symbolCountPerBlock);
	}

	// build encode tables
	std::vector<HuffmanEncodeTable>& symbolEncodeTables = pInstance->Encode.symbolEncodeTables;
	std::vector<uint> symbolCount(blockCount, symbolCountPerBlock);
	if(!HuffmanEncodeTable::design(pInstance, symbolEncodeTables.data(), blockCount, (const Symbol**)pdpSymbolStreams, symbolCount.data())) {
		pInstance->releaseBuffers(blockCount * 5);
		return false;
	}

	timer("Symbols:    Upload Huffman Tables");

	// fill stream infos, upload encode tables
	for(uint block = 0; block < blockCount; block++) {
		HuffmanGPUStreamInfo& streamInfo = pStreamInfos[block];

		streamInfo.dpSymbolStream = pdpSymbolStreams[block];
		streamInfo.symbolCount = symbolCount[block];

		symbolEncodeTables[block].uploadToGPUAsync(streamInfo.dpEncodeCodeWords, streamInfo.dpEncodeCodeWordLengths);
	}

	timer("Symbols:    Huffman Encode");

	// encode the symbols
	std::vector<uint> codewordBitsizeSymbols(blockCount);
	huffmanEncode(pInstance, pStreamInfos.data(), blockCount, pInstance->m_offsetInterval, codewordBitsizeSymbols.data());

	timer("Symbols:    Pack Offsets");

	// make offsets incremental and pack
	for(uint block = 0; block < blockCount; block++) {
		HuffmanGPUStreamInfo& streamInfo = pStreamInfos[block];
		packInc16(pInstance, streamInfo.dpOffsets, (ushort*)streamInfo.dpOffsets, getNumOffsets(symbolCountPerBlock, pInstance->m_offsetInterval));
	}

	timer("Symbols:    Download");

	// download encoded symbols and offsets
	for(uint block = 0; block < blockCount; block++) {
		HuffmanGPUStreamInfo& streamInfo = pStreamInfos[block];

		uint codewordUints = getNumUintsForBits(codewordBitsizeSymbols[block]);
		cudaSafeCall(cudaMemcpyAsync(pInstance->Encode.ppCodewordBuffers[block], streamInfo.dpCodewordStream, codewordUints * sizeof(uint), cudaMemcpyDeviceToHost));
		uint numOffsets = getNumOffsets(symbolCountPerBlock, pInstance->m_offsetInterval);
		cudaSafeCall(cudaMemcpyAsync(pInstance->Encode.ppOffsetBuffers[block], streamInfo.dpOffsets, numOffsets * sizeof(ushort), cudaMemcpyDeviceToHost));
		cudaSafeCall(cudaEventRecord(pInstance->Encode.pSyncEvents[block]));
	}

	timer();


	// write to bitstream
	for(uint block = 0; block < blockCount; block++) {
		BitStream& bitStream = *ppBitStreams[singleBitStream ? 0 : block];
		uint bitStreamPosStart = bitStream.getBitPosition();

		uint bitStreamPosOld = bitStreamPosStart;
		uint bitStreamPos;

		// write encode table
		symbolEncodeTables[block].writeToBitstream(bitStream);

		bitStreamPos = bitStream.getBitPosition();
		//g_bitsSymbolEncodeTables += bitStreamPos - bitStreamPosOld;
		bitStreamPosOld = bitStreamPos;

		// write codeword stream
		bitStream.writeAligned(&codewordBitsizeSymbols[block], 1);
		uint codewordUints = getNumUintsForBits(codewordBitsizeSymbols[block]);
		cudaSafeCall(cudaEventSynchronize(pInstance->Encode.pSyncEvents[block]));
		bitStream.writeAligned(pInstance->Encode.ppCodewordBuffers[block], codewordUints);

		bitStreamPos = bitStream.getBitPosition();
		//g_bitsSymbolCodewords += bitStreamPos - bitStreamPosOld;
		bitStreamPosOld = bitStreamPos;

		// write offsets
		uint numOffsets = getNumOffsets(symbolCountPerBlock, pInstance->m_offsetInterval);
		bitStream.writeAligned((ushort*)pInstance->Encode.ppOffsetBuffers[block], numOffsets);

		bitStreamPos = bitStream.getBitPosition();
		//g_bitsSymbolOffsets += bitStreamPos - bitStreamPosOld;
		bitStreamPosOld = bitStreamPos;


		//g_bitsTotal += bitStreamPos - bitStreamPosStart;
	}
	//g_totalEncodedCount += blockCount * symbolCountPerBlock;

	cudaSafeCall(cudaEventRecord(pInstance->Encode.encodeFinishedEvent));

	pInstance->releaseBuffers(blockCount * 5);

	return true;
}

bool decodeHuff(Instance* pInstance, BitStreamReadOnly* ppBitStreams[], bool singleBitStream, Symbol* pdpSymbolStreams[], uint blockCount, uint symbolCountPerBlock)
{
	uint symbolStreamMaxBytes = symbolCountPerBlock * sizeof(Symbol);
	uint offsetStreamMaxBytes = getNumOffsets(symbolCountPerBlock, pInstance->m_offsetInterval) * sizeof(uint);

	std::vector<HuffmanGPUStreamInfo> pStreamInfos(blockCount);
	for(uint block = 0; block < blockCount; block++) {
		HuffmanGPUStreamInfo& streamInfo = pStreamInfos[block];

		streamInfo.dpEncodeCodeWords       = pInstance->getBuffer<uint>(MAX_DISTINCT_SYMBOLS);
		streamInfo.dpEncodeCodeWordLengths = pInstance->getBuffer<uint>(MAX_DISTINCT_SYMBOLS);

		streamInfo.dpDecodeTable = pInstance->getBuffer<byte>(HuffmanDecodeTable::computeMaxGPUSize());

		streamInfo.dpCodewordStream = (uint*)pInstance->getBuffer<byte>(symbolStreamMaxBytes);
		streamInfo.dpOffsets        = (uint*)pInstance->getBuffer<byte>(offsetStreamMaxBytes);
	}

	cudaSafeCall(cudaEventSynchronize(pInstance->Encode.encodeFinishedEvent));

	CudaScopedTimer timer(pInstance->Encode.timerDecode);

	timer("Symbols:    Upload (+read BitStream)");

	std::vector<uint> symbolCount(blockCount, symbolCountPerBlock);

	// read and upload decode tables, upload codeword streams and offsets, and fill stream info
	std::vector<HuffmanDecodeTable>& symbolDecodeTables = pInstance->Encode.symbolDecodeTables;
	std::vector<const uint*> pCodewordStreams(blockCount);
	std::vector<uint> codewordUints(blockCount);
	for(uint block = 0; block < blockCount; block++) {
		BitStreamReadOnly& bitStream = *ppBitStreams[singleBitStream ? 0 : block];

		// read decode table
		symbolDecodeTables[block].readFromBitstream(bitStream);

		HuffmanGPUStreamInfo& streamInfo = pStreamInfos[block];

		// upload decode table
		symbolDecodeTables[block].uploadToGPUAsync(streamInfo.dpDecodeTable);
		streamInfo.decodeSymbolTableSize = symbolDecodeTables[block].m_symbolTableSize;

		// upload codewords
		uint codewordBitsize;
		bitStream.readAligned(&codewordBitsize, 1);
		codewordUints[block] = getNumUintsForBits(codewordBitsize);
		pCodewordStreams[block] = bitStream.getRaw() + bitStream.getBitPosition() / (sizeof(uint)*8);
		// assume the bitstream memory is already page-locked
		cudaSafeCall(cudaMemcpyAsync(streamInfo.dpCodewordStream, pCodewordStreams[block], codewordUints[block] * sizeof(uint), cudaMemcpyHostToDevice));
		bitStream.skipBits(codewordUints[block] * sizeof(uint) * 8);

		// upload offsets
		uint numOffsets = getNumOffsets(symbolCountPerBlock, pInstance->m_offsetInterval);
		bitStream.align<uint>();
		const uint* pOffset = bitStream.getRaw() + bitStream.getBitPosition() / (sizeof(uint)*8);
		// assume the bitstream memory is already page-locked
		cudaSafeCall(cudaMemcpyAsync(streamInfo.dpOffsets, pOffset, numOffsets * sizeof(ushort), cudaMemcpyHostToDevice));
		bitStream.skipAligned<ushort>(numOffsets);

		// unpack incremental offsets
		unpackInc16(pInstance, streamInfo.dpOffsets, (ushort*)streamInfo.dpOffsets, numOffsets);

		// fill stream infos
		streamInfo.dpSymbolStream = pdpSymbolStreams[block];
		streamInfo.symbolCount = symbolCount[block];
	}

	timer("Symbols:    Huffman Decode");

	// decode symbols
	memcpy(pInstance->Encode.pUpload0, pStreamInfos.data(), blockCount * sizeof(HuffmanGPUStreamInfo));
	huffmanDecode(pInstance, (HuffmanGPUStreamInfo*)pInstance->Encode.pUpload0, blockCount, pInstance->m_offsetInterval);

	timer();

	cudaSafeCall(cudaEventRecord(pInstance->Encode.encodeFinishedEvent));

	pInstance->releaseBuffers(blockCount * 5);

	return true;
}


//void encodeResetBitCounts()
//{
//	g_bitsSymbolEncodeTables = 0;
//	g_bitsSymbolCodewords = 0;
//	g_bitsSymbolOffsets = 0;
//	g_bitsZeroCountEncodeTables = 0;
//	g_bitsZeroCountCodewords = 0;
//	g_bitsZeroCountOffsets = 0;
//	g_bitsTotal = 0;
//	g_totalEncodedCount = 0;
//}

//static float getBitsPercent(uint bitCount)
//{
//	return 100.0f * float(bitCount) / float(g_bitsTotal);
//}

//void encodePrintBitCounts()
//{
//	float bps = float(g_bitsTotal) / float(g_totalEncodedCount);
//	printf("Encoded %u symbols to %u bits (%.2f bps)\n", g_totalEncodedCount, g_bitsTotal, bps);
//	printf("Symbol Encode Tables    : %9u bits (%5.2f%%)\n", g_bitsSymbolEncodeTables,    getBitsPercent(g_bitsSymbolEncodeTables));
//	printf("Symbol Codewords        : %9u bits (%5.2f%%)\n", g_bitsSymbolCodewords,       getBitsPercent(g_bitsSymbolCodewords));
//	printf("Symbol Offsets          : %9u bits (%5.2f%%)\n", g_bitsSymbolOffsets,         getBitsPercent(g_bitsSymbolOffsets));
//	printf("ZeroCount Encode Tables : %9u bits (%5.2f%%)\n", g_bitsZeroCountEncodeTables, getBitsPercent(g_bitsZeroCountEncodeTables));
//	printf("ZeroCount Codewords     : %9u bits (%5.2f%%)\n", g_bitsZeroCountCodewords,    getBitsPercent(g_bitsZeroCountCodewords));
//	printf("ZeroCount Offsets       : %9u bits (%5.2f%%)\n", g_bitsZeroCountOffsets,      getBitsPercent(g_bitsZeroCountOffsets));
//}

}
