#include "EncodeCPU.h"

#include <cassert>

#include "cudaUtil.h"
#include "util.h"

#include "HuffmanCPU.h"
#include "HuffmanTableCPU.h"
#include "RiceGolombCPU.h"
#include "RunLengthCPU.h"
#include "RBUCCPU.h"


namespace cudaCompress {

std::string getCoderName(ECoder coder)
{
	switch(coder) {
		case CODER_RUNLENGTH_HUFFMAN:
			return "RunLength-Huffman";
		case CODER_HUFFMAN:
			return "Huffman";
		case CODER_RUNLENGTH_GOLOMBRICE:
			return "RunLength-GolombRice";
		case CODER_GOLOMBRICE:
			return "GolombRice";
		case CODER_RBUC:
			return "RBUC";
		default:
			return false;
	}
}



bool encodeCPU(ECoder coder, BitStream* pBitStream, /*const*/ std::vector<Symbol>* pSymbolStream, uint offsetInterval)
{
	return encodeCPU(coder, &pBitStream, &pSymbolStream, 1, offsetInterval);
}

bool decodeCPU(ECoder coder, BitStreamReadOnly* pBitStream, std::vector<Symbol>* pSymbolStream, uint symbolCount, uint offsetInterval)
{
	return decodeCPU(coder, &pBitStream, &pSymbolStream, symbolCount, 1, offsetInterval);
}



bool encodeCPU(ECoder coder, BitStream* ppBitStreams[], /*const*/ std::vector<Symbol>* ppSymbolStreams[], uint blockCount, uint offsetInterval)
{
	switch(coder) {
		case CODER_RUNLENGTH_HUFFMAN:
			return encodeRLHuffCPU(ppBitStreams, ppSymbolStreams, blockCount, offsetInterval);
		case CODER_HUFFMAN:
			return encodeHuffCPU(ppBitStreams, ppSymbolStreams, blockCount, offsetInterval);
		case CODER_RUNLENGTH_GOLOMBRICE:
			return encodeRLRiceCPU(ppBitStreams, ppSymbolStreams, blockCount, offsetInterval);
		case CODER_GOLOMBRICE:
			return encodeRiceCPU(ppBitStreams, ppSymbolStreams, blockCount, offsetInterval);
		case CODER_RBUC:
			return encodeRBUCCPU(ppBitStreams, ppSymbolStreams, blockCount);
		default:
			return false;
	}
}

bool decodeCPU(ECoder coder, BitStreamReadOnly* ppBitStreams[], std::vector<Symbol>* ppSymbolStreams[], uint symbolCountPerBlock, uint blockCount, uint offsetInterval)
{
	switch(coder) {
		case CODER_RUNLENGTH_HUFFMAN:
			return decodeRLHuffCPU(ppBitStreams, ppSymbolStreams, symbolCountPerBlock, blockCount, offsetInterval);
		case CODER_HUFFMAN:
			return decodeHuffCPU(ppBitStreams, ppSymbolStreams, symbolCountPerBlock, blockCount, offsetInterval);
		case CODER_RUNLENGTH_GOLOMBRICE:
			return decodeRLRiceCPU(ppBitStreams, ppSymbolStreams, symbolCountPerBlock, blockCount, offsetInterval);
		case CODER_GOLOMBRICE:
			return decodeRiceCPU(ppBitStreams, ppSymbolStreams, symbolCountPerBlock, blockCount, offsetInterval);
		case CODER_RBUC:
			return decodeRBUCCPU(ppBitStreams, ppSymbolStreams, symbolCountPerBlock, blockCount);
		default:
			return false;
	}
}


bool encodeRLHuffCPU(BitStream* ppBitStreams[], /*const*/ std::vector<Symbol>* ppSymbolStreams[], uint blockCount, uint offsetInterval, bool zeroRunsOnly)
{
	for(uint block = 0; block < blockCount; block++) {
		BitStream& bitStream = *ppBitStreams[block];

		// run length encode
		std::vector<Symbol> symbolsCompact;
		std::vector<ushort> zeroCounts;
		if(zeroRunsOnly) {
			runLengthZeroesEncodeCPU(symbolsCompact, zeroCounts, *ppSymbolStreams[block], ZERO_COUNT_MAX);
		} else {
			runLengthEncodeCPU(symbolsCompact, zeroCounts, *ppSymbolStreams[block], ZERO_COUNT_MAX);
		}

		// write compacted symbol count
		uint compactedSymbolCount = (uint)symbolsCompact.size();
		bitStream.writeAligned(&compactedSymbolCount, 1);


		// 1. compacted symbols
		HuffmanTableCPU table;
		table.design(symbolsCompact);

		table.writeToBitstream(bitStream);

		HuffmanDecodeTableCPU decodeTable;
		decodeTable.build(table);

		HuffmanEncodeTableCPU encodeTable;
		encodeTable.build(decodeTable);

		// leave room for codeword stream bitsize
		bitStream.writeAlign<uint>();
		uint codewordBitsizeIndex = bitStream.getBitPosition() / (8 * sizeof(uint));
		uint codewordBitsize = 0;
		bitStream.writeAligned(&codewordBitsize, 1);

		// encode, write codeword stream to bitstream
		std::vector<uint> offsets;
		uint codewordStartBit = bitStream.getBitPosition();
		huffmanEncodeCPU(bitStream, symbolsCompact, encodeTable, offsets, offsetInterval);

		// write codeword bitsize to reserved position
		codewordBitsize = bitStream.getBitPosition() - codewordStartBit;
		bitStream.getRaw()[codewordBitsizeIndex] = codewordBitsize;

		bitStream.writeAlign<uint>();

		// make offsets incremental
		std::vector<ushort> offsetIncrements(offsets.size());
		for(uint i = 1; i < (uint)offsets.size(); i++) {
			offsetIncrements[i] = offsets[i] - offsets[i-1];
		}

		// write offsets
		//bitStream.writeAligned(&offsetBits, 1);
		//for(uint i = 0; i < offsets.size(); i++)
		//	bitStream.writeBits(offsets[i], offsetBits);
		//bitStream.writeAlign<uint>();
		bitStream.writeAligned(offsetIncrements.data(), (uint)offsetIncrements.size());


		// 2. zero counts
		table.design(zeroCounts);
		table.writeToBitstream(bitStream);

		decodeTable.build(table);
		encodeTable.build(decodeTable);

		// leave room for codeword stream bitsize
		bitStream.writeAlign<uint>();
		codewordBitsizeIndex = bitStream.getBitPosition() / (8 * sizeof(uint));
		codewordBitsize = 0;
		bitStream.writeAligned(&codewordBitsize, 1);

		// encode, write codeword stream to bitstream
		offsets.clear();
		codewordStartBit = bitStream.getBitPosition();
		huffmanEncodeCPU(bitStream, zeroCounts, encodeTable, offsets, offsetInterval);

		// write codeword bitsize to reserved position
		codewordBitsize = bitStream.getBitPosition() - codewordStartBit;
		bitStream.getRaw()[codewordBitsizeIndex] = codewordBitsize;

		bitStream.writeAlign<uint>();

		// make offsets incremental
		for(uint i = 1; i < (uint)offsets.size(); i++) {
			offsetIncrements[i] = offsets[i] - offsets[i-1];
		}

		// write offsets
		//bitStream.writeAligned(&offsetBits, 1);
		//for(uint i = 0; i < offsets.size(); i++)
		//	bitStream.writeBits(offsets[i], offsetBits);
		//bitStream.writeAlign<uint>();
		bitStream.writeAligned(offsetIncrements.data(), (uint)offsetIncrements.size());
	}

	return true;
}

bool decodeRLHuffCPU(BitStreamReadOnly* ppBitStreams[], std::vector<Symbol>* ppSymbolStreams[], uint symbolCountPerBlock, uint blockCount, uint offsetInterval, bool zeroRunsOnly)
{
	for(uint block = 0; block < blockCount; block++) {
		BitStreamReadOnly& bitStream = *ppBitStreams[block];

		// read compacted symbol count
		uint compactedSymbolCount;
		bitStream.readAligned(&compactedSymbolCount, 1);


		// 1. compacted symbols
		HuffmanTableCPU table;
		table.readFromBitstream(bitStream);

		HuffmanDecodeTableCPU decodeTable;
		decodeTable.build(table);

		uint codewordBitsize = 0;
		bitStream.readAligned(&codewordBitsize, 1);

		uint codewordStartBit = bitStream.getBitPosition();
		std::vector<Symbol> symbolsCompact;
		huffmanDecodeCPU(bitStream, compactedSymbolCount, symbolsCompact, decodeTable);

		assert(bitStream.getBitPosition() - codewordStartBit == codewordBitsize);

		bitStream.align<uint>();

		uint offsetCount = (compactedSymbolCount + offsetInterval - 1) / offsetInterval;
		// skip offsets
		//uint offsetBits = 0;
		//bitStream.readAligned(&offsetBits, 1);
		//bitStream.skipBits(offsetCount * offsetBits);
		//bitStream.align<uint>();
		bitStream.skipAligned<ushort>(offsetCount);


		// 2. zero counts
		table.readFromBitstream(bitStream);
		decodeTable.build(table);

		codewordBitsize = 0;
		bitStream.readAligned(&codewordBitsize, 1);

		codewordStartBit = bitStream.getBitPosition();
		std::vector<ushort> zeroCounts;
		huffmanDecodeCPU(bitStream, compactedSymbolCount, zeroCounts, decodeTable);

		assert(bitStream.getBitPosition() - codewordStartBit == codewordBitsize);

		bitStream.align<uint>();

		// skip offsets
		//offsetBits = 0;
		//bitStream.readAligned(&offsetBits, 1);
		//bitStream.skipBits(offsetCount * offsetBits);
		//bitStream.align<uint>();
		bitStream.skipAligned<ushort>(offsetCount);


		// run length decode
		if(zeroRunsOnly) {
			runLengthZeroesDecodeCPU(symbolsCompact, zeroCounts, *ppSymbolStreams[block], symbolCountPerBlock);
		} else {
			runLengthDecodeCPU(symbolsCompact, zeroCounts, *ppSymbolStreams[block], symbolCountPerBlock);
		}
	}

	return true;
}



bool encodeHuffCPU(BitStream* ppBitStreams[], /*const*/ std::vector<Symbol>* ppSymbolStreams[], uint blockCount, uint offsetInterval)
{
	for(uint block = 0; block < blockCount; block++) {
		BitStream& bitStream = *ppBitStreams[block];

		HuffmanTableCPU table;
		table.design(*ppSymbolStreams[block]);

		table.writeToBitstream(bitStream);

		HuffmanDecodeTableCPU decodeTable;
		decodeTable.build(table);

		HuffmanEncodeTableCPU encodeTable;
		encodeTable.build(decodeTable);

		// leave room for codeword stream bitsize
		bitStream.writeAlign<uint>();
		uint codewordBitsizeIndex = bitStream.getBitPosition() / (8 * sizeof(uint));
		uint codewordBitsize = 0;
		bitStream.writeAligned(&codewordBitsize, 1);

		// encode, write codeword stream to bitstream
		std::vector<uint> offsets;
		uint codewordStartBit = bitStream.getBitPosition();
		huffmanEncodeCPU(bitStream, *ppSymbolStreams[block], encodeTable, offsets, offsetInterval);

		// write codeword bitsize to reserved position
		codewordBitsize = bitStream.getBitPosition() - codewordStartBit;
		bitStream.getRaw()[codewordBitsizeIndex] = codewordBitsize;

		bitStream.writeAlign<uint>();

		// make offsets incremental
		std::vector<ushort> offsetIncrements(offsets.size());
		for(uint i = 1; i < (uint)offsets.size(); i++) {
			offsetIncrements[i] = offsets[i] - offsets[i-1];
		}

		// write offsets
		//bitStream.writeAligned(&offsetBits, 1);
		//for(uint i = 0; i < offsets.size(); i++)
		//	bitStream.writeBits(offsets[i], offsetBits);
		//bitStream.writeAlign<uint>();
		bitStream.writeAligned(offsetIncrements.data(), (uint)offsetIncrements.size());
	}

	return true;
}

bool decodeHuffCPU(BitStreamReadOnly* ppBitStreams[], std::vector<Symbol>* ppSymbolStreams[], uint symbolCountPerBlock, uint blockCount, uint offsetInterval)
{
	for(uint block = 0; block < blockCount; block++) {
		BitStreamReadOnly& bitStream = *ppBitStreams[block];

		HuffmanTableCPU table;
		table.readFromBitstream(bitStream);

		HuffmanDecodeTableCPU decodeTable;
		decodeTable.build(table);

		uint codewordBitsize = 0;
		bitStream.readAligned(&codewordBitsize, 1);

		uint codewordStartBit = bitStream.getBitPosition();
		huffmanDecodeCPU(bitStream, symbolCountPerBlock, *ppSymbolStreams[block], decodeTable);

		assert(bitStream.getBitPosition() - codewordStartBit == codewordBitsize);

		bitStream.align<uint>();

		uint offsetCount = (symbolCountPerBlock + offsetInterval - 1) / offsetInterval;
		// skip offsets
		//uint offsetBits = 0;
		//bitStream.readAligned(&offsetBits, 1);
		//bitStream.skipBits(offsetCount * offsetBits);
		//bitStream.align<uint>();
		bitStream.skipAligned<ushort>(offsetCount);
	}

	return true;
}



bool encodeRLRiceCPU(BitStream* ppBitStreams[], /*const*/ std::vector<Symbol>* ppSymbolStreams[], uint blockCount, uint offsetInterval, bool zeroRunsOnly)
{
	for(uint block = 0; block < blockCount; block++) {
		BitStream& bitStream = *ppBitStreams[block];

		// run length encode
		std::vector<Symbol> symbolsCompact;
		std::vector<ushort> zeroCounts;
		if(zeroRunsOnly) {
			runLengthZeroesEncodeCPU(symbolsCompact, zeroCounts, *ppSymbolStreams[block], ZERO_COUNT_MAX);
		} else {
			runLengthEncodeCPU(symbolsCompact, zeroCounts, *ppSymbolStreams[block], ZERO_COUNT_MAX);
		}

		// write compacted symbol count
		uint compactedSymbolCount = (uint)symbolsCompact.size();
		bitStream.writeAligned(&compactedSymbolCount, 1);


		// 1. compacted symbols

		// leave room for codeword stream bitsize
		bitStream.writeAlign<uint>();
		uint codewordBitsizeIndex = bitStream.getBitPosition() / (8 * sizeof(uint));
		uint codewordBitsize = 0;
		bitStream.writeAligned(&codewordBitsize, 1);

		// encode, write codeword stream to bitstream
		std::vector<uint> offsets;
		uint codewordStartBit = bitStream.getBitPosition();
		riceGolombEncodeCPU(bitStream, symbolsCompact, offsetInterval, offsets, 1);

		// write codeword bitsize to reserved position
		codewordBitsize = bitStream.getBitPosition() - codewordStartBit;
		bitStream.getRaw()[codewordBitsizeIndex] = codewordBitsize;

		bitStream.writeAlign<uint>();

		// make offsets incremental
		std::vector<ushort> offsetIncrements(offsets.size());
		for(uint i = 1; i < (uint)offsets.size(); i++) {
			offsetIncrements[i] = offsets[i] - offsets[i-1];
		}

		// write offsets
		bitStream.writeAligned(offsetIncrements.data(), (uint)offsetIncrements.size());


		// 2. zero counts

		// leave room for codeword stream bitsize
		bitStream.writeAlign<uint>();
		codewordBitsizeIndex = bitStream.getBitPosition() / (8 * sizeof(uint));
		codewordBitsize = 0;
		bitStream.writeAligned(&codewordBitsize, 1);

		// encode, write codeword stream to bitstream
		offsets.clear();
		codewordStartBit = bitStream.getBitPosition();
		riceGolombEncodeCPU(bitStream, zeroCounts, offsetInterval, offsets, 1);

		// write codeword bitsize to reserved position
		codewordBitsize = bitStream.getBitPosition() - codewordStartBit;
		bitStream.getRaw()[codewordBitsizeIndex] = codewordBitsize;

		bitStream.writeAlign<uint>();

		// make offsets incremental
		for(uint i = 1; i < (uint)offsets.size(); i++) {
			offsetIncrements[i] = offsets[i] - offsets[i-1];
		}

		// write offsets
		bitStream.writeAligned(offsetIncrements.data(), (uint)offsetIncrements.size());
	}

	return true;
}

bool decodeRLRiceCPU(BitStreamReadOnly* ppBitStreams[], std::vector<Symbol>* ppSymbolStreams[], uint symbolCountPerBlock, uint blockCount, uint offsetInterval, bool zeroRunsOnly)
{
	for(uint block = 0; block < blockCount; block++) {
		BitStreamReadOnly& bitStream = *ppBitStreams[block];

		// read compacted symbol count
		uint compactedSymbolCount;
		bitStream.readAligned(&compactedSymbolCount, 1);


		// 1. compacted symbols
		uint codewordBitsize = 0;
		bitStream.readAligned(&codewordBitsize, 1);

		uint codewordStartBit = bitStream.getBitPosition();
		std::vector<Symbol> symbolsCompact;
		riceGolombDecodeCPU(bitStream, compactedSymbolCount, symbolsCompact, offsetInterval);

		assert(bitStream.getBitPosition() - codewordStartBit == codewordBitsize);

		bitStream.align<uint>();

		uint offsetCount = (compactedSymbolCount + offsetInterval - 1) / offsetInterval;
		// skip offsets
		bitStream.skipAligned<ushort>(offsetCount);


		// 2. zero counts
		codewordBitsize = 0;
		bitStream.readAligned(&codewordBitsize, 1);

		codewordStartBit = bitStream.getBitPosition();
		std::vector<ushort> zeroCounts;
		riceGolombDecodeCPU(bitStream, compactedSymbolCount, zeroCounts, offsetInterval);

		assert(bitStream.getBitPosition() - codewordStartBit == codewordBitsize);

		bitStream.align<uint>();

		// skip offsets
		bitStream.skipAligned<ushort>(offsetCount);


		// run length decode
		if(zeroRunsOnly) {
			runLengthZeroesDecodeCPU(symbolsCompact, zeroCounts, *ppSymbolStreams[block], symbolCountPerBlock);
		} else {
			runLengthDecodeCPU(symbolsCompact, zeroCounts, *ppSymbolStreams[block], symbolCountPerBlock);
		}
	}

	return true;
}



bool encodeRiceCPU(BitStream* ppBitStreams[], /*const*/ std::vector<Symbol>* ppSymbolStreams[], uint blockCount, uint offsetInterval)
{
	for(uint block = 0; block < blockCount; block++) {
		BitStream& bitStream = *ppBitStreams[block];

		// leave room for codeword stream bitsize
		bitStream.writeAlign<uint>();
		uint codewordBitsizeIndex = bitStream.getBitPosition() / (8 * sizeof(uint));
		uint codewordBitsize = 0;
		bitStream.writeAligned(&codewordBitsize, 1);

		// encode, write codeword stream to bitstream
		std::vector<uint> offsets;
		uint codewordStartBit = bitStream.getBitPosition();
		riceGolombEncodeCPU(bitStream, *ppSymbolStreams[block], offsetInterval, offsets, 1);

		// write codeword bitsize to reserved position
		codewordBitsize = bitStream.getBitPosition() - codewordStartBit;
		bitStream.getRaw()[codewordBitsizeIndex] = codewordBitsize;

		bitStream.writeAlign<uint>();

		// make offsets incremental
		std::vector<ushort> offsetIncrements(offsets.size());
		for(uint i = 1; i < (uint)offsets.size(); i++) {
			offsetIncrements[i] = offsets[i] - offsets[i-1];
		}

		// write offsets
		bitStream.writeAligned(offsetIncrements.data(), (uint)offsetIncrements.size());
	}

	return true;
}

bool decodeRiceCPU(BitStreamReadOnly* ppBitStreams[], std::vector<Symbol>* ppSymbolStreams[], uint symbolCountPerBlock, uint blockCount, uint offsetInterval)
{
	for(uint block = 0; block < blockCount; block++) {
		BitStreamReadOnly& bitStream = *ppBitStreams[block];

		uint codewordBitsize = 0;
		bitStream.readAligned(&codewordBitsize, 1);

		uint codewordStartBit = bitStream.getBitPosition();
		riceGolombDecodeCPU(bitStream, symbolCountPerBlock, *ppSymbolStreams[block], offsetInterval);

		assert(bitStream.getBitPosition() - codewordStartBit == codewordBitsize);

		bitStream.align<uint>();

		uint offsetCount = (symbolCountPerBlock + offsetInterval - 1) / offsetInterval;
		// skip offsets
		bitStream.skipAligned<ushort>(offsetCount);
	}

	return true;
}



bool encodeRBUCCPU(BitStream* ppBitStreams[], /*const*/ std::vector<Symbol>* ppSymbolStreams[], uint blockCount)
{
	std::vector<uint> branchFactors;
	branchFactors.push_back(3);
	branchFactors.push_back(6);
	branchFactors.push_back(9);

	for(uint block = 0; block < blockCount; block++) {
		BitStream& bitStream = *ppBitStreams[block];

		// leave room for codeword stream bitsize
		bitStream.align<uint>();
		uint codewordBitsizeIndex = bitStream.getBitPosition() / (8 * sizeof(uint));
		uint codewordBitsize = 0;
		bitStream.writeAligned(&codewordBitsize, 1);

		// encode, write codeword stream to bitstream
		std::vector<uint> offsets;
		uint codewordStartBit = bitStream.getBitPosition();
		rbucEncodeCPU(bitStream, offsets, *ppSymbolStreams[block], branchFactors);

		// write codeword bitsize to reserved position
		codewordBitsize = bitStream.getBitPosition() - codewordStartBit;
		bitStream.getRaw()[codewordBitsizeIndex] = codewordBitsize;

		bitStream.align<uint>();

		// make offsets incremental
		std::vector<ushort> offsetIncrements(offsets.size());
		for(uint i = 1; i < (uint)offsets.size(); i++) {
			offsetIncrements[i] = offsets[i] - offsets[i-1];
		}

		// write offsets
		//bitStream.writeAligned(&offsetBits, 1);
		//for(uint i = 0; i < offsets.size(); i++)
		//	bitStream.writeBits(offsets[i], offsetBits);
		//bitStream.align<uint>();
		bitStream.writeAligned(offsetIncrements.data(), (uint)offsetIncrements.size());
	}

	return true;
}

bool decodeRBUCCPU(BitStreamReadOnly* ppBitStreams[], std::vector<Symbol>* ppSymbolStreams[], uint symbolCountPerBlock, uint blockCount)
{
	std::vector<uint> branchFactors;
	branchFactors.push_back(3);
	branchFactors.push_back(6);
	branchFactors.push_back(9);
	uint offsetInterval = 3 * 6 * 9;

	for(uint block = 0; block < blockCount; block++) {
		BitStreamReadOnly& bitStream = *ppBitStreams[block];

		uint codewordBitsize = 0;
		bitStream.readAligned(&codewordBitsize, 1);

		uint codewordStartBit = bitStream.getBitPosition();
		rbucDecodeCPU(bitStream, *ppSymbolStreams[block], symbolCountPerBlock, branchFactors);

		//FIXME assert(bitStream.getBitPosition() - codewordStartBit == codewordBitsize);

		bitStream.align<uint>();

		uint offsetCount = (symbolCountPerBlock + offsetInterval - 1) / offsetInterval;
		// skip offsets
		//uint offsetBits = 0;
		//bitStream.readAligned(&offsetBits, 1);
		//bitStream.skipBits(offsetCount * offsetBits);
		//bitStream.align<uint>();
		bitStream.skipAligned<ushort>(offsetCount);
	}

	return true;
}

}
