#include "Huffman.h"

#include <cassert>

#include <cuda_runtime.h>
//#include <thrust/device_ptr.h>
//#include <thrust/scan.h>

#include "tools/Vec.h"

#include "cudaUtil.h"
#include "util.h"

#include "Instance.h"

using namespace tum3D;


#include "scan/scan_app.cui"

#include "HuffmanKernels.cui"


namespace cudaCompress {


static uint getStreamCountMax(const Instance* pInstance)
{
	bool is3D = (pInstance->m_sizeMax.z() > 1);

	uint countMax = (is3D ? 7 : 4);
	countMax *= max(1u, pInstance->m_channelCountMax);
	return countMax;
}

static uint getSymbolCountPerStreamMax(const Instance* pInstance)
{
	Vec3i blockSizeMax = (pInstance->m_sizeMax + 1) / 2;
	return blockSizeMax.volume();
}


size_t huffmanGetRequiredMemory(const Instance* pInstance)
{
	uint streamCountMax = getStreamCountMax(pInstance);
	uint symbolCountPerStreamMax = getSymbolCountPerStreamMax(pInstance);

	size_t sizeEncode = 0;
	size_t sizeDecode = 0;

	// encode: dpScratch
	for(uint i = 0; i < streamCountMax; i++) {
		sizeEncode += getAlignedSize((getPrefixCount(symbolCountPerStreamMax) + 1) * sizeof(uint), 128);
	}

	// encode: dpScanTotal
	sizeEncode += getAlignedSize(streamCountMax * sizeof(uint), 128);

	// decode: dpStreamInfos
	sizeDecode += getAlignedSize(sizeof(HuffmanGPUStreamInfo) * streamCountMax, 128);

	return max(sizeEncode, sizeDecode);
}

bool huffmanInit(Instance* pInstance)
{
	uint streamCountMax = getStreamCountMax(pInstance);
	uint symbolCountPerStreamMax = getSymbolCountPerStreamMax(pInstance);

	cudaSafeCall(cudaMallocHost(&pInstance->Huffman.pReadback, streamCountMax * sizeof(uint)));

	cudaSafeCall(cudaEventCreateWithFlags(&pInstance->Huffman.syncEventReadback, cudaEventDisableTiming));

	return true;
}

bool huffmanShutdown(Instance* pInstance)
{
	cudaSafeCall(cudaEventDestroy(pInstance->Huffman.syncEventReadback));
	pInstance->Huffman.syncEventReadback = 0;

	cudaSafeCall(cudaFreeHost(pInstance->Huffman.pReadback));
	pInstance->Huffman.pReadback = nullptr;

	return true;
}


bool huffmanEncode(Instance* pInstance, const HuffmanGPUStreamInfo* pStreamInfos, uint streamCount, uint offsetInterval, uint* pCompressedSizeBits)
{
	uint streamCountMax = getStreamCountMax(pInstance);
	uint symbolCountPerStreamMax = getSymbolCountPerStreamMax(pInstance);

	assert(streamCount <= streamCountMax);

	std::vector<uint*> dpScratch(streamCount);
	for(uint i = 0; i < streamCount; i++) {
		dpScratch[i] = pInstance->getBuffer<uint>(getPrefixCount(pStreamInfos[i].symbolCount) + 1);
	}
	uint* dpScanTotal = pInstance->getBuffer<uint>(streamCountMax);

	CudaScopedTimer timer(pInstance->Huffman.timerEncode);

	timer("Codewords / Offsets");

	for(uint i = 0; i < streamCount; i++) {
		const HuffmanGPUStreamInfo& streamInfo = pStreamInfos[i];

		assert(streamInfo.symbolCount <= symbolCountPerStreamMax);

		if(streamInfo.symbolCount == 0) {
			continue;
		}

		uint prefixCount = getPrefixCount(streamInfo.symbolCount);

		// get codeword lengths (of COMPACTIFY_ELEM_PER_THREAD consecutive codewords)
		uint threadCountTotal = prefixCount;
		uint blockSize = WORDS_TO_LENGTH_THREADS_PER_BLOCK;
		uint blockCount = (threadCountTotal + blockSize - 1) / blockSize;

		huffmanEncodeWordsToLengthKernel<<<blockCount, blockSize>>>(dpScratch[i], streamInfo.symbolCount, streamInfo.dpSymbolStream, streamInfo.dpEncodeCodeWordLengths);
		cudaCheckMsg("huffmanEncodeWordsToLengthKernel execution failed");

		// scan codeword lengths to get output indices
		//thrust::device_ptr<uint> pScratchThrust(dpScratch[i]);
		//thrust::exclusive_scan(pScratchThrust, pScratchThrust + prefixCount + 1, pScratchThrust);
		//cudaCheckMsg("huffmanEncode: Error in thrust::exclusive_scan");
		scanArray<uint, uint, true>(dpScratch[i], dpScratch[i], prefixCount + 1, pInstance->m_pScanPlan);
		cudaCheckMsg("huffmanEncode: Error in scanArray");

		cudaSafeCall(cudaMemcpyAsync(dpScanTotal + i, dpScratch[i] + prefixCount, sizeof(uint), cudaMemcpyDeviceToDevice));

		// write out offsets
		assert(offsetInterval >= COMPACTIFY_ELEM_PER_THREAD);
		assert(offsetInterval % COMPACTIFY_ELEM_PER_THREAD == 0);
		threadCountTotal = (streamInfo.symbolCount + offsetInterval - 1) / offsetInterval;
		blockSize = 128;
		blockCount = (threadCountTotal + blockSize - 1) / blockSize;

		huffmanEncodeCollectOffsetsKernel<<<blockCount, blockSize>>>(streamInfo.dpOffsets, threadCountTotal, dpScratch[i], offsetInterval / COMPACTIFY_ELEM_PER_THREAD);
		cudaCheckMsg("huffmanEncodeCollectOffsetsKernel execution failed");
	}

	//timer("Readback Compressed Size");

	// Read compressed size
	cudaSafeCall(cudaMemcpyAsync(pInstance->Huffman.pReadback, dpScanTotal, streamCount * sizeof(uint), cudaMemcpyDeviceToHost));
	cudaSafeCall(cudaEventRecord(pInstance->Huffman.syncEventReadback));

	timer("Compactify");

	for(uint i = 0; i < streamCount; i++) {
		const HuffmanGPUStreamInfo& streamInfo = pStreamInfos[i];

		if(streamInfo.symbolCount == 0) {
			continue;
		}

		cudaSafeCall(cudaMemsetAsync(streamInfo.dpCodewordStream, 0, streamInfo.symbolCount * sizeof(Symbol)));

		uint threadCountTotal = getPrefixCount(streamInfo.symbolCount);
		uint blockSize = COMPACTIFY_THREADS_PER_BLOCK;
		uint blockCount = (threadCountTotal + blockSize - 1) / blockSize;

		huffmanEncodeCompactifyKernel<<<blockCount, blockSize>>>(
			streamInfo.dpCodewordStream, streamInfo.symbolCount, streamInfo.dpSymbolStream,
			dpScratch[i], streamInfo.dpEncodeCodeWords, streamInfo.dpEncodeCodeWordLengths);
		cudaCheckMsg("huffmanEncodeCompactifyKernel execution failed");
	}

	timer("Readback Sync");

	cudaSafeCall(cudaEventSynchronize(pInstance->Huffman.syncEventReadback));
	for(uint i = 0; i < streamCount; i++) {
		const HuffmanGPUStreamInfo& streamInfo = pStreamInfos[i];

		if(streamInfo.symbolCount == 0) {
			pCompressedSizeBits[i] = 0;
		} else {
			pCompressedSizeBits[i] = pInstance->Huffman.pReadback[i];
		}
	}

	timer();

	pInstance->releaseBuffers(streamCount + 1);

	return true;
}

bool huffmanDecode(Instance* pInstance, const HuffmanGPUStreamInfo* pStreamInfosPinned, uint streamCount, uint offsetInterval)
{
	uint streamCountMax = getStreamCountMax(pInstance);

	assert(streamCount <= streamCountMax);

	HuffmanGPUStreamInfo* dpStreamInfos = pInstance->getBuffer<HuffmanGPUStreamInfo>(streamCount);

	CudaScopedTimer timer(pInstance->Huffman.timerDecode);

	timer("Upload Info");

	// upload stream infos (assume pStreamInfos is in pinned memory)
	cudaSafeCall(cudaMemcpyAsync(dpStreamInfos, pStreamInfosPinned, sizeof(HuffmanGPUStreamInfo) * streamCount, cudaMemcpyHostToDevice));
	// note: we don't sync on this upload - we trust that the caller won't overwrite/delete the array...

	timer("Decode");

	// get max number of symbols
	uint symbolCountPerStreamMax = 0;
	for(uint i = 0; i < streamCount; i++)
		symbolCountPerStreamMax = max(symbolCountPerStreamMax, pStreamInfosPinned[i].symbolCount);

	if(symbolCountPerStreamMax == 0) {
		pInstance->releaseBuffer();
		return true;
	}

	// launch decode kernel
	uint threadCountPerStream = (symbolCountPerStreamMax + offsetInterval - 1) / offsetInterval;
	const uint blockSize = 192;
	static_assert(blockSize >= HUFFMAN_LOOKUP_SIZE, "blockSize must be >= HUFFMAN_LOOKUP_SIZE");
	dim3 blockCount((threadCountPerStream + blockSize - 1) / blockSize, streamCount);

	huffmanDecodeKernel<<<blockCount, blockSize>>>(dpStreamInfos, offsetInterval);
	cudaCheckMsg("huffmanDecodeKernel execution failed");

	timer("Transpose");

	// launch transpose kernel
	dim3 blockSizeTranspose(TRANSPOSE_BLOCKDIM_X, TRANSPOSE_BLOCKDIM_Y);
	dim3 blockCountTranspose((symbolCountPerStreamMax + WARP_SIZE * offsetInterval - 1) / (WARP_SIZE * offsetInterval), streamCount);

	switch(offsetInterval) {
		case 32:
			huffmanDecodeTransposeKernel<32><<<blockCountTranspose, blockSizeTranspose>>>(dpStreamInfos);
			break;
		case 64:
			huffmanDecodeTransposeKernel<64><<<blockCountTranspose, blockSizeTranspose>>>(dpStreamInfos);
			break;
		case 128:
			huffmanDecodeTransposeKernel<128><<<blockCountTranspose, blockSizeTranspose>>>(dpStreamInfos);
			break;
		case 256:
			huffmanDecodeTransposeKernel<256><<<blockCountTranspose, blockSizeTranspose>>>(dpStreamInfos);
			break;
		default:
			assert(false);
	}
	cudaCheckMsg("huffmanDecodeTransposeKernel execution failed");

	timer();

	pInstance->releaseBuffer();

	return true;
}

}
