#include "HuffmanTable.h"

#include <cassert>
#include <queue>

//#include <Windows.h>

#include <cuda_runtime.h>
//#include <thrust/device_ptr.h>
//#include <thrust/reduce.h>

#include "tools/Operator.h"

#include "cudaUtil.h"

#include "Instance.h"
#include "Histogram.h"
#include "HuffmanDesign.h"

#include "reduce/reduce_app.cui"


namespace cudaCompress {

HuffmanDecodeTable::HuffmanDecodeTable()
	: m_pSymbolTable(0), m_symbolTableSize(0)
{
	cudaSafeCall(cudaMallocHost(&m_pCodewordIndex, CODEWORD_INDEX_SIZE * sizeof(int), cudaHostAllocWriteCombined));
	m_pCodewordFirstIndexPerLength = m_pCodewordIndex;
	m_pCodewordMinPerLength = m_pCodewordIndex + MAX_CODEWORD_BITS;
	m_pCodewordMaxPerLength = m_pCodewordIndex + 2 * MAX_CODEWORD_BITS;
	clear();
}

HuffmanDecodeTable::HuffmanDecodeTable(HuffmanDecodeTable&& other)
{
	// copy state from other
	memcpy(this, &other, sizeof(HuffmanDecodeTable));

	// clear other
	other.m_pCodewordIndex = nullptr;
	other.m_pCodewordFirstIndexPerLength = nullptr;
	other.m_pCodewordMinPerLength = nullptr;
	other.m_pCodewordMaxPerLength = nullptr;
	other.m_pSymbolTable = nullptr;
	other.m_symbolTableSize = 0;
}

HuffmanDecodeTable::~HuffmanDecodeTable()
{
	cudaSafeCall(cudaFreeHost(m_pCodewordIndex));
	delete[] m_pSymbolTable;
}

HuffmanDecodeTable& HuffmanDecodeTable::operator=(HuffmanDecodeTable&& other)
{
	if(this == &other)
		return *this;

	// release our own resources
	cudaSafeCall(cudaFreeHost(m_pCodewordIndex));
	delete[] m_pSymbolTable;

	// copy state from other
	memcpy(this, &other, sizeof(HuffmanDecodeTable));

	// clear other
	other.m_pCodewordIndex = nullptr;
	other.m_pCodewordFirstIndexPerLength = nullptr;
	other.m_pCodewordMinPerLength = nullptr;
	other.m_pCodewordMaxPerLength = nullptr;
	other.m_pSymbolTable = nullptr;
	other.m_symbolTableSize = 0;

	return *this;
}

void HuffmanDecodeTable::clear()
{
	delete[] m_pSymbolTable;
	m_pSymbolTable = nullptr;
	m_symbolTableSize = 0;

	for(uint i = 0; i < CODEWORD_INDEX_SIZE; i++) {
		m_pCodewordIndex[i] = -1;
	}
}

void HuffmanDecodeTable::readFromBitstream(BitStreamReadOnly& bitstream)
{
	clear();

	// read codeword count per length
	std::vector<uint> codewordCountPerLength;
	uint codewordCountPerLengthSize = 0;
	bitstream.readBits(codewordCountPerLengthSize, LOG2_MAX_CODEWORD_BITS);
	codewordCountPerLength.reserve(codewordCountPerLengthSize);
	for(uint i = 0; i < codewordCountPerLengthSize; i++) {
		uint codewordCount = 0;
		bitstream.readBits(codewordCount, LOG2_MAX_DISTINCT_SYMBOLS);
		codewordCountPerLength.push_back(codewordCount);
	}

	// read symbol table
	m_symbolTableSize = 0;
	uint symbolBits = 0;
	bitstream.readBits(m_symbolTableSize, LOG2_MAX_DISTINCT_SYMBOLS);
	bitstream.readBits(symbolBits, LOG2_MAX_SYMBOL_BITS);
	m_pSymbolTable = new Symbol[m_symbolTableSize];
	for(uint i = 0; i < m_symbolTableSize; i++) {
		uint symbol = 0;
		bitstream.readBits(symbol, symbolBits);
		m_pSymbolTable[i] = Symbol(symbol);
	}

	build(codewordCountPerLength);
}

uint HuffmanDecodeTable::computeMaxGPUSize()
{
	uint size = 0;

	size += CODEWORD_INDEX_SIZE * sizeof(int);
	size += MAX_DISTINCT_SYMBOLS * sizeof(Symbol);

	return size;
}

uint HuffmanDecodeTable::computeGPUSize() const
{
	uint size = 0;

	size += CODEWORD_INDEX_SIZE * sizeof(int);
	size += m_symbolTableSize * sizeof(Symbol);

	return size;
}

void HuffmanDecodeTable::uploadToGPU(byte* dpTable) const
{
	size_t size = 0;
	byte* dpTarget = dpTable;

	size = CODEWORD_INDEX_SIZE * sizeof(int);
	cudaSafeCall(cudaMemcpy(dpTarget, m_pCodewordIndex, size, cudaMemcpyHostToDevice));
	dpTarget += size;

	size = m_symbolTableSize * sizeof(Symbol);
	cudaSafeCall(cudaMemcpy(dpTarget, m_pSymbolTable, size, cudaMemcpyHostToDevice));
	dpTarget += size;

	assert(dpTarget - dpTable == computeGPUSize());
}

void HuffmanDecodeTable::uploadToGPUAsync(byte* dpTable) const
{
	size_t size = 0;
	byte* dpTarget = dpTable;

	size = CODEWORD_INDEX_SIZE * sizeof(int);
	cudaSafeCall(cudaMemcpyAsync(dpTarget, m_pCodewordIndex, size, cudaMemcpyHostToDevice));
	dpTarget += size;

	size = m_symbolTableSize * sizeof(Symbol);
	cudaSafeCall(cudaMemcpyAsync(dpTarget, m_pSymbolTable, size, cudaMemcpyHostToDevice));
	dpTarget += size;

	assert(dpTarget - dpTable == computeGPUSize());
}

void HuffmanDecodeTable::build(const std::vector<uint>& codewordCountPerLength)
{
	if(m_symbolTableSize == 0)
		return;

	// count total number of codewords
	uint codewordCount = 0;
	for(uint i = 0; i < codewordCountPerLength.size(); i++) {
		codewordCount += codewordCountPerLength[i];
	}
	if(codewordCountPerLength.empty()) {
		// this can happen when all symbols are the same -> only a single "codeword" with length 0
		codewordCount++;
	}

	assert(codewordCount == m_symbolTableSize);

	// find codeword lengths
	std::vector<uint> codewordLength;
	codewordLength.reserve(codewordCount);

	uint codewordIndexLocal = 0; // within current length
	uint codewordLengthCur = 1;
	if(!codewordCountPerLength.empty()) {
		for(uint entry = 0; entry < codewordCount; entry++, codewordIndexLocal++) {
			// exhausted all codewords of this length?
			while(codewordIndexLocal >= codewordCountPerLength[codewordLengthCur - 1]) {
				codewordLengthCur++;
				codewordIndexLocal = 0;
			}

			codewordLength.push_back(codewordLengthCur);
		}
	}

	// find codewords
	std::vector<int> codeword;
	codeword.reserve(codewordCount);

	int codewordCur = 0;
	for(uint entry = 0;;) {
		// assign codeword
		codeword.push_back(codewordCur);
		codewordCur++;
		entry++;

		// done?
		if(entry >= codewordCount)
			break;

		// if next codeword is longer, make room on the right for next bit(s)
		if(codewordLength[entry] > codewordLength[entry-1]) {
			uint lengthDiff = codewordLength[entry] - codewordLength[entry-1];
			codewordCur <<= lengthDiff;
		}
	}

	// build indices (by codeword length) into table
	uint codewordMaxLength = uint(codewordCountPerLength.size());
	assert(codewordMaxLength <= MAX_CODEWORD_BITS);
	// loop over codeword lengths (actually (length-1))
	for(uint codewordLength = 0, entry = 0; codewordLength < codewordMaxLength; codewordLength++) {
		if(codewordCountPerLength[codewordLength] > 0) {
			// current entry is first codeword of this length
			m_pCodewordFirstIndexPerLength[codewordLength] = entry;
			// store value of first codeword of this length
			m_pCodewordMinPerLength[codewordLength] = codeword[entry];
			// move to last codeword of this length
			entry += codewordCountPerLength[codewordLength] - 1;
			// store value of last codeword of this length
			m_pCodewordMaxPerLength[codewordLength] = codeword[entry];
			// move to first codeword of next length
			entry++;
		} else {
			m_pCodewordFirstIndexPerLength[codewordLength] = -1;
			m_pCodewordMinPerLength[codewordLength] = -1;
			m_pCodewordMaxPerLength[codewordLength] = -1;
		}
	}
}


//////////////////////////////////////////////////////////////////////////////


static uint getTableCountMax(const Instance* pInstance)
{
	bool is3D = (pInstance->m_sizeMax.z() > 1);

	uint countMax = (is3D ? 7 : 4);
	countMax *= max(1u, pInstance->m_channelCountMax);
	return countMax;
}

size_t HuffmanEncodeTable::getRequiredMemory(const Instance* pInstance)
{
	uint tableCountMax = getTableCountMax(pInstance);

	size_t size = 0;

	// dpHistograms
	for(uint i = 0; i < tableCountMax; i++) {
		size += getAlignedSize(MAX_DISTINCT_SYMBOLS * sizeof(uint), 128);
	}

	// dpReduceOut
	size += getAlignedSize(tableCountMax * sizeof(ushort), 128);

	return size;
}

void HuffmanEncodeTable::init(Instance* pInstance)
{
	uint tableCountMax = getTableCountMax(pInstance);

	cudaSafeCall(cudaMallocHost(&pInstance->HuffmanTable.pReadback, tableCountMax * MAX_DISTINCT_SYMBOLS * sizeof(uint)));

	pInstance->HuffmanTable.pSyncEvents.resize(tableCountMax);
	for(uint i = 0; i < tableCountMax; i++) {
		cudaSafeCall(cudaEventCreateWithFlags(&pInstance->HuffmanTable.pSyncEvents[i], cudaEventDisableTiming));
	}
}

void HuffmanEncodeTable::shutdown(Instance* pInstance)
{
	for(uint i = 0; i < pInstance->HuffmanTable.pSyncEvents.size(); i++) {
		cudaSafeCall(cudaEventDestroy(pInstance->HuffmanTable.pSyncEvents[i]));
	}
	pInstance->HuffmanTable.pSyncEvents.clear();

	cudaSafeCall(cudaFreeHost(pInstance->HuffmanTable.pReadback));
	pInstance->HuffmanTable.pReadback = nullptr;
}


HuffmanEncodeTable::HuffmanEncodeTable()
	: m_symbolMax(0), m_codewordTableSize(0)
{
	uint codewordTableSizeMax = MAX_DISTINCT_SYMBOLS;
	cudaSafeCall(cudaMallocHost(&m_pCodewords,       codewordTableSizeMax * sizeof(uint), cudaHostAllocWriteCombined));
	cudaSafeCall(cudaMallocHost(&m_pCodewordLengths, codewordTableSizeMax * sizeof(uint), cudaHostAllocWriteCombined));
}

HuffmanEncodeTable::HuffmanEncodeTable(HuffmanEncodeTable&& other)
{
	m_symbolMax = other.m_symbolMax;
	m_symbols.swap(other.m_symbols);
	m_codewordCountPerLength.swap(other.m_codewordCountPerLength);

	m_pCodewords = other.m_pCodewords;
	other.m_pCodewords = nullptr;
	m_pCodewordLengths = other.m_pCodewordLengths;
	other.m_pCodewordLengths = nullptr;
	m_codewordTableSize = other.m_codewordTableSize;
	other.m_codewordTableSize = 0;
}

HuffmanEncodeTable::~HuffmanEncodeTable()
{
	clear();

	cudaSafeCall(cudaFreeHost(m_pCodewords));
	cudaSafeCall(cudaFreeHost(m_pCodewordLengths));
}


HuffmanEncodeTable& HuffmanEncodeTable::operator=(HuffmanEncodeTable&& other)
{
	if(this == &other)
		return *this;

	m_symbolMax = other.m_symbolMax;
	m_symbols.swap(other.m_symbols);
	m_codewordCountPerLength.swap(other.m_codewordCountPerLength);

	cudaSafeCall(cudaFreeHost(m_pCodewords));
	m_pCodewords = other.m_pCodewords;
	other.m_pCodewords = nullptr;
	cudaSafeCall(cudaFreeHost(m_pCodewordLengths));
	m_pCodewordLengths = other.m_pCodewordLengths;
	other.m_pCodewordLengths = nullptr;
	m_codewordTableSize = other.m_codewordTableSize;
	other.m_codewordTableSize = 0;

	return *this;
}


void HuffmanEncodeTable::clear()
{
	m_codewordTableSize = 0;

	m_symbolMax = 0;
	m_symbols.clear();
	m_codewordCountPerLength.clear();
}


bool HuffmanEncodeTable::design(Instance* pInstance, HuffmanEncodeTable* pTables, uint tableCount, const Symbol** pdpSymbolStreams, uint* pSymbolCountPerStream)
{
	std::vector<uint*> dpHistograms(tableCount);
	for(uint i = 0; i < tableCount; i++) {
		dpHistograms[i] = pInstance->getBuffer<uint>(MAX_DISTINCT_SYMBOLS);
	}
	ushort* dpReduceOut = pInstance->getBuffer<ushort>(tableCount);


	for(uint i = 0; i < tableCount; i++)
		pTables[i].clear();

	// find max symbol
	Symbol symbolMax = 0;
	for(uint i = 0; i < tableCount; i++) {
		if(pSymbolCountPerStream[i] == 0) {
			cudaSafeCall(cudaMemsetAsync(dpReduceOut + i, 0, sizeof(ushort)));
		} else if(pSymbolCountPerStream[i] == 1) {
			// apparently, cudppReduce produces corrupt results for size==1 ?!
			// 17.2.2012, CUDA 4.0
			cudaSafeCall(cudaMemcpyAsync(dpReduceOut + i, pdpSymbolStreams[i], sizeof(ushort), cudaMemcpyDeviceToDevice));
		} else {
			//thrust::device_ptr<const Symbol> pSymbolStreamThrust(pdpSymbolStreams[i]);
			//pTables[i].m_symbolMax = thrust::reduce(pSymbolStreamThrust, pSymbolStreamThrust + pSymbolCountPerStream[i], uint(0), thrust::maximum<Symbol>());
			//cudaCheckMsg("HuffmanEncodeTable::design: Error in thrust::reduce");
			reduceArray<ushort, OperatorMax<ushort>>(dpReduceOut + i, pdpSymbolStreams[i], pSymbolCountPerStream[i], pInstance->m_pReducePlan);
			cudaCheckMsg("HuffmanEncodeTable::design: Error in reduceArray");
		}
	}
	Symbol* pTableSymbolMax = (Symbol*)pInstance->HuffmanTable.pReadback;
	cudaSafeCall(cudaMemcpy(pTableSymbolMax, dpReduceOut, tableCount * sizeof(ushort), cudaMemcpyDeviceToHost));
	for(uint i = 0; i < tableCount; i++) {
		pTables[i].m_symbolMax = pTableSymbolMax[i];
		symbolMax = max(symbolMax, pTables[i].m_symbolMax);
	}

	//cudaEvent_t s, e;
	//cudaEventCreate(&s);
	//cudaEventCreate(&e);
	//cudaEventRecord(s);

	// find symbol probabilities
	uint distinctSymbolCount = symbolMax + 1;
	if(distinctSymbolCount > MAX_DISTINCT_SYMBOLS) {
		//TODO maybe clamp values instead of failing?
		printf("WARNING: distinctSymbolCount == %u > %u, huffman table design failed.\n", distinctSymbolCount, MAX_DISTINCT_SYMBOLS);
		pInstance->releaseBuffers(tableCount + 1);
		return false;
	}
	assert(distinctSymbolCount <= MAX_DISTINCT_SYMBOLS);
	histogram(pInstance, dpHistograms.data(), tableCount, pdpSymbolStreams, pSymbolCountPerStream, distinctSymbolCount);

	//cudaEventRecord(e);
	//cudaEventSynchronize(e);
	//float ms;
	//cudaEventElapsedTime(&ms, s, e);
	//static float acc = 0.0f;
	//acc+=ms;
	//printf("hist: %.2f ms\n", acc);
	//cudaEventDestroy(s);
	//cudaEventDestroy(e);

	//LARGE_INTEGER start;
	//QueryPerformanceCounter(&start);

	for(uint i = 0; i < tableCount; i++) {
		// download symbol probabilities (asynchronously)
		uint distinctSymbolCountThisTable = pTables[i].m_symbolMax + 1;
		cudaSafeCall(cudaMemcpyAsync(pInstance->HuffmanTable.pReadback + i * MAX_DISTINCT_SYMBOLS, dpHistograms[i], distinctSymbolCountThisTable * sizeof(uint), cudaMemcpyDeviceToHost));
		// record event so we can wait for completion of the download later
		cudaSafeCall(cudaEventRecord(pInstance->HuffmanTable.pSyncEvents[i]));
	}

	for(uint i = 0; i < tableCount; i++) {
		// wait for async download to complete
		cudaSafeCall(cudaEventSynchronize(pInstance->HuffmanTable.pSyncEvents[i]));

		// build actual encode table
		uint distinctSymbolCountThisTable = pTables[i].m_symbolMax + 1;
		pTables[i].build(pInstance->HuffmanTable.pReadback + i * MAX_DISTINCT_SYMBOLS, distinctSymbolCountThisTable);
	}

	//LARGE_INTEGER end, freq;
	//QueryPerformanceCounter(&end);
	//QueryPerformanceFrequency(&freq);

	//float t = 1000.0f * float(end.QuadPart - start.QuadPart) / float(freq.QuadPart);
	//printf("Time Huffman Design: %.2f ms\n", t);
	
	pInstance->releaseBuffers(tableCount + 1);

	return true;
}

void HuffmanEncodeTable::uploadToGPU(uint* dpCodewords, uint* dpCodewordLengths) const
{
	size_t size = m_codewordTableSize * sizeof(uint);
	cudaSafeCall(cudaMemcpy(dpCodewords,       m_pCodewords,       size, cudaMemcpyHostToDevice));
	cudaSafeCall(cudaMemcpy(dpCodewordLengths, m_pCodewordLengths, size, cudaMemcpyHostToDevice));
}

void HuffmanEncodeTable::uploadToGPUAsync(uint* dpCodewords, uint* dpCodewordLengths) const
{
	size_t size = m_codewordTableSize * sizeof(uint);
	cudaSafeCall(cudaMemcpyAsync(dpCodewords,       m_pCodewords,       size, cudaMemcpyHostToDevice));
	cudaSafeCall(cudaMemcpyAsync(dpCodewordLengths, m_pCodewordLengths, size, cudaMemcpyHostToDevice));
}

void HuffmanEncodeTable::writeToBitstream(BitStream& bitstream) const
{
	// compute needed bits per symbol
	Symbol symbolMax = m_symbolMax;
	uint symbolBits = 0;
	while(symbolMax > 0) {
		symbolMax >>= 1;
		symbolBits++;
	}

	// write
	bitstream.writeBits(uint(m_codewordCountPerLength.size()), LOG2_MAX_CODEWORD_BITS);
	for(uint i = 0; i < m_codewordCountPerLength.size(); i++) {
		bitstream.writeBits(m_codewordCountPerLength[i], LOG2_MAX_DISTINCT_SYMBOLS);
	}
	bitstream.writeBits(uint(m_symbols.size()), LOG2_MAX_DISTINCT_SYMBOLS);
	bitstream.writeBits(symbolBits, LOG2_MAX_SYMBOL_BITS);
	for(uint i = 0; i < m_symbols.size(); i++) {
		bitstream.writeBits(m_symbols[i], symbolBits);
	}
}

void HuffmanEncodeTable::build(const uint* pSymbolProbabilities, uint distinctSymbolCount)
{
	std::vector<HuffmanTreeNode> huffmanNodes(2 * distinctSymbolCount - 1);
	uint nextNodeIndex = 0;

	// build list of all used symbols, packed in HuffmanTreeNodes
	// these will be the leaves of the huffman tree
	std::vector<HuffmanTreeNode*> treeLeaves;
	for(uint symbol = 0; symbol < distinctSymbolCount; symbol++) {
		if(pSymbolProbabilities[symbol] > 0) {
			huffmanNodes[nextNodeIndex].init(symbol, pSymbolProbabilities[symbol]);
			treeLeaves.push_back(&huffmanNodes[nextNodeIndex]);
			nextNodeIndex++;
		}
	}

	if(treeLeaves.empty())
		return;

	// list of huffman nodes to process
	std::priority_queue<HuffmanTreeNode*, std::vector<HuffmanTreeNode*>, HuffmanTreeNodeProbabilityInvComparer> treeNodesTodo(treeLeaves.begin(), treeLeaves.end());

	// build the huffman tree by successively combining the lowest-probability nodes
	while(treeNodesTodo.size() > 1) {
		uint newNodeIndex = nextNodeIndex++;
		HuffmanTreeNode& newNode = huffmanNodes[newNodeIndex];

		newNode.init(INVALID_SYMBOL, 0);

		// get nodes with lowest probability as children
		HuffmanTreeNode* pLeftChild = treeNodesTodo.top();
		treeNodesTodo.pop();
		HuffmanTreeNode* pRightChild = treeNodesTodo.top();
		treeNodesTodo.pop();

		newNode.m_pLeftChild = pLeftChild;
		newNode.m_pRightChild = pRightChild;

		// combine probabilities
		newNode.m_probability = pLeftChild->m_probability + pRightChild->m_probability;

		// insert into todo list
		treeNodesTodo.push(&newNode);
	}

	HuffmanTreeNode& rootNode = *treeNodesTodo.top();

	// assign codeword length = tree level
	int codewordLengthMax = rootNode.assignCodewordLength(0);

	// sort leaves (ie actual symbols) by codeword length
	std::sort(treeLeaves.begin(), treeLeaves.end(), HuffmanTreeNodeCodewordLengthComparer());

	// fill codeword count list and symbol list from leaves list
	m_codewordCountPerLength.resize(codewordLengthMax);
	m_symbols.resize(treeLeaves.size());
	for(uint i = 0; i < treeLeaves.size(); i++) {
		const HuffmanTreeNode& node = *treeLeaves[i];

		if(node.m_codewordLength > 0)
			m_codewordCountPerLength[node.m_codewordLength - 1]++;
		m_symbols[i] = node.m_symbol;
	}

	// count total number of codewords
	uint codewordCount = 0;
	for(uint i = 0; i < m_codewordCountPerLength.size(); i++) {
		codewordCount += m_codewordCountPerLength[i];
	}
	if(m_codewordCountPerLength.empty()) {
		// this can happen when all symbols are the same -> only a single "codeword" with length 0
		codewordCount++;
	}

	// find codeword lengths (ordered by codeword index)
	uint* pCodewordLengthsByIndex = new uint[codewordCount];
	uint codewordIndexLocal = 0; // within current length
	uint codewordLength = 1;
	if(!m_codewordCountPerLength.empty()) {
		for(uint entry = 0; entry < codewordCount; entry++, codewordIndexLocal++) {
			// exhausted all codewords of this length?
			while(codewordIndexLocal >= m_codewordCountPerLength[codewordLength - 1]) {
				codewordLength++;
				codewordIndexLocal = 0;
			}

			pCodewordLengthsByIndex[entry] = codewordLength;
		}
	} else {
		pCodewordLengthsByIndex[0] = 0;
	}

	// find codewords (ordered by codeword index)
	uint* pCodewordsByIndex = new uint[codewordCount];
	uint codeword = 0;
	for(uint entry = 0;;) {
		// assign codeword
		pCodewordsByIndex[entry] = codeword;
		codeword++;
		entry++;

		// done?
		if(entry >= codewordCount)
			break;

		// if next codeword is longer, make room on the right for next bit(s)
		if(pCodewordLengthsByIndex[entry] > pCodewordLengthsByIndex[entry-1]) {
			uint lengthDiff = pCodewordLengthsByIndex[entry] - pCodewordLengthsByIndex[entry-1];
			codeword <<= lengthDiff;
		}
	}

	m_codewordTableSize = distinctSymbolCount;

	// fill tables with invalid values
	memset(m_pCodewords,        0, m_codewordTableSize * sizeof(uint));
	memset(m_pCodewordLengths, -1, m_codewordTableSize * sizeof(uint));

	// reorder codewords and lengths by symbol
	for(uint entry = 0; entry < codewordCount; entry++) {
		m_pCodewords      [m_symbols[entry]] = pCodewordsByIndex      [entry];
		m_pCodewordLengths[m_symbols[entry]] = pCodewordLengthsByIndex[entry];
	}

	delete[] pCodewordsByIndex;
	delete[] pCodewordLengthsByIndex;
}

}
