#include "Instance.h"

#include <algorithm>
#include <cassert>

#include "util.h"
#include "cudaUtil.h"

#include "reduce/reduce_plan.h"
#include "scan/scan_plan.h"

#include "DWT.h"
#include "Histogram.h"
#include "Huffman.h"
#include "HuffmanTable.h"
#include "RunLength.h"
#include "PackInc.h"
#include "Encode.h"
#include "CompressImage.h"
#include "CompressHeightfield.h"
#include "CompressVolume.h"

using namespace tum3D;


namespace cudaCompress
{

Instance::Instance()
	: m_cudaDevice(-1)
	, m_sizeMax(0, 0, 0)
	, m_channelCountMax(0), m_offsetInterval(0)
	, m_pReducePlan(nullptr), m_pScanPlan(nullptr)
	, m_timingDetail(TIMING_DETAIL_NONE)
	, m_dpBuffer(nullptr), m_bufferSize(0), m_bufferOffset(0)
{
}

Instance::~Instance()
{
	assert(m_cudaDevice == -1);
}

bool Instance::create(int cudaDevice, uint sizeMaxX, uint sizeMaxY, uint sizeMaxZ, uint channelCountMax, uint offsetInterval)
{
	if(cudaDevice >= 0) {
		m_cudaDevice = cudaDevice;
		cudaSafeCall(cudaSetDevice(m_cudaDevice));
	} else {
		// value < 0 means "use current device"
		cudaSafeCall(cudaGetDevice(&m_cudaDevice));
	}

	m_sizeMax = Vec3i(sizeMaxX, sizeMaxY, sizeMaxZ);
	m_channelCountMax = channelCountMax;
	m_offsetInterval = offsetInterval;


	uint elemCountTotal = m_sizeMax.volume();

	Vec3i encodingBlockSize = (m_sizeMax + 1) / 2;
	uint elemCountPerBlock = encodingBlockSize.volume();

	bool is3D = (sizeMaxZ > 1);

	uint encodingBlockCountMax = (is3D ? 7 : 4);
	encodingBlockCountMax *= max(1u, channelCountMax);

	uint offsetCountMax = (elemCountPerBlock + offsetInterval - 1) / offsetInterval;


	m_pScanPlan = new ScanPlan(sizeof(uint), elemCountPerBlock + 1); // "+ 1" for total
	m_pReducePlan = new ReducePlan(sizeof(uint), elemCountTotal);


	size_t sizeTier0 = 0;
	sizeTier0 = max(sizeTier0, runLengthGetRequiredMemory(this));
	sizeTier0 = max(sizeTier0, huffmanGetRequiredMemory(this));
	// HuffmanEncodeTable uses histogram...
	sizeTier0 = max(sizeTier0, HuffmanEncodeTable::getRequiredMemory(this) + histogramGetRequiredMemory(this));
	sizeTier0 = max(sizeTier0, packIncGetRequiredMemory(this));
	size_t sizeTier1 = 0;
	sizeTier1 = max(sizeTier1, encodeGetRequiredMemory(this));
	size_t sizeTier2 = 0;
	sizeTier2 = max(sizeTier2, compressImageGetRequiredMemory(this));
	sizeTier2 = max(sizeTier2, compressHeightfieldGetRequiredMemory(this));
	sizeTier2 = max(sizeTier2, compressVolumeGetRequiredMemory(this));

	m_bufferSize = sizeTier0 + sizeTier1 + sizeTier2;
	//TODO don't use cudaSafeCall, but manually check for out of memory!
	cudaSafeCall(cudaMalloc(&m_dpBuffer, m_bufferSize));


	runLengthInit(this);
	huffmanInit(this);
	HuffmanEncodeTable::init(this);
	histogramInit(this);
	packIncInit(this);

	encodeInit(this);

	compressImageInit(this);
	compressHeightfieldInit(this);
	compressVolumeInit(this);


	return true;
}

void Instance::destroy()
{
	assert(m_bufferOffset == 0);
	assert(m_allocatedSizes.empty());


	cudaSafeCall(cudaFree(m_dpBuffer));
	m_dpBuffer = nullptr;
	m_bufferSize = 0;
	m_bufferOffset = 0;


	compressVolumeShutdown(this);
	compressHeightfieldShutdown(this);
	compressImageShutdown(this);

	encodeShutdown(this);

	packIncShutdown(this);
	runLengthShutdown(this);
	HuffmanEncodeTable::shutdown(this);
	huffmanShutdown(this);
	histogramShutdown(this);


	delete m_pReducePlan;
	m_pReducePlan = nullptr;

	delete m_pScanPlan;
	m_pScanPlan = nullptr;


	m_sizeMax = Vec3i(0, 0, 0);
	m_channelCountMax = 0;
	m_offsetInterval = 0;

	m_cudaDevice = -1;
}


void Instance::setTimingDetail(ETimingDetail detail)
{
	m_timingDetail = detail;

	bool enableTier0 = (m_timingDetail >= TIMING_DETAIL_LOW);
	bool enableTier1 = (m_timingDetail >= TIMING_DETAIL_MEDIUM);
	bool enableTier2 = (m_timingDetail >= TIMING_DETAIL_HIGH);

	CompressImage.timerEncode.setEnabled(enableTier0);
	CompressImage.timerDecode.setEnabled(enableTier0);
	CompressHeightfield.timerEncode.setEnabled(enableTier0);
	CompressHeightfield.timerDecode.setEnabled(enableTier0);
	CompressVolume.timerEncode.setEnabled(enableTier0);
	CompressVolume.timerDecode.setEnabled(enableTier0);

	Encode.timerEncode.setEnabled(enableTier1);
	Encode.timerDecode.setEnabled(enableTier1);

	Huffman.timerEncode.setEnabled(enableTier2);
	Huffman.timerDecode.setEnabled(enableTier2);
	RunLength.timerEncode.setEnabled(enableTier2);
	RunLength.timerDecode.setEnabled(enableTier2);
}

void Instance::printTimings()
{
	std::vector<std::string> eventNames;
	std::vector<float>       eventTimes;

	bool enableTier0 = (m_timingDetail >= TIMING_DETAIL_LOW);
	bool enableTier1 = (m_timingDetail >= TIMING_DETAIL_MEDIUM);
	bool enableTier2 = (m_timingDetail >= TIMING_DETAIL_HIGH);

	if(enableTier0)
	{
		CompressImage.timerEncode.getAccumulatedTimes(eventNames, eventTimes, true);
		CompressImage.timerDecode.getAccumulatedTimes(eventNames, eventTimes, true);
		CompressHeightfield.timerEncode.getAccumulatedTimes(eventNames, eventTimes, true);
		CompressHeightfield.timerDecode.getAccumulatedTimes(eventNames, eventTimes, true);
		CompressVolume.timerEncode.getAccumulatedTimes(eventNames, eventTimes, true);
		CompressVolume.timerDecode.getAccumulatedTimes(eventNames, eventTimes, true);
	}

	if(enableTier1)
	{
		Encode.timerEncode.getAccumulatedTimes(eventNames, eventTimes, true);
		Encode.timerDecode.getAccumulatedTimes(eventNames, eventTimes, true);
	}

	if(enableTier2)
	{
		Huffman.timerEncode.getAccumulatedTimes(eventNames, eventTimes, true);
		Huffman.timerDecode.getAccumulatedTimes(eventNames, eventTimes, true);
		RunLength.timerEncode.getAccumulatedTimes(eventNames, eventTimes, true);
		RunLength.timerDecode.getAccumulatedTimes(eventNames, eventTimes, true);
	}


	size_t eventCount = min(eventNames.size(), eventTimes.size());

	uint nameLengthMax = 0;
	for(uint i = 0; i < eventCount; i++) {
		nameLengthMax = max(nameLengthMax, (uint)eventNames[i].length());
	}
	for(uint i = 0; i < eventCount; i++) {
		printf("%-*s: %6.2f ms\n", nameLengthMax+1, eventNames[i].c_str(), eventTimes[i]);
	}
}

void Instance::resetTimings()
{
	CompressImage.timerEncode.reset();
	CompressImage.timerDecode.reset();
	CompressHeightfield.timerEncode.reset();
	CompressHeightfield.timerDecode.reset();
	CompressVolume.timerEncode.reset();
	CompressVolume.timerDecode.reset();

	Encode.timerEncode.reset();
	Encode.timerDecode.reset();

	Huffman.timerEncode.reset();
	Huffman.timerDecode.reset();
	RunLength.timerEncode.reset();
	RunLength.timerDecode.reset();
}


byte* Instance::getByteBuffer(size_t bytes)
{
	assert(m_bufferOffset + bytes <= m_bufferSize);
	if(m_bufferOffset + bytes > m_bufferSize) {
		printf("ERROR: Instance::getByteBuffer: out of memory!\n");
		return nullptr;
	}

	byte* dpResult = m_dpBuffer + m_bufferOffset;
	m_allocatedSizes.push_back(bytes);
	m_bufferOffset += getAlignedSize(bytes, 128);

	return dpResult;
}

void Instance::releaseBuffer()
{
	assert(!m_allocatedSizes.empty());
	if(m_allocatedSizes.empty()) {
		printf("ERROR: Instance::releaseBuffer: no more buffers to release\n");
		return;
	}

	size_t lastSize = m_allocatedSizes.back();
	m_allocatedSizes.pop_back();

	m_bufferOffset -= getAlignedSize(lastSize, 128);
	assert(m_bufferOffset % 128 == 0);
}

void Instance::releaseBuffers(uint bufferCount)
{
	for(uint i = 0; i < bufferCount; i++) {
		releaseBuffer();
	}
}

}
