#ifndef __TUM3D_CUDACOMPRESS__INSTANCE_H__
#define __TUM3D_CUDACOMPRESS__INSTANCE_H__


#include "global.h"

#include <vector>

#include <cuda_runtime.h>

#include "tools/CudaTimer.h"
#include "tools/Vec.h"

#include "HuffmanTable.h"
#include "Timing.h"


class ReducePlan;
class ScanPlan;


namespace cudaCompress
{

class Instance
{
public:
	Instance();
	~Instance();

	// pass -1 as the cudaDevice to use the current one
	bool create(int cudaDevice, uint sizeMaxX, uint sizeMaxY, uint sizeMaxZ, uint channelCountMax, uint offsetIntervalMin);
	void destroy();

	void setTimingDetail(ETimingDetail detail);
	ETimingDetail getTimingDetail() const { return m_timingDetail; }
	void printTimings();
	void resetTimings();

	// get a buffer of the specified size in GPU memory
	byte* getByteBuffer(size_t bytes);
	template<typename T>
	T* getBuffer(size_t count) { return (T*)getByteBuffer(count * sizeof(T)); }
	// release the last buffer returned from getBuffer
	void releaseBuffer();
	void releaseBuffers(uint bufferCount);


	int m_cudaDevice;

	tum3D::Vec3i m_sizeMax;
	uint m_channelCountMax;
	uint m_offsetInterval;


	ReducePlan* m_pReducePlan;
	ScanPlan*   m_pScanPlan;


	// TIER 0
	struct CompressImageResources
	{
		byte* pUpload;
		cudaEvent_t syncEventUpload;

		CudaTimerResources timerEncode;
		CudaTimerResources timerDecode;

		CompressImageResources()
			: pUpload(nullptr), syncEventUpload(0) {}
	} CompressImage;

	struct CompressHeightfieldResources
	{
		uint* pBitsUpload;
		uint* pBitsDownload;

		cudaEvent_t syncEventUpload;
		cudaEvent_t syncEventDownload;

		CudaTimerResources timerEncode;
		CudaTimerResources timerDecode;

		CompressHeightfieldResources()
			: pBitsUpload(nullptr), pBitsDownload(nullptr)
			, syncEventUpload(0), syncEventDownload(0) {}
	} CompressHeightfield;

	struct CompressVolumeResources
	{
		byte* pUpload;
		cudaEvent_t syncEventUpload;

		CudaTimerResources timerEncode;
		CudaTimerResources timerDecode;

		CompressVolumeResources()
			: pUpload(nullptr), syncEventUpload(0) {}
	} CompressVolume;


	// TIER 1
	struct EncodeResources
	{
		std::vector<uint*> ppCodewordBuffers;
		std::vector<uint*> ppOffsetBuffers;

		std::vector<HuffmanEncodeTable> symbolEncodeTables;
		std::vector<HuffmanEncodeTable> zeroCountEncodeTables;

		std::vector<HuffmanDecodeTable> symbolDecodeTables;
		std::vector<HuffmanDecodeTable> zeroCountDecodeTables;

		std::vector<cudaEvent_t> pSyncEvents;

		byte* pUpload0;
		byte* pUpload1;

		cudaEvent_t encodeFinishedEvent;

		CudaTimerResources timerEncode;
		CudaTimerResources timerDecode;

		EncodeResources()
			: pUpload0(nullptr), pUpload1(nullptr), encodeFinishedEvent(0) {}
	} Encode;


	// TIER 2
	struct HistogramResources
	{
		byte* pUpload;
		cudaEvent_t syncEvent;

		HistogramResources()
			: pUpload(nullptr), syncEvent(0) {}
	} Histogram;

	struct HuffmanResources
	{
		uint* pReadback;
		cudaEvent_t syncEventReadback;

		CudaTimerResources timerEncode;
		CudaTimerResources timerDecode;

		HuffmanResources()
			: pReadback(nullptr), syncEventReadback(0) {}
	} Huffman;

	struct HuffmanTableResources
	{
		uint* pReadback;

		std::vector<cudaEvent_t> pSyncEvents;

		HuffmanTableResources()
			: pReadback(nullptr) {}
	} HuffmanTable;

	struct RunLengthResources
	{
		uint* pReadback;
		cudaEvent_t syncEvent;

		CudaTimerResources timerEncode;
		CudaTimerResources timerDecode;

		RunLengthResources()
			: pReadback(nullptr), syncEvent(0) {}
	} RunLength;

private:
	ETimingDetail m_timingDetail;

	byte* m_dpBuffer;
	size_t m_bufferSize;

	size_t m_bufferOffset;
	std::vector<size_t> m_allocatedSizes;
};

}


#endif
