#include "RiceGolombCPU.h"

#include <cmath>

#include "util.h"


namespace cudaCompress {

static const uint GLIMIT = 32; // maximum allowable codeword length

void encodeSymbol(BitStream& bitStream, Symbol symbol, uint k, uint qmax)
{
	uint m = 1 << k;
	uint q = symbol / m;

	if(q < GLIMIT - qmax - 1) {
		// regular rice-golomb coding

		uint r = symbol - q * m;

		// write q zeros followed by a single one
		while(q > 31) {
			bitStream.writeBits(0, 32);
			q -= 32;
		}
		bitStream.writeBits(1, q + 1);

		// write r in k bits
		bitStream.writeBits(r, k);
	} else {
		// length-limited mode

		// write glimit-qmax-1 zeros followed by a single one
		uint zeros = GLIMIT - qmax - 1;
		while(zeros > 31) {
			bitStream.writeBits(0, 32);
			zeros -= 32;
		}
		bitStream.writeBits(1, zeros + 1);

		// write symbol in qmax bits
		bitStream.writeBits(symbol - 1, qmax);
	}
}

Symbol decodeSymbol(BitStreamReadOnly& bitStream, uint k, uint qmax)
{
	// read until first one bit; number of zeros == q
	uint q = 0;
	uint temp = 0;
	bitStream.readBit(temp);
	while(!temp) {
		++q;
		bitStream.readBit(temp);
	}

	if(q == GLIMIT - qmax - 1) {
		// length-limited mode

		// read symbol
		uint val = 0;
		bitStream.readBits(val, qmax);
		return Symbol(val + 1);
	} else {
		// regular golomb-rice coding

		// read r (k bits)
		uint r = 0;
		bitStream.readBits(r, k);

		uint m = 1 << k;
		return Symbol(q * m + r);
	}
}

bool riceGolombEncodeCPU(BitStream& bitStream, const std::vector<Symbol>& symbolStream, uint kBlockSize, std::vector<uint>& offsets, uint offsetIntervalInBlocks)
{
	uint offsetBase = bitStream.getBitPosition();
	for(uint block = 0, blockOffset = 0; blockOffset < symbolStream.size(); ++block, blockOffset += kBlockSize) {
		if(offsetIntervalInBlocks > 0 && block % offsetIntervalInBlocks == 0) {
			offsets.push_back(bitStream.getBitPosition() - offsetBase);
		}

		uint blockSizeCur = min(blockOffset + kBlockSize, (uint)symbolStream.size()) - blockOffset;

		// compute "reasonable" k
		uint64 symbolSum = 0;
		Symbol symbolMax = 0;
		for(uint index = blockOffset; index < blockOffset + blockSizeCur; index++) {
			symbolSum += symbolStream[index];
			symbolMax = max(symbolMax, symbolStream[index]);
		}
		double symbolAvg = double(symbolSum) / double(blockSizeCur);
		uint k = (uint)clamp(int(ceil(log(0.5 * symbolAvg) / log(2.0))), 0, 15);
		bitStream.writeBits(k, 4);

		// number of bits in (symbolMax-1)
		uint qmax = getRequiredBits(max(uint(symbolMax), 1u) - 1u);
		bitStream.writeBits(qmax, 4);

		// encode symbols
		for(uint index = blockOffset; index < blockOffset + blockSizeCur; index++) {
			encodeSymbol(bitStream, symbolStream[index], k, qmax);
		}
	}

	return true;
}

bool riceGolombDecodeCPU(BitStreamReadOnly& bitStream, uint symbolCount, std::vector<Symbol>& symbolStream, uint kBlockSize)
{
	for(uint offset = 0; offset < symbolCount; offset += kBlockSize) {
		uint blockSizeCur = min(offset + kBlockSize, symbolCount) - offset;

		// read k and qmax
		uint k = 0;
		bitStream.readBits(k, 4);
		uint qmax = 0;
		bitStream.readBits(qmax, 4);

		// decode symbols
		for(uint index = offset; index < offset + blockSizeCur; index++) {
			symbolStream.push_back(decodeSymbol(bitStream, k, qmax));
		}
	}

	return true;
}

}
