#include "RunLength.h"

#include <cassert>
#include <string>

#include <cuda_runtime.h>

//#include <thrust/device_ptr.h>
//#include <thrust/scan.h>

#include "tools/Vec.h"

#include "cudaUtil.h"
#include "util.h"

#include "Instance.h"

using namespace tum3D;


#include "scan/scan_app.cui"

#include "RunLengthKernels.cui"


namespace cudaCompress {


static uint getStreamCountMax(const Instance* pInstance)
{
	bool is3D = (pInstance->m_sizeMax.z() > 1);

	uint countMax = (is3D ? 7 : 4);
	countMax *= max(1u, pInstance->m_channelCountMax);
	return countMax;
}

static uint getSymbolCountPerStreamMax(const Instance* pInstance)
{
	Vec3i blockSizeMax = (pInstance->m_sizeMax + 1) / 2;
	return blockSizeMax.volume();
}

size_t runLengthGetRequiredMemory(const Instance* pInstance)
{
	uint streamCountMax = getStreamCountMax(pInstance);
	uint symbolCountMax = getSymbolCountPerStreamMax(pInstance);

	size_t size = 0;

	// encode and decode: dpValidSymbolIndices
	size += getAlignedSize(symbolCountMax * sizeof(uint), 128);
	// encode: dpOutputIndices
	size += getAlignedSize((symbolCountMax + 1) * sizeof(uint), 128);
	// encode: dpScanTotal
	size += getAlignedSize(streamCountMax * sizeof(uint), 128);

	return size;
}

bool runLengthInit(Instance* pInstance)
{
	uint streamCountMax = getStreamCountMax(pInstance);
	uint symbolCountMax = getSymbolCountPerStreamMax(pInstance);

	cudaSafeCall(cudaMallocHost(&pInstance->RunLength.pReadback, streamCountMax * sizeof(uint)));

	cudaSafeCall(cudaEventCreateWithFlags(&pInstance->RunLength.syncEvent, cudaEventDisableTiming));

	return true;
}

bool runLengthShutdown(Instance* pInstance)
{
	cudaSafeCall(cudaEventDestroy(pInstance->RunLength.syncEvent));
	pInstance->RunLength.syncEvent = 0;

	cudaSafeCall(cudaFreeHost(pInstance->RunLength.pReadback));
	pInstance->RunLength.pReadback = nullptr;

	return true;
}

bool runLengthEncode(Instance* pInstance, Symbol** pdpSymbolsCompact, ushort** pdpZeroCounts, const Symbol** pdpSymbols, const uint* pSymbolCount, uint streamCount, uint zeroCountMax, uint* pSymbolCountCompact)
{
	uint streamCountMax = getStreamCountMax(pInstance);
	uint symbolCountMax = getSymbolCountPerStreamMax(pInstance);

	assert(streamCount <= streamCountMax);

	uint* dpOutputIndices      = pInstance->getBuffer<uint>(symbolCountMax + 1);
	uint* dpValidSymbolIndices = pInstance->getBuffer<uint>(symbolCountMax);
	uint* dpScanTotal          = pInstance->getBuffer<uint>(streamCountMax);

	uint blockSize = 0;
	uint blockCount = 0;

	for(uint i = 0; i < streamCount; i++) {
		assert(pSymbolCount[i] <= symbolCountMax);

		CudaScopedTimer timer(pInstance->RunLength.timerEncode);

		timer("Scan Valid Flags");

		// run prefix sum on symbol non-zero flags to get output indices
		//TODO ballot scan!
		scanArray<ushort, uint, true, FunctorFlagTrue<ushort, uint>>(dpOutputIndices, pdpSymbols[i], pSymbolCount[i] + 1, pInstance->m_pScanPlan);
		cudaCheckMsg("runLengthEncode: Error in scanArray");

		timer("Get Valid Symbol Indices");

		// last element of outputindices == compact symbol count, start readback
		uint* dpCompactSymbolCount = dpOutputIndices + pSymbolCount[i];
		cudaSafeCall(cudaMemcpyAsync(pInstance->RunLength.pReadback + i, dpCompactSymbolCount, sizeof(uint), cudaMemcpyDeviceToHost));
		cudaSafeCall(cudaEventRecord(pInstance->RunLength.syncEvent));

		// get indices of valid (non-zero) symbols
		blockSize = 384;
		blockCount = min((pSymbolCount[i] + blockSize - 1) / blockSize, 256);
		runLengthEncodeGetValidSymbolIndices<<<blockCount, blockSize>>>(dpOutputIndices, dpValidSymbolIndices, pSymbolCount[i]);
		cudaCheckMsg("runLengthEncodeGetValidSymbolIndices execution failed");

		timer("Get # Extra Zeros");

		// compute number of extra zero symbols to insert in order to respect zeroCountMax
		// choose blockCount based on original (non-compact) symbol count, so we can wait a bit longer before syncing on the download
		blockSize = 384;
		blockCount = min((pSymbolCount/*Compact*/[i] + blockSize - 1) / blockSize, 256);

		runLengthEncodeExtraZeroSymbolCountsKernel<<<blockCount, blockSize>>>(dpValidSymbolIndices, dpOutputIndices, dpCompactSymbolCount, zeroCountMax);
		cudaCheckMsg("runLengthEncodeExtraZeroSymbolCountsKernel execution failed");

		timer("Sync Readback");

		// wait for download of compacted symbol count - need it for the next scan
		cudaSafeCall(cudaEventSynchronize(pInstance->RunLength.syncEvent));
		pSymbolCountCompact[i] = pInstance->RunLength.pReadback[i];

		timer("Scan # Extra Zeros");

		// run prefix sum on extra zero symbol counts to get output offsets
		//thrust::device_ptr<uint> pWriteOffsetsThrust(dpOutputIndices);
		//thrust::exclusive_scan(pWriteOffsetsThrust, pWriteOffsetsThrust + pSymbolCountCompact[i] + 1, pWriteOffsetsThrust);
		//cudaCheckMsg("runLengthEncode: Error in thrust::exclusive_scan");
		scanArray<uint, uint, true>(dpOutputIndices, dpOutputIndices, pSymbolCountCompact[i] + 1, pInstance->m_pScanPlan);
		cudaCheckMsg("runLengthEncode: Error in scanArray");

		timer("Download # Extra Zeros");

		// last write offset == total number of extra zeroes to be inserted
		cudaSafeCall(cudaMemcpyAsync(dpScanTotal + i, dpOutputIndices + pSymbolCountCompact[i], sizeof(uint), cudaMemcpyDeviceToDevice));
		// if this was the last stream, start readback to cpu
		if(i == streamCount - 1) {
			cudaSafeCall(cudaMemcpyAsync(pInstance->RunLength.pReadback, dpScanTotal, streamCount * sizeof(uint), cudaMemcpyDeviceToHost));
			cudaSafeCall(cudaEventRecord(pInstance->RunLength.syncEvent));
		}

		// if there are no non-zero symbols, we can bail out here
		if(pSymbolCountCompact[i] == 0) {
			continue;
		}

		timer("Compact");

		// copy non-zero symbols to output, pad with extra zero symbols where necessary
		blockSize = 384;
		blockCount = min((pSymbolCountCompact[i] + blockSize - 1) / blockSize, 256);

		//cudaMemset(pdpSymbolsCompact[i], 0, sizeof(Symbol)*pSymbolCount[i]);
		runLengthEncodeCompactKernel<<<blockCount, blockSize>>>(pdpSymbols[i], dpValidSymbolIndices, dpOutputIndices, pdpSymbolsCompact[i], pdpZeroCounts[i], pSymbolCountCompact[i], zeroCountMax);
		cudaCheckMsg("runLengthEncodeCompactKernel execution failed");
	}

	// add extra zeros to compacted symbol count
	cudaSafeCall(cudaEventSynchronize(pInstance->RunLength.syncEvent));
	for(uint i = 0; i < streamCount; i++) {
		pSymbolCountCompact[i] += pInstance->RunLength.pReadback[i];
	}

	pInstance->releaseBuffers(3);

	return true;
}

bool runLengthDecode(Instance* pInstance, const Symbol** pdpSymbolsCompact, const ushort** pdpZeroCounts, const uint* pSymbolCountCompact, Symbol** pdpSymbols, const uint* pSymbolCount, uint streamCount)
{
	uint streamCountMax = getStreamCountMax(pInstance);
	uint symbolCountMax = getSymbolCountPerStreamMax(pInstance);

	uint* dpValidSymbolIndices = pInstance->getBuffer<uint>(symbolCountMax);

	for(uint i = 0; i < streamCount; i++) {
		CudaScopedTimer timer(pInstance->RunLength.timerDecode);

		timer("Memset Zeros");

		// set all symbols to 0
		cudaSafeCall(cudaMemsetAsync(pdpSymbols[i], 0, pSymbolCount[i] * sizeof(Symbol)));

		timer("Scan Zero Counts");

		// if there are no symbols, we're done here
		if(pSymbolCountCompact[i] == 0) {
			//if(timings) updateTimings(g_decodeEvent, g_decodeEventTime, nextEvent-1);
			continue;
		}

		// run prefix sum on zero counts to get valid symbol indices
		scanArray<ushort, uint, false>(dpValidSymbolIndices, pdpZeroCounts[i], pSymbolCountCompact[i], pInstance->m_pScanPlan);
		cudaCheckMsg("runLengthDecode: Error in scanArray");

		timer("Scatter Symbols");

		// expand symbol stream - scattered write of non-zero symbols
		uint blockSize = 384;
		uint blockCount = min((pSymbolCountCompact[i] + blockSize - 1) / blockSize, 256);

		runLengthDecodeKernel<<<blockCount, blockSize>>>(pdpSymbolsCompact[i], dpValidSymbolIndices, pSymbolCountCompact[i], pdpSymbols[i]);
		cudaCheckMsg("runLengthDecodeKernel execution failed");
	}

	pInstance->releaseBuffer();

	return true;
}

}
