#include "global.h"

#include <fstream>
#include <string>
#include <vector>

#include <cuda_runtime.h>
#include <cuda_profiler_api.h>

#include "tthread/tinythread.h"

#include "tools/entropy.h"
#include "tools/imgtools.h"
#include "tools/rawfile.h"
#include "tools/PinnedAllocator.h"
#include "tools/Vec.h"

#include "util.h"

#include "Init.h"
#include "Bits.h"
#include "CompressImage.h"
#include "CompressHeightfield.h"
#include "CompressVolume.h"
#include "Huffman.h"
#include "HuffmanTable.h"
#include "RunLength.h"
#include "Encode.h"
#include "Quantize.h"
#include "YCoCg.h"
#include "YCoCgCPU.h"
#include "DWT.h"
#include "EncodeCPU.h"
#include "RunLengthCPU.h"
#include "Timing.h"

using namespace cudaCompress;
using namespace tum3D;


#pragma warning( push )
#pragma warning( disable : 4996 )


int compressImage(const std::string& filenameOrig, int width, int height, int levelCount, float quantStep)
{
	int byteCount = width * height * 3;

	// read file
	byte* pData = new byte[byteCount];
	if(!readByteRaw(filenameOrig, byteCount, pData)) {
		printf("Failed opening file %s\n", filenameOrig.c_str());
		delete[] pData;
		return -1;
	}

	byte** pdpData = new byte*[levelCount + 1];
	uint byteCountLevel = byteCount;
	for(int level = 0; level <= levelCount; level++) {
		cudaSafeCall(cudaMalloc(&pdpData[level], byteCountLevel));
		byteCountLevel /= 4;
	}

	cudaMemcpy(pdpData[0], pData, width * height * 3, cudaMemcpyHostToDevice);


	void* pInstance = createInstance(-1, 2048, 2048, 1, 3, 128);


	convertRGBtoYCoCg(pdpData[0], pdpData[0], width * height);

	// propagate
	for(int level = 0; level < levelCount; level++) {
		int widthLevel  = width  >> level;
		int heightLevel = height >> level;
		getLowpassImage(pInstance, pdpData[level], widthLevel, heightLevel, 3, 3, pdpData[level+1]);
	}

	// compress and decompress
	size_t compressedSize = 0;
	std::vector<uint>* pBitStream = new std::vector<uint>[levelCount];
	for(int level = levelCount - 1; level >= 0; level--) {
		int widthLevel  = width  >> level;
		int heightLevel = height >> level;
		compressImageOneLevel(pInstance, pdpData[level], widthLevel, heightLevel, 3, 3, pdpData[level+1], pBitStream[level], quantStep);

		compressedSize += pBitStream[level].size() * sizeof(uint);

		cudaMemset(pdpData[level], 0, widthLevel * heightLevel * 3);
		decompressImageOneLevel(pInstance, pdpData[level], widthLevel, heightLevel, 3, 3, pdpData[level+1], pBitStream[level], quantStep);
	}
	compressedSize += (width / (1 << levelCount)) * (height / (1 << levelCount)) * 3;

	convertYCoCgtoRGB(pdpData[0], pdpData[0], width * height);


	destroyInstance(pInstance);


	cudaMemcpy(pData, pdpData[0], width * height * 3, cudaMemcpyDeviceToHost);


	writeByteRaw("out.raw", width * height * 3, pData);

	printf("PSNR: %.2f", computePSNRByteRaws(filenameOrig, "out.raw", width * height * 3));
	printf("     Compressed size: %8llu   bpp: %.2f\n", compressedSize, float(compressedSize * 8) / float(width * height));


	delete[] pBitStream;

	for(int level = 0; level <= levelCount; level++) {
		cudaSafeCall(cudaFree(pdpData[level]));
	}
	delete[] pdpData;

	delete[] pData;

	return 0;
}


int testOneLevelImage(const std::string& filenameOrig, int width, int height, float quantizationStep, uint iterations, ETimingDetail timingDetail = TIMING_DETAIL_NONE)
{
	const Vec2i size(width, height);
	const Vec2i blockSize(size.x()/2, size.y()/2);
	const uint elemCountTotal = size.area();
	const uint elemCountPerBlock = blockSize.area();
	const uint blockCount = 3;

	//HuffmanEncodeTable::initTimes();

	// read file
	byte* pData3 = new byte[elemCountTotal*3];
	if(!readByteRaw(filenameOrig, elemCountTotal*3, pData3)) {
		printf("Failed opening file %s\n", filenameOrig.c_str());
		delete[] pData3;
		return -1;
	}

	convertRGBToYCoCgCPU(pData3, pData3, elemCountTotal);

	byte* pData = new byte[elemCountTotal*4];
	for(uint i = 0; i < elemCountTotal; i++) {
		pData[4*i  ] = pData3[3*i  ];
		pData[4*i+1] = pData3[3*i+1];
		pData[4*i+2] = pData3[3*i+2];
		pData[4*i+3] = 0;
	}

	// allocate GPU arrays and upload data
	byte* dpImage = 0;
	cudaSafeCall(cudaMalloc(&dpImage, elemCountTotal * 4 * sizeof(byte)));
	cudaSafeCall(cudaMemcpy(dpImage, pData, elemCountTotal * 4 * sizeof(byte), cudaMemcpyHostToDevice));
	byte* dpLowpass = 0;
	cudaSafeCall(cudaMalloc(&dpLowpass, elemCountPerBlock * 4 * sizeof(byte)));

	cudaEvent_t start, end;
	cudaSafeCall(cudaEventCreate(&start));
	cudaSafeCall(cudaEventCreate(&end));
	float time = 0.0f;

	void* pInstance = createInstance(-1, size.x(), size.y(), 1, 3, 128);
	setTimingDetail(pInstance, timingDetail);

	std::vector<uint> bitStream;

	cudaProfilerStart();
	cudaSafeCall(cudaEventRecord(start));

	for(uint i = 0; i < iterations; i++) {
		getLowpassImage(pInstance, dpImage, size.x(), size.y(), 4, 3, dpLowpass); // 1.1 ms
		compressImageOneLevel(pInstance, dpImage, size.x(), size.y(), 4, 3, dpLowpass, bitStream, quantizationStep);
	}

	cudaSafeCall(cudaEventRecord(end));
	cudaProfilerStop();

	cudaSafeCall(cudaEventSynchronize(end));
	cudaSafeCall(cudaEventElapsedTime(&time, start, end));
	float throughput = iterations * float(elemCountPerBlock * blockCount) * 1000.0f / (time * 1024.0f * 1024.0f);
	printf("Compress:   %6.2f ms  (%7.2f MPix/s)\n", time / float(iterations), throughput);
	if(timingDetail != TIMING_DETAIL_NONE) {
		printf("Detailed Timings (sum for %i iterations):\n", iterations);
		printTimings(pInstance);
		resetTimings(pInstance);
		printf("\n");
	}

	//encodePrintBitCounts();
	//printf("\n");

	cudaSafeCall(cudaMemset(dpImage, 0, elemCountTotal * 4 * sizeof(byte)));

	cudaSafeCall(cudaHostRegister(bitStream.data(), bitStream.size() * sizeof(uint), cudaHostRegisterDefault));

	cudaProfilerStart();
	cudaSafeCall(cudaEventRecord(start));

	for(uint i = 0; i < iterations; i++) {
		decompressImageOneLevel(pInstance, dpImage, size.x(), size.y(), 4, 3, dpLowpass, bitStream, quantizationStep);
	}

	cudaSafeCall(cudaEventRecord(end));
	cudaProfilerStop();

	cudaSafeCall(cudaHostUnregister(bitStream.data()));

	cudaSafeCall(cudaEventSynchronize(end));
	cudaSafeCall(cudaEventElapsedTime(&time, start, end));
	throughput = iterations * float(elemCountPerBlock * blockCount) * 1000.0f / (time * 1024.0f * 1024.0f);
	printf("Decompress: %6.2f ms  (%7.2f MPix/s)\n", time / float(iterations), throughput);
	if(timingDetail != TIMING_DETAIL_NONE) {
		printf("Detailed Timings (sum for %i iterations):\n", iterations);
		printTimings(pInstance);
		resetTimings(pInstance);
		printf("\n");
	}

	memset(pData, 0, elemCountTotal * 4 * sizeof(byte));
	cudaSafeCall(cudaMemcpy(pData, dpImage, elemCountTotal * 4 * sizeof(byte), cudaMemcpyDeviceToHost));
	for(uint i = 0; i < elemCountTotal; i++) {
		pData3[3*i  ] = pData[4*i  ];
		pData3[3*i+1] = pData[4*i+1];
		pData3[3*i+2] = pData[4*i+2];
	}
	convertYCoCgToRGBCPU(pData3, pData3, elemCountTotal);
	FILE* file=fopen("out.raw", "wb");
	fwrite(pData3,elemCountTotal*3,1,file);
	fclose(file);

	destroyInstance(pInstance);

	cudaSafeCall(cudaFree(dpLowpass));
	cudaSafeCall(cudaFree(dpImage));

	delete[] pData3;
	delete[] pData;

	//HuffmanEncodeTable::printTimes();

	//printf("Highpass size: %u\n", bitStream.getRawSizeBytes());
	printf("PSNR: %.2f\n", computePSNRByteRaws(filenameOrig, "out.raw", elemCountTotal*3));
	printf("\n\n");

	return 0;
}

int testOneLevelHeightfield(const std::string& filenameOrig, int width, int height, uint iterations, ETimingDetail timingDetail = TIMING_DETAIL_NONE)
{
	const Vec2i size(width, height);
	const Vec2i blockSize(size.x()/2, size.y()/2);
	const uint elemCountTotal = size.area();
	const uint elemCountPerBlock = blockSize.area();
	const uint blockCount = 3;

	//HuffmanEncodeTable::initTimes();

	// read file
	short* pData = new short[elemCountTotal];
	if(!readShortRaw(filenameOrig, elemCountTotal, pData)) {
		printf("Failed opening file %s\n", filenameOrig.c_str());
		delete[] pData;
		return -1;
	}

	//// upsample test: leave only upper left quadrant and upsample
	//for(int y=0;y<size.y();y++){
	//	for(int x=0;x<size.x();x++){
	//		if(x>=size.x()/2 || y >= size.y()/2){
	//			pData[x+y*size.x()]=0;
	//		}
	//	}
	//}
	//short*pTemp=new short[elemCountTotal];
	//unliftCDF53_y(pData, pTemp, size.xy());
	//unliftCDF53_x(pTemp, pData, size.xy());
	//delete[]pTemp;

	// allocate GPU arrays and upload data
	short* dpImage = 0;
	cudaSafeCall(cudaMalloc(&dpImage, elemCountTotal * sizeof(short)));
	cudaSafeCall(cudaMemcpy(dpImage, pData, elemCountTotal * sizeof(short), cudaMemcpyHostToDevice));
	short* dpLowpass = 0;
	cudaSafeCall(cudaMalloc(&dpLowpass, elemCountPerBlock * sizeof(short)));

	cudaEvent_t start, end;
	cudaSafeCall(cudaEventCreate(&start));
	cudaSafeCall(cudaEventCreate(&end));
	float time = 0.0f;

	void* pInstance = createInstance(-1, size.x(), size.y(), 1, 1, 128);
	setTimingDetail(pInstance, timingDetail);

	std::vector<uint> bitStream;

	cudaProfilerStart();
	cudaSafeCall(cudaEventRecord(start));

	for(uint i = 0; i < iterations; i++) {
		compressHeightfieldOneLevel(pInstance, dpImage, size.x(), size.y(), dpLowpass, 0, bitStream);
	}

	cudaSafeCall(cudaEventRecord(end));
	cudaProfilerStop();

	cudaSafeCall(cudaEventSynchronize(end));
	cudaSafeCall(cudaEventElapsedTime(&time, start, end));
	float throughput = iterations * float(elemCountPerBlock * blockCount) * 1000.0f / (time * 1024.0f * 1024.0f);
	printf("Compress:   %6.2f ms  (%7.2f MPix/s)\n", time / float(iterations), throughput);
	if(timingDetail != TIMING_DETAIL_NONE) {
		printf("Detailed Timings (sum for %i iterations):\n", iterations);
		printTimings(pInstance);
		resetTimings(pInstance);
		printf("\n");
	}

	//encodePrintBitCounts();
	//printf("\n");

	cudaSafeCall(cudaMemset(dpImage, 0, elemCountTotal * sizeof(short)));

	// decompressHeightfieldOneLevel modifies the lowpass image, so make a backup first
	short* dpLowpassBak = 0;
	cudaSafeCall(cudaMalloc(&dpLowpassBak, elemCountPerBlock * sizeof(short)));
	cudaSafeCall(cudaMemcpy(dpLowpassBak, dpLowpass, elemCountPerBlock * sizeof(short), cudaMemcpyDeviceToDevice));

	// note: cudaHostRegister on non-4k aligned memory is not allowed in CUDA 4.0, but seems to work fine
	cudaSafeCall(cudaHostRegister(bitStream.data(), bitStream.size() * sizeof(uint), cudaHostRegisterDefault));

	cudaProfilerStart();
	cudaSafeCall(cudaEventRecord(start));

	for(uint i = 0; i < iterations; i++) {
		cudaSafeCall(cudaMemcpy(dpLowpass, dpLowpassBak, elemCountPerBlock * sizeof(short), cudaMemcpyDeviceToDevice));
		decompressHeightfieldOneLevel(pInstance, dpImage, size.x(), size.y(), dpLowpass, 0, bitStream);
	}

	cudaSafeCall(cudaEventRecord(end));
	cudaProfilerStop();

	cudaSafeCall(cudaHostUnregister(bitStream.data()));

	cudaSafeCall(cudaFree(dpLowpassBak));

	cudaSafeCall(cudaEventSynchronize(end));
	cudaSafeCall(cudaEventElapsedTime(&time, start, end));
	throughput = iterations * float(elemCountPerBlock * blockCount) * 1000.0f / (time * 1024.0f * 1024.0f);
	printf("Decompress: %6.2f ms  (%7.2f MPix/s)\n", time / float(iterations), throughput);
	if(timingDetail != TIMING_DETAIL_NONE) {
		printf("Detailed Timings (sum for %i iterations):\n", iterations);
		printTimings(pInstance);
		resetTimings(pInstance);
		printf("\n");
	}

	memset(pData, 0, elemCountTotal * sizeof(short));
	cudaSafeCall(cudaMemcpy(pData, dpImage, elemCountTotal * sizeof(short), cudaMemcpyDeviceToHost));

	writeShortRaw("out.raw", elemCountTotal, pData);

	destroyInstance(pInstance);

	cudaSafeCall(cudaFree(dpLowpass));
	cudaSafeCall(cudaFree(dpImage));

	delete[] pData;

	//HuffmanEncodeTable::printTimes();

	//printf("Highpass size: %u\n", bitStream.getRawSizeBytes());
	printf("%s\n", compareShortRaws(filenameOrig, "out.raw", elemCountTotal) ? "EQUAL" : "NOT EQUAL" );
	printf("\n\n");

	return 0;
}

int testVolumeLossless(const std::string& filenameOrig, bool short12, int width, int height, int depth, int numLevels, ETimingDetail timingDetail = TIMING_DETAIL_NONE)
{
	const Vec3i size(width, height, depth);
	const Vec3i blockSize(size.x()/2, size.y()/2, size.z()/2);
	const uint elemCountTotal = size.volume();
	const uint elemCountPerBlock = blockSize.volume();

	const uint iterations = 10;

	// read file
	short* pData = new short[elemCountTotal];
	bool fail = true;
	if(short12) {
		fail = !readShortRaw(filenameOrig, elemCountTotal, pData);
		// convert 12bit ushort (4 LSBs are zero) to zero-centered sshort
		for(uint i = 0; i < elemCountTotal; i++) {
			ushort val = ((ushort*)pData)[i];
			pData[i] = short(val >> 4) - 2048;
			//pData[i] = short(val >> 6) - 512;
		}
	} else {
		fail = !readByteRawAsShort(filenameOrig, elemCountTotal, pData);
	}
	if(fail) {
		printf("Failed opening file %s\n", filenameOrig.c_str());
		delete[] pData;
		return -1;
	}

	// allocate GPU arrays and upload data
	short* dpImage = 0;
	cudaSafeCall(cudaMalloc(&dpImage, elemCountTotal * sizeof(short)));
	cudaSafeCall(cudaMemcpy(dpImage, pData, elemCountTotal * sizeof(short), cudaMemcpyHostToDevice));

	cudaEvent_t start, end;
	cudaSafeCall(cudaEventCreate(&start));
	cudaSafeCall(cudaEventCreate(&end));
	float time = 0.0f;

	void* pInstance = createInstance(-1, size.x(), size.y(), size.z(), 1, 128);
	setTimingDetail(pInstance, timingDetail);

	std::vector<uint> bitStream;

	cudaProfilerStart();
	cudaSafeCall(cudaEventRecord(start));

	for(uint i = 0; i < iterations; i++) {
		compressVolumeLossless(pInstance, dpImage, size.x(), size.y(), size.z(), numLevels, bitStream);
	}

	cudaSafeCall(cudaEventRecord(end));
	cudaProfilerStop();

	cudaSafeCall(cudaEventSynchronize(end));
	cudaSafeCall(cudaEventElapsedTime(&time, start, end));
	float throughput = iterations * float(elemCountTotal) * 1000.0f / (time * 1024.0f * 1024.0f);
	printf("Compress:   %6.2f ms  (%7.2f MPix/s)\n", time / float(iterations), throughput);
	if(timingDetail != TIMING_DETAIL_NONE) {
		printf("Detailed Timings (sum for %i iterations):\n", iterations);
		printTimings(pInstance);
		resetTimings(pInstance);
		printf("\n");
	}

	//encodePrintBitCounts();
	//printf("\n");

	cudaSafeCall(cudaMemset(dpImage, 0, elemCountTotal * sizeof(short)));

	// note: cudaHostRegister on non-4k aligned memory is not allowed in CUDA 4.0, but seems to work fine
	cudaSafeCall(cudaHostRegister(bitStream.data(), bitStream.size() * sizeof(uint), cudaHostRegisterDefault));

	cudaProfilerStart();
	cudaSafeCall(cudaEventRecord(start));

	for(uint i = 0; i < iterations; i++) {
		//decompressVolumeLosslessOneLevel(pInstance, dpImage, size.x(), size.y(), size.z(), dpLowpass, bitStream);
		decompressVolumeLossless(pInstance, dpImage, size.x(), size.y(), size.z(), numLevels, bitStream);
	}

	cudaSafeCall(cudaEventRecord(end));
	cudaProfilerStop();

	cudaSafeCall(cudaHostUnregister(bitStream.data()));

	cudaSafeCall(cudaEventSynchronize(end));
	cudaSafeCall(cudaEventElapsedTime(&time, start, end));
	throughput = iterations * float(elemCountTotal) * 1000.0f / (time * 1024.0f * 1024.0f);
	printf("Decompress: %6.2f ms  (%7.2f MPix/s)\n", time / float(iterations), throughput);
	if(timingDetail != TIMING_DETAIL_NONE) {
		printf("Detailed Timings (sum for %i iterations):\n", iterations);
		printTimings(pInstance);
		resetTimings(pInstance);
		printf("\n");
	}

	memset(pData, 0, elemCountTotal * sizeof(short));
	cudaSafeCall(cudaMemcpy(pData, dpImage, elemCountTotal * sizeof(short), cudaMemcpyDeviceToHost));

	if(short12) {
		// convert zero-centered sshort to 12bit ushort (4 LSBs are zero)
		for(uint i = 0; i < elemCountTotal; i++) {
			short val = pData[i];
			((ushort*)pData)[i] = ushort(val + 2048) << 4;
			//((ushort*)pData)[i] = ushort(val + 512) << 6;
		}
		writeShortRaw("out.raw", elemCountTotal, pData);
	} else {
		writeByteRawFromShort("out.raw", elemCountTotal, pData);
	}

	destroyInstance(pInstance);

	//cudaSafeCall(cudaFree(dpLowpass));
	cudaSafeCall(cudaFree(dpImage));

	delete[] pData;

	uint originalSizeBytes = elemCountTotal;
	if(short12) originalSizeBytes += originalSizeBytes / 2; // * 1.5
	uint compressedSizeBytes = uint(bitStream.size()) * sizeof(uint);
	printf("Compressed size: %u B  (%.2f : 1)\n", compressedSizeBytes, float(originalSizeBytes) / float(compressedSizeBytes));
	printf("%s\n", compareByteRaws(filenameOrig, "out.raw", elemCountTotal) ? "EQUAL" : "NOT EQUAL");
	printf("\n\n");

	return 0;
}

int testVolumeLossy(const std::string& filenameOrig, int width, int height, int depth, int numLevels, float quantStep, ETimingDetail timingDetail = TIMING_DETAIL_NONE)
{
	const Vec3i size(width, height, depth);
	const Vec3i blockSize(size.x()/2, size.y()/2, size.z()/2);
	const uint elemCountTotal = size.volume();
	const uint elemCountPerBlock = blockSize.volume();

	const uint iterations = 1;

	// read file
	byte* pData = new byte[elemCountTotal];
	if(!readByteRaw(filenameOrig, elemCountTotal, pData)) {
		printf("Failed opening file %s\n", filenameOrig.c_str());
		delete[] pData;
		return -1;
	}

	// allocate GPU arrays and upload data
	byte* dpImage = 0;
	cudaSafeCall(cudaMalloc(&dpImage, elemCountTotal * sizeof(byte)));
	cudaSafeCall(cudaMemcpy(dpImage, pData, elemCountTotal * sizeof(byte), cudaMemcpyHostToDevice));

	cudaEvent_t start, end;
	cudaSafeCall(cudaEventCreate(&start));
	cudaSafeCall(cudaEventCreate(&end));
	float time = 0.0f;

	void* pInstance = createInstance(-1, size.x(), size.y(), size.z(), 1, 128);
	setTimingDetail(pInstance, timingDetail);

	std::vector<uint> bitStream;

	cudaProfilerStart();
	cudaSafeCall(cudaEventRecord(start));

	for(uint i = 0; i < iterations; i++) {
		compressVolume(pInstance, dpImage, size.x(), size.y(), size.z(), numLevels, bitStream, quantStep);
	}

	cudaSafeCall(cudaEventRecord(end));
	cudaProfilerStop();

	cudaSafeCall(cudaEventSynchronize(end));
	cudaSafeCall(cudaEventElapsedTime(&time, start, end));
	float throughput = iterations * float(elemCountTotal) * 1000.0f / (time * 1024.0f * 1024.0f);
	printf("Compress:   %6.2f ms  (%7.2f MPix/s)\n", time / float(iterations), throughput);
	if(timingDetail != TIMING_DETAIL_NONE) {
		printf("Detailed Timings (sum for %i iterations):\n", iterations);
		printTimings(pInstance);
		resetTimings(pInstance);
		printf("\n");
	}

	//encodePrintBitCounts();
	//printf("\n");

	cudaSafeCall(cudaMemset(dpImage, 0, elemCountTotal * sizeof(byte)));

	// note: cudaHostRegister on non-4k aligned memory is not allowed in CUDA 4.0, but seems to work fine
	cudaSafeCall(cudaHostRegister(bitStream.data(), bitStream.size() * sizeof(uint), cudaHostRegisterDefault));

	cudaProfilerStart();
	cudaSafeCall(cudaEventRecord(start));

	for(uint i = 0; i < iterations; i++) {
		decompressVolume(pInstance, dpImage, size.x(), size.y(), size.z(), numLevels, bitStream, quantStep);
	}

	cudaSafeCall(cudaEventRecord(end));
	cudaProfilerStop();

	cudaSafeCall(cudaHostUnregister(bitStream.data()));

	cudaSafeCall(cudaEventSynchronize(end));
	cudaSafeCall(cudaEventElapsedTime(&time, start, end));
	throughput = iterations * float(elemCountTotal) * 1000.0f / (time * 1024.0f * 1024.0f);
	printf("Decompress: %6.2f ms  (%7.2f MPix/s)\n", time / float(iterations), throughput);
	if(timingDetail != TIMING_DETAIL_NONE) {
		printf("Detailed Timings (sum for %i iterations):\n", iterations);
		printTimings(pInstance);
		resetTimings(pInstance);
		printf("\n");
	}

	memset(pData, 0, elemCountTotal * sizeof(byte));
	cudaSafeCall(cudaMemcpy(pData, dpImage, elemCountTotal * sizeof(byte), cudaMemcpyDeviceToHost));

	writeByteRaw("out.raw", elemCountTotal, pData);

	destroyInstance(pInstance);

	cudaSafeCall(cudaFree(dpImage));

	delete[] pData;

	uint compressedSizeBytes = uint(bitStream.size()) * sizeof(uint);
	printf("Compressed size: %u B  (%.2f : 1)\n", compressedSizeBytes, float(elemCountTotal) / float(compressedSizeBytes));
	float psnr = 0.0f;
	computePSNRByteRaws(filenameOrig, "out.raw", elemCountTotal, &psnr, 1);
	printf("PSNR: %.2f\n", psnr );
	printf("\n\n");

	return 0;
}

int testVolumeFloat(const std::string& filenameOrig, int width, int height, int depth, int numLevels, float quantStep, uint iterations, ETimingDetail timingDetail = TIMING_DETAIL_NONE, const std::string& filenameOut = "out.raw", const std::string& filenameComp = "")
{
	const Vec3i size(width, height, depth);
	const Vec3i blockSize(size.x()/2, size.y()/2, size.z()/2);
	const uint elemCountTotal = size.volume();
	const uint elemCountPerBlock = blockSize.volume();

	// read file
	float* pData = new float[elemCountTotal];
	if(!readFloatRaw(filenameOrig, elemCountTotal, pData)) {
		printf("Failed opening file %s\n", filenameOrig.c_str());
		delete[] pData;
		return -1;
	}

	// allocate GPU arrays and upload data
	float* dpImage = 0;
	cudaSafeCall(cudaMalloc(&dpImage, elemCountTotal * sizeof(float)));
	cudaSafeCall(cudaMemcpy(dpImage, pData, elemCountTotal * sizeof(float), cudaMemcpyHostToDevice));

	cudaEvent_t start, end;
	cudaSafeCall(cudaEventCreate(&start));
	cudaSafeCall(cudaEventCreate(&end));
	float time = 0.0f;

	void* pInstance = createInstance(-1, size.x(), size.y(), size.z(), 1, 128);
	setTimingDetail(pInstance, timingDetail);

	std::vector<uint> bitStream;

	cudaProfilerStart();
	cudaSafeCall(cudaEventRecord(start));

	for(uint i = 0; i < iterations; i++) {
		compressVolumeFloat(pInstance, dpImage, size.x(), size.y(), size.z(), numLevels, bitStream, quantStep);
	}

	cudaSafeCall(cudaEventRecord(end));
	cudaProfilerStop();

	cudaSafeCall(cudaEventSynchronize(end));
	cudaSafeCall(cudaEventElapsedTime(&time, start, end));
	float throughput = iterations * float(elemCountTotal) * 1000.0f / (time * 1024.0f * 1024.0f);
	printf("Compress:   %6.2f ms  (%7.2f MPix/s)\n", time / float(iterations), throughput);
	if(timingDetail != TIMING_DETAIL_NONE) {
		printf("Detailed Timings (sum for %i iterations):\n", iterations);
		printTimings(pInstance);
		resetTimings(pInstance);
		printf("\n");
	}

	//encodePrintBitCounts();
	//printf("\n");

	if(!filenameComp.empty()) {
		FILE* fileComp = 0;
		fopen_s(&fileComp, filenameComp.c_str(), "wb");
		if(!fileComp) {
			printf("Failed to open file %s for writing.\n", filenameComp.c_str());
		} else {
			fwrite(bitStream.data(), sizeof(uint), bitStream.size(), fileComp);
			fclose(fileComp);
		}
	}

	cudaSafeCall(cudaMemset(dpImage, 0, elemCountTotal * sizeof(float)));

	cudaSafeCall(cudaHostRegister(bitStream.data(), bitStream.size() * sizeof(uint), cudaHostRegisterDefault));

	cudaProfilerStart();
	cudaSafeCall(cudaEventRecord(start));

	for(uint i = 0; i < iterations; i++) {
		decompressVolumeFloat(pInstance, dpImage, size.x(), size.y(), size.z(), numLevels, bitStream, quantStep);
		//decompressVolumeFloatLowpass(dpImage, size.x(), size.y(), size.z(), numLevels, bitStream, quantStep);
	}

	cudaSafeCall(cudaEventRecord(end));
	cudaProfilerStop();

	cudaSafeCall(cudaHostUnregister(bitStream.data()));

	//FILE* fileLow = 0;
	//fopen_s(&fileLow, "low.raw", "wb");
	//if(!fileLow) {
	//	exit(-42);
	//} else {
	//	std::vector<float> data;
	//	data.resize(64*64*64);
	//	cudaMemcpy(data.data(), dpImage, 64*64*64*4, cudaMemcpyDeviceToHost);
	//	fwrite(data.data(), sizeof(float), data.size(), fileLow);
	//	fclose(fileLow);
	//}

	cudaSafeCall(cudaEventSynchronize(end));
	cudaSafeCall(cudaEventElapsedTime(&time, start, end));
	throughput = iterations * float(elemCountTotal) * 1000.0f / (time * 1024.0f * 1024.0f);
	printf("Decompress: %6.2f ms  (%7.2f MPix/s)\n", time / float(iterations), throughput);
	if(timingDetail != TIMING_DETAIL_NONE) {
		printf("Detailed Timings (sum for %i iterations):\n", iterations);
		printTimings(pInstance);
		resetTimings(pInstance);
		printf("\n");
	}

	memset(pData, 0, elemCountTotal * sizeof(float));
	cudaSafeCall(cudaMemcpy(pData, dpImage, elemCountTotal * sizeof(float), cudaMemcpyDeviceToHost));

	writeFloatRaw(filenameOut, elemCountTotal, pData);

	destroyInstance(pInstance);

	cudaSafeCall(cudaFree(dpImage));

	delete[] pData;

	uint compressedSizeBytes = uint(bitStream.size()) * sizeof(uint);
	float compressionFactor = sizeof(float) * float(elemCountTotal) / float(compressedSizeBytes);
	float range, maxe, rmse, psnr, snr;
	computeStatsFloatRaws(filenameOrig, filenameOut, elemCountTotal, &range, &maxe, &rmse, &psnr, &snr);

	printf("Compressed size: %u B  (%.2f : 1)\n", compressedSizeBytes, compressionFactor);
	printf("Range: %.3f   MaxE: %.4f   RMSE: %.4f   PSNR: %.2f   SNR: %.2f\n", range, maxe, rmse, psnr, snr);
	printf("\n\n");

	std::ofstream info("info.txt", std::ios_base::app | std::ios_base::ate);
	info << filenameOut << std::endl;
	info << "Quantization step: " << quantStep << std::endl;
	info << "Compressed size: " << compressedSizeBytes << " B  (" << compressionFactor << " : 1)" << std::endl;
	info << "Range: " << range << "\nRMSE: " << rmse << "\nPSNR: " << psnr << "\nSNR: " << snr << "\n" << std::endl;
	info.close();

	return 0;
}

std::vector<int> getCudaDevices()
{
	std::vector<int> usableDevices;

	// loop over all cuda devices, collect those that we can use
	int deviceCount = 0;
	cudaSafeCall(cudaGetDeviceCount(&deviceCount));

	for(int device = 0; device < deviceCount; device++)
	{
		cudaDeviceProp prop;
		cudaSafeCall(cudaGetDeviceProperties(&prop, device));

		bool deviceOK = (prop.major >= 2);
		if(deviceOK)
		{
			printf("Found usable CUDA device %i: %s, compute %i.%i\n", device, prop.name, prop.major, prop.minor);
			usableDevices.push_back(device);
		}
	}

	return usableDevices;
}

struct TestVolumeFloatArgs
{
	TestVolumeFloatArgs()
		: step(0)
		, device(-1)
		, pData(nullptr)
		, width(0), height(0), depth(0), numLevels(0), quantStep(0.0f), iterations(0) {}

	TestVolumeFloatArgs(int device, float* pData, int width, int height, int depth, int numLevels, float quantStep, uint iterations, ETimingDetail timingDetail)
		: step(0)
		, device(device)
		, pData(pData)
		, width(width), height(height), depth(depth), numLevels(numLevels), quantStep(quantStep)
		, iterations(iterations)
		, timingDetail(timingDetail) {}

	void set(int device, float* pData, int width, int height, int depth, int numLevels, float quantStep, uint iterations, ETimingDetail timingDetail)
	{
		this->device = device;
		this->pData = pData;
		this->width = width;
		this->height = height;
		this->depth = depth;
		this->numLevels = numLevels;
		this->quantStep = quantStep;
		this->iterations = iterations;
		this->timingDetail = timingDetail;
	}

	int step;
	tthread::mutex mutex;
	tthread::condition_variable stepDone;

	int device;

	float* pData;
	int width;
	int height;
	int depth;
	int numLevels;
	float quantStep;
	uint iterations;
	ETimingDetail timingDetail;
};

void testVolumeFloatMultiGPUThreadFunc(void* pArgsRaw)
{
	TestVolumeFloatArgs* pArgs = (TestVolumeFloatArgs*)pArgsRaw;

	const Vec3i size(pArgs->width, pArgs->height, pArgs->depth);
	const Vec3i blockSize(size.x()/2, size.y()/2, size.z()/2);
	const uint elemCountTotal = size.volume();
	const uint elemCountPerBlock = blockSize.volume();

	cudaSafeCall(cudaSetDevice(pArgs->device));

	// allocate GPU arrays and upload data
	float* dpImage = 0;
	cudaSafeCall(cudaMalloc(&dpImage, elemCountTotal * sizeof(float)));
	cudaSafeCall(cudaMemcpy(dpImage, pArgs->pData, elemCountTotal * sizeof(float), cudaMemcpyHostToDevice));

	cudaEvent_t start, end;
	cudaSafeCall(cudaEventCreate(&start));
	cudaSafeCall(cudaEventCreate(&end));
	float time = 0.0f;

	void* pInstance = createInstance(-1, size.x(), size.y(), size.z(), 1, 128);
	setTimingDetail(pInstance, pArgs->timingDetail);

	// signal that initialization is done and encoding is about to start
	{ tthread::lock_guard<tthread::mutex> guard(pArgs->mutex);
		pArgs->step++;
	}
	pArgs->stepDone.notify_all();

	std::vector<uint> bitStream;

	cudaProfilerStart();
	cudaSafeCall(cudaEventRecord(start));

	for(uint i = 0; i < pArgs->iterations; i++) {
		compressVolumeFloat(pInstance, dpImage, size.x(), size.y(), size.z(), pArgs->numLevels, bitStream, pArgs->quantStep);
	}

	cudaSafeCall(cudaEventRecord(end));
	cudaProfilerStop();

	// signal that encoding is done
	{ tthread::lock_guard<tthread::mutex> guard(pArgs->mutex);
		pArgs->step++;
	}
	pArgs->stepDone.notify_all();

	cudaSafeCall(cudaEventSynchronize(end));
	cudaSafeCall(cudaEventElapsedTime(&time, start, end));
	float throughput = pArgs->iterations * float(elemCountTotal) * 1000.0f / (time * 1024.0f * 1024.0f);
	printf("Device %i: Compress:   %6.2f ms  (%7.2f MPix/s)\n", pArgs->device, time / float(pArgs->iterations), throughput);
	if(pArgs->timingDetail != TIMING_DETAIL_NONE) {
		printf("Detailed Timings (sum for %i iterations):\n", pArgs->iterations);
		printTimings(pInstance);
		resetTimings(pInstance);
		printf("\n");
	}

	//encodePrintBitCounts();
	//printf("\n");

	cudaSafeCall(cudaMemset(dpImage, 0, elemCountTotal * sizeof(float)));

	cudaSafeCall(cudaHostRegister(bitStream.data(), bitStream.size() * sizeof(uint), cudaHostRegisterDefault));

	// signal that decoding is about to start
	{ tthread::lock_guard<tthread::mutex> guard(pArgs->mutex);
		pArgs->step++;
	}
	pArgs->stepDone.notify_all();

	cudaProfilerStart();
	cudaSafeCall(cudaEventRecord(start));

	for(uint i = 0; i < pArgs->iterations; i++) {
		decompressVolumeFloat(pInstance, dpImage, size.x(), size.y(), size.z(), pArgs->numLevels, bitStream, pArgs->quantStep);
	}

	cudaSafeCall(cudaEventRecord(end));
	cudaProfilerStop();

	// signal that decoding is done
	{ tthread::lock_guard<tthread::mutex> guard(pArgs->mutex);
		pArgs->step++;
	}
	pArgs->stepDone.notify_all();

	cudaSafeCall(cudaHostUnregister(bitStream.data()));

	cudaSafeCall(cudaEventSynchronize(end));
	cudaSafeCall(cudaEventElapsedTime(&time, start, end));
	throughput = pArgs->iterations * float(elemCountTotal) * 1000.0f / (time * 1024.0f * 1024.0f);
	printf("Device %i: Decompress: %6.2f ms  (%7.2f MPix/s)\n", pArgs->device, time / float(pArgs->iterations), throughput);
	if(pArgs->timingDetail != TIMING_DETAIL_NONE) {
		printf("Detailed Timings (sum for %i iterations):\n", pArgs->iterations);
		printTimings(pInstance);
		resetTimings(pInstance);
		printf("\n");
	}

	//memset(pArgs->pData, 0, elemCountTotal * sizeof(float));
	//cudaSafeCall(cudaMemcpy(pArgs->pData, dpImage, elemCountTotal * sizeof(float), cudaMemcpyDeviceToHost));

	destroyInstance(pInstance);

	cudaSafeCall(cudaFree(dpImage));

	//uint compressedSizeBytes = uint(bitStream.size()) * sizeof(uint);
	//float compressionFactor = sizeof(float) * float(elemCountTotal) / float(compressedSizeBytes);
	//float range, maxe, rmse, psnr, snr;
	//computeStatsFloatRaws(filenameOrig, filenameOut, elemCountTotal, &range, &maxe, &rmse, &psnr, &snr);

	//printf("Compressed size: %u B  (%.2f : 1)\n", compressedSizeBytes, compressionFactor);
	//printf("Range: %.3f   MaxE: %.4f   RMSE: %.4f   PSNR: %.2f   SNR: %.2f\n", range, maxe, rmse, psnr, snr);
	//printf("\n\n");
}

int testVolumeFloatMultiGPU(const std::string& filenameOrig, int width, int height, int depth, int numLevels, float quantStep, uint iterations, ETimingDetail timingDetail = TIMING_DETAIL_NONE)
{
	const Vec3i size(width, height, depth);
	const Vec3i blockSize(size.x()/2, size.y()/2, size.z()/2);
	const uint elemCountTotal = size.volume();
	const uint elemCountPerBlock = blockSize.volume();

	// read file
	float* pData = new float[elemCountTotal];
	if(!readFloatRaw(filenameOrig, elemCountTotal, pData)) {
		printf("Failed opening file %s\n", filenameOrig.c_str());
		delete[] pData;
		return -1;
	}

	std::vector<int> devices = getCudaDevices();

	TestVolumeFloatArgs* args = new TestVolumeFloatArgs[devices.size()];
	//std::vector<TestVolumeFloatArgs> args;
	for(size_t i = 0; i < devices.size(); i++)
	{
		args[i].set(devices[i], pData, width, height, depth, numLevels, quantStep, iterations, timingDetail);
	}

	// start threads
	std::vector<tthread::thread*> threads;
	for(size_t i = 0; i < devices.size(); i++)
	{
		threads.push_back(new tthread::thread(testVolumeFloatMultiGPUThreadFunc, &args[i]));
	}

	// wait for threads to finish their steps one by one
	int stepMax = 4;
	for(int step = 1; step <= stepMax; step++)
	{
		for(size_t i = 0; i < threads.size(); i++)
		{
			tthread::lock_guard<tthread::mutex> guard(args[i].mutex);
			while(args[i].step < stepMax)
			{
				args[i].stepDone.wait(args[i].mutex);
			}
		}
	}

	// wait for threads to shut down
	for(size_t i = 0; i < threads.size(); i++)
	{
		threads[i]->join();
		delete threads[i];
	}

	delete[] args;

	printf("DONE.\n\n");

	return 0;
}


void convertFloatRawToShort(const std::string& filenameIn, const std::string& filenameOut, float quantStep, uint elemCount)
{
	FILE* file = fopen(filenameIn.c_str(), "rb");
	if(!file) {
		printf("Can't open file %s\n", filenameIn.c_str());
		exit(-1);
	}

	float* pDataFloat = new float[elemCount];
	fread(pDataFloat, sizeof(float), elemCount, file);

	fclose(file);

	float heightMin = 1e10;
	float heightMax = -1e10;
	for(uint i = 0; i < elemCount; i++) {
		if(pDataFloat[i] < heightMin) heightMin = pDataFloat[i];
		if(pDataFloat[i] > heightMax) heightMax = pDataFloat[i];
	}

	short shift = short((heightMax - heightMin) / 2.0f + 0.5f);

	short* pDataShort = new short[elemCount];
	for(uint i = 0; i < elemCount; i++) {
		float valf = pDataFloat[i] - heightMin; // make >= 0
		pDataShort[i] = short(valf / quantStep + 0.5f) - shift;
	}

	file = fopen(filenameOut.c_str(), "wb");
	if(!file) {
		printf("Can't create file %s\n", filenameOut.c_str());
		exit(-1);
	}

	fwrite(pDataShort, sizeof(short), elemCount, file);
	fclose(file);

	delete[] pDataShort;
	delete[] pDataFloat;
}


void compareSymbolStreams(const std::vector<Symbol>& symbols1, const std::vector<Symbol>& symbols2)
{
	uint count = (uint)min(symbols1.size(), symbols2.size());
	for(uint i = 0; i < count; i++) {
		if(symbols1[i] != symbols2[i]) {
			printf("ERROR at index %i: %i vs. %i\n", i, int(symbols1[i]), int(symbols2[i]));
			break;
		}
	}
}

bool testVolumeEncodingCPU(const std::string& filename, uint sizeX, uint sizeY, uint sizeZ, uint numLevels, float quantizationStepLevel0)
{
	uint offsetInterval = 128;


	FILE* file = fopen(filename.c_str(), "rb");
	if(!file) return false;

	Vec3i size(sizeX, sizeY, sizeZ);
	uint elemCountTotal = size.volume();

	int bytes = elemCountTotal * sizeof(float);

	float* pData = new float[elemCountTotal];
	fread(pData, sizeof(float), elemCountTotal, file);
	fclose(file);

	float* dpData = nullptr;
	cudaSafeCall(cudaMalloc(&dpData, bytes));
	cudaSafeCall(cudaMemcpy(dpData, pData, bytes, cudaMemcpyHostToDevice));

	void* pInstance = createInstance(-1, sizeX, sizeY, sizeZ, 1, offsetInterval);




	const uint blockCount = 7;

	Vec3i blockSizeMax = ((size + 1) / 2);
	uint elemCountPerBlockMax = blockSizeMax.volume();

	std::vector<Symbol*> dpSymbolStreams(blockCount);
	for(uint i = 0; i < blockCount; i++) {
		cudaSafeCall(cudaMalloc(&dpSymbolStreams[i], elemCountPerBlockMax * sizeof(Symbol)));
	}

	float* dpDWT[2];
	cudaSafeCall(cudaMalloc(&dpDWT[0], elemCountTotal * sizeof(float)));
	cudaSafeCall(cudaMalloc(&dpDWT[1], elemCountTotal * sizeof(float)));


	// perform full dwt
	int d = 0;
	dwtFloat3DForward(dpDWT[d], dpDWT[1-d], dpDWT[d], dpData, sizeX, sizeY, sizeZ);
	uint factor = 2;
	for(uint level = 1; level < numLevels; level++) {
		dwtFloat3DForward(dpDWT[1-d], dpDWT[d], dpDWT[1-d], dpDWT[d], sizeX / factor, sizeY / factor, sizeZ / factor, 1, sizeX, sizeX * sizeY, sizeX, sizeX * sizeY);
		d = 1 - d;
		factor *= 2;
	}


	uint sizeTotal[CODER_COUNT] = {};


	std::vector<Symbol> symbolStream;
	std::vector<Symbol> symbolStreamReconst;
	BitStream bitStream;


	// quantize and encode lowpass band
	float quantizationStep = quantizationStepLevel0 / float(factor/2);
	quantizeToSymbols3D(dpSymbolStreams[0], dpDWT[d], quantizationStep, sizeX / factor, sizeY / factor, sizeZ / factor, sizeX, sizeX * sizeY);
	//bool result = encodeHuff(&pBitStreams, true, &dpSymbolStreams[0], 1, size.volume() / (factor * factor * factor));
	symbolStream.resize(size.volume() / (factor * factor * factor));
	cudaSafeCall(cudaMemcpy(symbolStream.data(), dpSymbolStreams[0], symbolStream.size() * sizeof(Symbol), cudaMemcpyDeviceToHost));
	for(ECoder coder = ECoder(0); coder < CODER_COUNT; coder = ECoder(coder + 1)) {
		printf("%20s level %i:", getCoderName(coder).c_str(), numLevels);

		printf(" enc...");
		bitStream.setBitSize(0);
		encodeCPU(coder, &bitStream, &symbolStream, offsetInterval);
		sizeTotal[coder] += bitStream.getBitSize();

		printf(" dec...");
		bitStream.setBitPosition(0);
		decodeCPU(coder, &bitStream, &symbolStreamReconst, (uint)symbolStream.size(), offsetInterval);
		printf(" comp...");
		compareSymbolStreams(symbolStream, symbolStreamReconst);
		printf(" done.\n");
	}

	// highpass bands
	for(int level = numLevels - 1; level >= 0; level--) {
		quantizationStep = quantizationStepLevel0 / float(factor/2);

		Vec3i blockSize = size / factor;

		// symbolize highpass bands
		uint nextSymbolStream = 0;
		for(uint block = 1; block < 1 + blockCount; block++) {
			uint x = block % 2;
			uint y = block / 2 % 2;
			uint z = block / 4;
			uint offset = x * blockSize.x() + y * blockSize.y() * size.x() + z * blockSize.z() * size.x() * size.y();

			// make (unsigned!) symbols
			quantizeToSymbols3D(dpSymbolStreams[nextSymbolStream++], dpDWT[d] + offset, quantizationStep, blockSize.x(), blockSize.y(), blockSize.z(), size.x(), size.x() * size.y());
		}

		// encode
		//bool result = encodeRLHuff(&pBitStreams, true, &dpSymbolStreams[0], blockCount, blockSize.volume());
		symbolStream.resize(blockSize.volume());
		for(uint i = 0; i < blockCount; i++) {
			cudaSafeCall(cudaMemcpy(symbolStream.data(), dpSymbolStreams[i], symbolStream.size() * sizeof(Symbol), cudaMemcpyDeviceToHost));
			for(ECoder coder = ECoder(0); coder < CODER_COUNT; coder = ECoder(coder + 1)) {
				printf("%20s level %u block %u:", getCoderName(coder).c_str(), level, i);

				printf(" enc...");
				bitStream.setBitSize(0);
				encodeCPU(coder, &bitStream, &symbolStream, offsetInterval);
				sizeTotal[coder] += bitStream.getBitSize();

				printf(" dec...");
				bitStream.setBitPosition(0);
				decodeCPU(coder, &bitStream, &symbolStreamReconst, (uint)symbolStream.size(), offsetInterval);
				printf(" comp...");
				compareSymbolStreams(symbolStream, symbolStreamReconst);
				printf(" done.\n");
			}
		}

		factor /= 2;
	}


	destroyInstance(pInstance);

	cudaSafeCall(cudaFree(dpDWT[1]));
	cudaSafeCall(cudaFree(dpDWT[0]));

	for(uint i = 0; i < blockCount; i++) {
		cudaSafeCall(cudaFree(dpSymbolStreams[i]));
	}

	cudaSafeCall(cudaFree(dpData));

	delete[] pData;


	for(ECoder coder = ECoder(0); coder < CODER_COUNT; coder = ECoder(coder + 1)) {
		uint bytes = (sizeTotal[coder] + 7) / 8;
		float kB = float(bytes) / 1024.0f;
		float factor = float(size.volume() * sizeof(float)) / float(bytes);
		printf("Total size %20s: %10u b %10u B %9.2f kB %5.1f:1\n", getCoderName(coder).c_str(), sizeTotal[coder], bytes, kB, factor);
	}


	return true;
}


#pragma warning( pop )
